#!NOMODULE
// Remote access backend.

// $Id: acbackend_remote.pike,v 1.7 1999/08/17 00:59:41 mast Exp $

#include <config.h>
#include <module.h>
inherit "module";
inherit "roxenlib";
inherit "aclib";

inherit ACDBBackend;

constant handshake_timeout = 5;	// Approx max seconds to complete handshake.
constant retry_interval = 120;	// Seconds between connect retries.
constant heartbeat = 900;	// Seconds between client side keepalive checks.

object acadm;
function(object(AC_DB):void) init_callback;

#ifdef AC_REMOTE_DEBUG
#define DEBUGMSG(msg) catch (roxen_perror (get_modname (acadm) + ": "+ (msg) + "\n"))
#else
#define DEBUGMSG(msg)
#endif

string shared_key, remote_id;

void create (object _acadm, object(AC_DB) _acdb,
	     function(object(AC_DB):void) _init_callback,
	     string _shared_key, string _remote_id)
{
  acadm = _acadm;
  ACDBBackend::create (_acdb);
  init_callback = _init_callback;
  shared_key = _shared_key;
  remote_id = _remote_id;
  acdb->local_id = lambda() {return remote_id;};
}


// Override Remote.context to get lookup to the local acdb.

class ACDBRemoteContext
{
  inherit Remote.context;

  void create (object ctx)
  {
    foreach (indices (ctx), string id) catch (this_object()[id] = ctx[id]);
  }

#define CTX_AC_RECORD 5321
#define CTX_AC_TABLE 5322

  array encode (mixed val)
  {
    if (objectp (val)) {
      object table;
      if (val->ac_record == "ac_record" && (table = val->table()))
	return ({CTX_AC_RECORD, table->name, val->id()});
      else if (val->ac_table == "ac_table")
	return ({CTX_AC_TABLE, val->name});
      else
	return ::encode (val);
    }
    else
      return ::encode (val);
  }

  mixed decode (array a)
  {
    switch (a[0]) {
      case CTX_AC_RECORD:
	return acdb[a[1]][a[2]];
      case CTX_AC_TABLE:
	return acdb[a[1]];
      default:
	return ::decode (a);
    }
  }
}


// Data.

int want_server, want_client;

object(Remote.Client) remclnt;
object(Remote.Server) remsrvr;

string srvrhost; int srvrport;

object server;
array(object) clients = ({});

object(Remote.connection) server_conn;
array(object(Remote.connection)) client_conns = ({}); // Same order as in clients.

void close_server()
{
#ifdef AC_REMOTE_DEBUG
  if (remsrvr) DEBUGMSG ("Closing server");
#endif
  object(Remote.Server) srvr = remsrvr;
  remsrvr = 0;
  catch (srvr->close_all());
}

void close_client()
{
#ifdef AC_REMOTE_DEBUG
  if (remclnt) DEBUGMSG ("Closing client");
#endif
  object(Remote.Client) clnt = remclnt;
  remclnt = 0;
  catch (clnt->close());
}

void close_callback (object(Remote.connection) conn)
{
  if (conn == server_conn) {
    if (remclnt)
      report_notice ("AC Remote: Lost connection to server " +
		     srvrhost + ":" + srvrport + "\n");
    invalidate_bind (server);
    server = 0;
    server_conn = 0;
    if (want_client) {
      remove_call_out (client_on);
      call_out (client_on, retry_interval, srvrhost, srvrport);
    }
  }
  else {
    if (remsrvr)
      report_notice ("AC Remote: Lost connection to client on " +
		     acadm->format_netaddr (conn->con->query_address(), 1) + "\n");
    int i = search (client_conns, conn);
    invalidate_bind (clients[i]);
    clients = clients[..i-1] + clients[i+1..];
    client_conns = client_conns[..i-1] + client_conns[i+1..];
  }
}


// Handshake.

// This obscure way of accessing authenticate_me() in Handshake is
// necessary since it must be static to not be callable remotely.
static function handshake;

static void got_connection (object rem, object(Remote.connection) conn)
{
  DEBUGMSG ("got_connection");
  if (!conn) {		// Client side.
    server = rem;
    conn = server_conn = remclnt->con;
  }
  else if (search (clients, rem) < 0) {
    report_notice ("AC Remote: Accepted connection from client on " +
		   acadm->format_netaddr (conn->con->query_address(), 1) + "\n");
    clients += ({rem});
    client_conns += ({conn});
    conn->add_close_callback (close_callback, conn);
  }
  conn->ctx = ACDBRemoteContext (conn->ctx);
}

static class Handshake
{
  // Challenge-response protocol taken from Applied Cryptography 2nd
  // ed, Bruce Schneier, p 54.

  static object secret;
  static object(Remote.connection) conn;

  static void create (void|object(Remote.connection) _conn)
  {
    DEBUGMSG ("Handshake.create");
    conn = _conn;
    secret = function_object (got_connection);
    if (!conn) handshake = authenticate_me;
  }

  static class Data
  {
    string request;
    string make_request()
    {
      return request = Crypto.sha()->update (
	Crypto.randomness.reasonably_random()->read (8) + shared_key)->digest();
    }
    string challenge;
    string make_challenge()
    {
      return challenge = Crypto.randomness.reasonably_random()->read (8);
    }
    string response()
    {
      if (!request || !challenge || sizeof (challenge) != 8)
	throw (({"AC Remote: Protocol error.\n", backtrace()}));
      return Crypto.sha()->update (request + challenge + shared_key)->digest();
    }
    int trusted;
    object secret;
    int created;
    void create() {created = time();}
  }

  static mapping(object(Handshake):object(Data)) connections = ([]);

  static void close_connection (object(Handshake) other)
  {
    DEBUGMSG ("close_connection");
    if (conn && conn->close) conn->close();
    m_delete (connections, other);
    if (!sizeof (connections)) remove_call_out (garb);
  }

  static void garb()
  {
    int limit;
    if (catch (limit = time (1) - handshake_timeout)) limit = time (1);
    foreach (indices (connections), object(Handshake) other) {
      object(Data) data = connections[other];
      if (data->created <= limit) {
	close_connection (other);
	catch (acadm->failed_connections++);
	DEBUGMSG ("closed old connection");
      }
    }
    if (sizeof (connections)) call_out (garb, handshake_timeout);
  }

  static object(Data) get_data (object(Handshake) other)
  {
    object(Data) data = connections[other];
    if (!data) {
      DEBUGMSG ("created new data record");
      data = Data();
      if (!sizeof (connections)) call_out (garb, handshake_timeout);
      connections[other] = data;
    }
    return data;
  }

  static object authenticate_me (object(Handshake) other)
  {
    DEBUGMSG ("authenticate_me");
    object(Data) data = get_data (other);
    data->challenge = other->make_challenge (this_object(), data->make_request());
    return other->check_response (this_object(), data->response(),
				  data->trusted && secret);
  }

  string make_challenge (object(Handshake) other, string request)
  {
    DEBUGMSG (sprintf ("make_challenge %O", request));
    object(Data) data = get_data (other);
    data->request = request;
    return data->make_challenge();
  }

  object check_response (object(Handshake) other, string response,
			 void|object other_secret)
  {
    DEBUGMSG (sprintf ("check_response %O", response));
    object(Data) data = connections[other];
    if (data) {
      if (response == data->response()) {
	// We trust other. Now let it authenticate us if we didn't get its secret.
	DEBUGMSG ("check_response ok");
	int authme = !other_secret && !data->trusted;
	data->trusted = 1;
	if (authme) other_secret = authenticate_me (other);
	if (other_secret) {
	  got_connection (other_secret, conn);
	  m_delete (connections, other);
	  if (!sizeof (connections)) remove_call_out (garb);
	}
	return secret;
      }
    }
    close_connection (other);
    acadm->failed_connections++;
    DEBUGMSG ("closed invalid connection");
    return 0;
  }
}


// Helpers.

void server_error (object client, array err, string msg)
{
  int i = search (clients, client);

  string name = "(unknown)";
  catch (name = acadm->format_netaddr (client_conns[i]->con->query_address(), 1));
  msg = replace (msg, "$client", name);
  report_error (msg + ": " + err[0]);
#ifdef AC_REMOTE_DEBUG
  roxen_perror (describe_backtrace (err));
#endif

  catch (client_conns[i]->close());

  // If the connection already is closed, close_callback() won't be
  // called to do this.
  i = search (clients, client);
  if (i >= 0) {
    clients = clients[..i-1] + clients[i+1..];
    client_conns = client_conns[..i-1] + client_conns[i+1..];
  }
}

void client_error (array err, string msg)
{
  report_error (msg + ": " + err[0]);
#ifdef AC_REMOTE_DEBUG
  roxen_perror (describe_backtrace (err));
#endif
  close_client();
  remove_call_out (client_on);
  call_out (client_on, retry_interval, srvrhost, srvrport);
}

static mapping(object:object) remobjs = ([]);

object check_destructed (object obj)
// Checks locally and remote whether obj is destructed. If it is, it's
// destructed locally too. Returns 0 if it's destructed, otherwise the
// object for the server it exists in. Something like this should
// probably be a service in Remote.
{
  if (!obj) {
    DEBUGMSG ("check_destructed: Object doesn't exist locally");
    return 0;
  }
  object rem;
  if (zero_type (rem = remobjs[obj]) == 1) {
    if (server_conn && search (server_conn->ctx->other, obj))
      rem = remobjs[obj] = server;
    else for (int i = 0; i < sizeof (client_conns); i++)
      if (search (client_conns[i]->ctx->other, obj)) {
	rem = remobjs[obj] = clients[i];
	break;
      }
    if (!rem) rem = remobjs[obj] = 0; // Is local.
  }
  if (!rem) {
    DEBUGMSG ("check_destructed: Object exists locally");
    return this_object();
  }
  if (!catch (rem = rem->check_destructed (obj)) && rem) {
    DEBUGMSG ("check_destructed: Object exists remote");
    return rem;
  }
  DEBUGMSG ("check_destructed: Object doesn't exist remote; destructed locally");
  destruct (obj);
  return 0;
}


// Resolver.

static mapping(string|object:object|string) bind;

object|string resolve (string|object what, void|int verify)
{
  if (!bind) {
    DEBUGMSG ("resolve: Resolver not initialized");
    return 0;
  }

  object|string res = bind[what];
  if (res && (!verify || check_destructed (objectp (res) ? res : what))) {
    DEBUGMSG ("resolve: " + (stringp (what) ?
			     "Object for " + what + " found in local bind" :
			     "Object resolved to " + res + " from local bind"));
    return res;
  }
  m_delete (bind, what);

  if (stringp (what)) {
    if (!server || catch (res = server->resolve (what)) || !res) {
      DEBUGMSG ("resolve: Object for " + what + " not found");
      return 0;
    }
    DEBUGMSG ("resolve: Object for " + what + " found in server");
    return bind[what] = res;
  }

  else {
    object rem = check_destructed (what);
    if (!rem) {
      m_delete (bind, what);
      return 0;
    }
    if (rem == this_object())
      return bind[what] = remote_id;
    return bind[what] = rem->resolve (what);
  }
}

static multiset(string) pre_bind_ids = (<>);
// Used to detect connection loops.

int register_id (string id, object rem)
{
  if (server) {
    if (pre_bind_ids[id]) {
      DEBUGMSG ("register_id: Remote connection loop");
      return -1;
    }
    pre_bind_ids[id] = 1;
    int res;
    array err = catch (res = server->register_id (id, rem));
    pre_bind_ids[id] = 0;
    if (err) throw (err);
    if (res <= 0) {
      DEBUGMSG ("register_id: Server rejects " + id + ": " + res);
      return res;
    }
  }
  else
    if (bind[id] && bind[id] != rem && check_destructed (bind[id])) {
      DEBUGMSG ("register_id: Remote ID " + id + " already in use");
      return 0;
    }

  bind[id] = rem;
  DEBUGMSG ("register_id: Remote ID " + id + " added");
  return 1;
}

static void init_id()
{
  array(string) client_ids;
  if (bind) client_ids = Array.filter (indices (bind), stringp);
  else bind = ([]);

  if (!sizeof (remote_id)) {
    remote_id = "";
    int res;
    do remote_id += sprintf ("%c", random ('z' - 'a' + 1) + 'a');
    while (!(res = register_id (remote_id, this_object())));
    if (res < 0) {
      close_client();
      throw (({"Remote connection loop detected\n", backtrace()}));
    }
    acadm->set_remote_id (remote_id);
  }
  else {
    int res = register_id (remote_id, this_object());
    if (res <= 0) {
      close_client();
      throw (({res < 0 ? "Remote connection loop detected\n" :
	       "Remote ID " + remote_id + " already in use\n", backtrace()}));
    }
  }
  acdb->remote_resolve = resolve;
  // acdb->local_id already fixed in create().

  if (client_ids)
    foreach (client_ids, string id)
      if (id != remote_id && !register_id (id, bind[id])) {
	close_client();
	throw (({"Client remote ID " + id + " already in use\n", backtrace()}));
      }
}

void invalidate_bind (object stale, void|string id)
{
  if (!sizeof (remote_id) || !bind[remote_id]) return; // Ignore if we're not registered.
  if (!id) {
    if (server && server != stale) {
      DEBUGMSG ("invalidate_bind: Propagating to server");
      catch (server->invalidate_bind (stale));
      return;
    }
    id = search (bind, stale);
    if (!id) {
      DEBUGMSG ("invalidate_bind: ID for stale object not found");
      return;
    }
  }

  DEBUGMSG ("invalidate_bind: Removing stale id " + id);
  m_delete (bind, id);
  object obj;
  while (obj = search (bind, id)) m_delete (bind, obj);

#ifdef AC_REMOTE_DEBUG
  if (sizeof (clients))
    DEBUGMSG ("invalidate_bind: Broadcasting to " + sizeof (clients) + " clients");
#endif
  foreach (clients, object client)
    if (client != stale) catch (client->invalidate_bind->async (stale, id));
}


// Updating.

static int _updatecount;
// Total number of updates in the server since remote server init.

mapping(string:multiset(int)) updated = ([]);
// The records in every table that has been updated on the server.

int updatecount()
{
  DEBUGMSG ("Reporting updatecount " + _updatecount +
	    " (" + sizeof (bind) + " entries in bind)");
  return _updatecount;
}

string dump_values (void|string s_ids)
{
  mapping(string:multiset(int)) ids = s_ids && decode_value (s_ids);

  DEBUGMSG ("Dumping " + (ids ? "record(s) in " + indices (ids) * ", " : "complete db"));

  mapping(string:mapping(int:array)) val;
  {
#ifdef THREADS
    // Won't touch mac here, so we don't have to observe its locking order.
    object dblock;
    catch (dblock = acdb->lock());
#endif
    val = acdb->dump_values (ids);
  }
  return encode_value (val);
}

void assign_values (string s_val)
{
  mapping(string:mapping(int:array)) val = decode_value (s_val);

  DEBUGMSG ("Got update(s) in " + indices (val) * ", " + " from client");

#ifdef THREADS
  // Won't touch mac here, so we don't have to observe its locking order.
  object dblock;
  catch (dblock = acdb->lock());
#endif
  acdb->assign_values (val);
  acdb->commit();
}

void notify_updated (string s_ids)
{
  mapping(string:multiset(int)) ids = decode_value (s_ids);

  DEBUGMSG ("Got notification of updated record(s) in " + indices (ids) * ", ");
  foreach (indices (ids), string table)
    if (updated[table]) updated[table] |= ids[table];
    else updated[table] = ids[table];
  _updatecount++;

#ifdef AC_REMOTE_DEBUG
  if (sizeof (clients))
    DEBUGMSG ("Broadcasting notification of updated record(s) in " +
	      indices (ids) * ", " + " to " + sizeof (clients) + " clients");
#endif
  foreach (clients, object client) {
    array err = catch (client->notify_updated->async (s_ids));
    if (err) server_error (client, err, "AC Remote: Error notifying client $client");
  }
}

void get_db (void|int is_initialized)
{
  DEBUGMSG ("Requesting complete db transfer from server");

  // Reinitialize unconditionally so that no records from other
  // backends exists.
  acdb->initialize();

  array err = catch {
    // FIXME: Divide this transfer into smaller pieces to avoid
    // potential blocking.
    acdb->assign_values (decode_value (server->dump_values()));
    updated = ([]);
    _updatecount = server->updatecount();
    acdb->commit (commit);
  };
  if (err) client_error (err, "AC Remote: Error reading db from server");
}

void begin (void|int fromscratch)
{
  if (!fromscratch && !sizeof (updated)) return;

  if (server)
    if (fromscratch) {
      DEBUGMSG ("Local db reinitialized");
      get_db (fromscratch);
    }

    else {
      int server_updatecount;
      array err = catch (server_updatecount = server->updatecount());
      if (err) client_error (err, "AC Remote: Error querying server");

      if (_updatecount != server_updatecount) {
	DEBUGMSG ("Updatecount mismatch (local: " + _updatecount + ", server: " +
		  server_updatecount + ")");
	get_db (fromscratch);
      }

      else {
	DEBUGMSG ("Requesting updated record(s) in " + indices (updated) * ", ");
	array err = catch {
	  acdb->assign_values (decode_value (
	    server->dump_values (encode_value (updated))));
	  updated = ([]);
	  acdb->commit (commit);
	};
	if (err)
	  client_error (err, "AC Remote: Error reading updated record(s) from server");
      }
    }
}

int commit (mapping(string:multiset(int)) ids)
{
  ids = ids + ([]);
  foreach (indices (ids), string table)
    if (!sizeof (ids[table])) m_delete (ids, table);

  if (server) {
    DEBUGMSG ("Pushing updated record(s) in " + indices (ids) * ", " + " to server");
    array err = catch {
      // Has to be done async to avoid possible deadlocks.
      server->assign_values->async (encode_value (acdb->dump_values (ids)));
      // The server will broadcast a notification now and we will
      // reread the record(s) from it unnecessarily, but it's not
      // worth bothering about.
    };
    if (err) {
      client_error (err, "AC Remote: Error updating record(s) to server");
      return 0;
    }
  }
  else {
    // Only notify if we're the root server.
    notify_updated (encode_value (ids));
  }

  return 1;
}


// Misc.

string name()
{
  //DEBUGMSG ("name");
  object conf = acadm->config;
  return conf->name +
    (conf->query_name() != conf->name ? " (" + conf->query_name() + ")" : "");
}

object(AC_DB) get_acdb()
// For remote use.
{
  //DEBUGMSG ("get_acdb");
  return acdb;
}


// Init and shutdown.

void server_on (string interface, int port)
{
  want_server = 1;

  if (!remsrvr) {
    close_server();
    array err = catch {
#if __VERSION__ < 0.7
      remsrvr = Remote.Server (
	lower_case (interface || "any") != "any" && interface, port);
#else
      remsrvr = Remote.Server (
	lower_case (interface || "any") != "any" && interface, port, 4);
#endif
      init_id();
      remsrvr->provide ("AC Remote 1", Handshake);
    };
    if (err) {
      report_error ("AC Remote: Error opening server port " + port + ": " + err[0]);
#ifdef AC_REMOTE_DEBUG
      roxen_perror (describe_backtrace (err));
#endif
      close_server();
      return;
    }
    report_notice ("AC Remote: Listening to " +
		   replace (remsrvr->port->query_address(), " ", ":") + "\n");
    _updatecount = 1;
  }
}

void server_off()
{
  want_server = 0;
  if (remsrvr) {
    close_server();
    report_notice ("AC Remote: Closed server port\n");
  }
}

void client_on (string host, int port)
{
  want_client = 1;
  srvrhost = host, srvrport = port;

#ifdef AC_REMOTE_DEBUG
  if (server) DEBUGMSG ("checking connection (" + sizeof (bind) + " entries in bind)");
#endif
  int server_updatecount;
  array err = catch (server_updatecount = server->updatecount());

  if (err) {
    close_client();
    err = catch {
#if __VERSION__ < 0.7
      remclnt = Remote.Client (
	host, port, 0, 10);	// Rather short timeout since this'll block.
#else
      remclnt = Remote.Client (
	host, port, 0, 10, 4);	// Rather short timeout since this'll block.
#endif
      if (!handshake) Handshake();
      if (!handshake (remclnt->get ("AC Remote 1")))
	throw (({"Authentication failed\n", backtrace()}));
      init_id();
      server_conn->add_close_callback (close_callback, server_conn);
      server_updatecount = server->updatecount();
    };
    if (err) {
      client_error (err, "AC Remote: Error connecting to server " + host + ":" + port);
      return;
    }
    report_notice ("AC Remote: Got remote connection to " + host + ":" + port + "\n");
    _updatecount = 0;
  }

  remove_call_out (client_on);
  call_out (client_on, heartbeat, host, port);

  if (_updatecount != server_updatecount) {
#ifdef THREADS
    // Won't touch mac here, so we don't have to observe its locking order.
    object dblock;
    catch (dblock = acdb->lock());
#endif
    begin();
    if (_updatecount != server_updatecount) {
      DEBUGMSG ("Updatecount mismatch (local: " + _updatecount + ", server: " +
		server_updatecount + ")");
      get_db();
#ifdef THREADS
      catch (destruct (dblock));
#endif
      if (acdb->initialized) init_callback (acdb);
    }
  }
}

void client_off()
{
  want_client = 0;
  if (remclnt) {
    close_client();
    report_notice ("AC Remote: Closed remote connection\n");
  }
  remove_call_out (client_on);
}

void destroy()
{
  server_off();
  client_off();
}
