//  WebLayout Templates
//
//  Author:  Jonas Walldn, jonasw@idonex.se


//  Suggestions
//
//  - An implementation of replace_nocase in C for higher performance.
//
//  - A nice addition would be a way to test for macro arguments
//    (e.g. if they're undefined or equal to some specific value) and
//    show/hide some elements as a result.
//
//  - Batch mode for storing result in file. This is not trivial since we
//    need the Roxen environment (or emulate it?) for calling parsing
//    functions.
//
//  - Add support for nested <template> containers in the content
//    file. This works intermittently now, but caching may prevent the
//    latest version from appearing. An idea is to extract and process
//    all <template> tags (calling parse_html?) before computing MD5.


//  Undocumented features (except for here, that is...)
//
//  - There is a BASE attribute to <template> which prefixes the search
//    path for the template directory set in the configuration pages. This
//    means we'll look in the directory BASE+QUERY(tmpl_storage_path) for
//    template files.
//
//  - The tag <tmplinsertexcept tag="tag1,tag2,..."> inserts all data in
//    the current source container except for those tags given as input
//    parameters.
//
//  ...and possibly even more which is only mentioned in the source below!


string cvs_version = "$Id: weblayout.pike,v 1.21 1999/08/26 06:37:51 wing Exp $";


#include <module.h>
inherit "module";
inherit "roxenlib";

import Stdio;
import Array;
import Stack;



#define GAUGE_PERF         0   //  Set to 1 to output timing info in log
#define DISABLE_CACHE      0   //  Set to 1 to disable caching
#define COMMENT_STRIPPING  0   //  Set to 1 to enable optional removal of
                               //  template comments (still too slow... :-( )
#define OLD_GLOBALS        0   //  Set to 1 to disable thread-safe globals
#define HONOR_RELOAD	   1   //  Set to 1 to let reloads override the cache 


#if !OLD_GLOBALS
  //  Enable multithreading if we know all globals are removed. Note that
  //  all globals are per-request and not shared among several requests
  //  so we never risk any race conditions.
  constant  thread_safe = 1;
#endif




//  ===================================================================
//
//  Configuration interface code


array (mixed) register_module()
{
  return ({
    MODULE_PARSER | MODULE_LOCATION | MODULE_PROVIDER,
      "Templates",
      
      //  Special hack supported by the configuration interface; if there
      //  is a newline before the first <p> the remaining text can be
      //  shown or hidden using a button.
      "Templates implements template-based web pages. The template "
      "and content data are kept in separate files and merged on-the-fly "
      "when a page is requested.\n<p>"
      
      "Templates supports the following tags for content files:<p>"

      "<ul><li><b>&lt;template [name=<i>template_file</i>]&gt;</b> and "
      "<b>&lt;/template&gt</b>. These tags signal that the data "
      "within uses tags defined by the template <i>template_file</i>. "
      "If <i>template_file</i> is unspecified the default template "
      "(chosen by the administrator) is used. If <i>template_file</i> "
      "is an absolute path it will override the normal template path. To "
      "override the template settings for a content file, append "
      "<b>?tmpl=<i>template_file</i></b> to the URL when accessing the file."
      "</ul><p>"

      "For template files the following tags are added:<p>"

      "<ul><li><b>&lt;tmplinsert tag=<i>custom_tag</i>&gt;</b>. This tag "
      "inserts the data contained in the first occurrence of "
      "<b>&lt;<i>custom_tag</i>&gt;</b> <i>data</i> <b>&lt;/<i>custom_tag"
      "</i>&gt;</b> at the current location in in the layout template. Use "
      "the predefined names <b>tmpldatafile</b> and <b>tmpldatafile_raw</b> "
      "to insert the name of the content file currently being processed "
      "(the latter being the URL before the stripping of the optional "
      "<b>?tmpl</b> argument). "

      "<li><b>&lt;tmplinsertall&gt;</b> processes and inserts all "
      "content data. It is only allowed outside of template blocks."
      
      "<li><b>&lt;tmpldefault tag=<i>custom_tag</i>&gt;</b> and <b>&lt;"
      "/tmpldefault&gt;</b> define a container for providing default "
      "data in case the content file doesn't contain any instances of "
      "<b>&lt;<i>custom_tag</i>&gt;</b>."

      "<li><b>&lt;tmplblock tag=<i>custom_tag</i> [params=<i>name</i>,...,"
      "<i>name</i>] [singletag]&gt;</b> and <b>&lt;/tmplblock&gt;</b>. This "
      "container defines a block whose contents can be inserted repeatedly. "
      "Inside this block there should be a string <b>$<i>custom_tag</i>$</b> "
      "that outputs the content data. Insertion of other blocks is also "
      "allowed. The parameters given to <b>params</b> act as substitution "
      "strings within the block; insert the value using the syntax "
      "<b>$<i>name</i>$</b>. Parameters passed from the content file tag "
      "but not specified in the comma-separated list can be inserted using "
      "the word <b>$varargs$</b>. A block can be redefined to allow for "
      "repeated data insertion with varying appearance. Including the "
      "attribute <b>singletag</b> specifies that the tag doesn't need a "
      "closing tag (such as e.g. <b>&lt;hr&gt;</b>); in this case "
      "the content data will be empty. Redefining built-in HTML and RXML "
      "tags is supported."
      
      "<li><b>&lt;tmpldefaultparam param=<i>name</i>&gt;</b> and <b>&lt;"
      "/tmpldefaultparam&gt;</b> is only valid within a template block "
      "definition. It will set a default value for a parameter within "
      "that block in case no value is given in the content file."

      "<li><b>&lt;tmplinsertblock tag=<i>block_name</i>,...,<i>block_name</i>"
      "&gt;</b>. This takes a comma-separated list of arguments, each one "
      "being a name of a block definition. The data will be extracted from "
      "the matching containers in the content file and inserted using the "
      "appropriate block definition. "

      "</ul>"
      
      //  Ok, this isn't pretty, but it works...
      "<p><i>Example:</i><p>"
      "<table border=0 cellspacing=0 cellpadding=0>"
      "<tr><td><b>Template file</b></td><td><b>Content file</b></td></tr>"
      "<tr><td colspan=3><hr></td></tr>"
      "<tr><td valign=top>"
      "<font size=-1>"
      "<pre>"
      "&lt;html&gt;&lt;body&gt;</pre><p><pre>"
      "&lt;!-- Blocks for heading and text --&gt;<br>"
      "&lt;tmplblock tag=heading&gt;<br>"
      "  &lt;b&gt;$heading$&lt;/b&gt;&lt;p&gt;<br>"
      "&lt;/tmplblock&gt;<br>"
      "&lt;tmplblock tag=text&gt;<br>"
      "  $text$&lt;p&gt;<br>"
      "&lt;/tmplblock&gt;</pre><p><pre>"
      "&lt;!-- Block for chapter using headings and texts --&gt;    <br>"
      "&lt;tmplblock tag=chapter params=chname&gt;<br>"
      "  &lt;h2&gt;&lt;hr noshade size=2&gt;$chname$&lt;hr noshade size=1&gt;"
         "&lt;/h2&gt;     <br>"
      "  &lt;tmplinsertblock tag=heading,text&gt;&lt;p&gt;<br>"
      "&lt;/tmplblock&gt;</pre><p><pre>"
      "&lt;!-- Insert all chapters in content file --&gt;<br>"
      "&lt;tmplinsertblock tag=chapter&gt;</pre><p><pre>"
      "&lt;/body&gt;&lt;/html&gt;</pre>"
      "</font>"
      "</td><td valign=top>"
      "<font size=-1>"
      "<pre>"
      "&lt;template mytemplate.tmpl&gt;</pre><p><pre>"
      "  &lt;chapter chname=\"Introduction\"&gt;<br>"
      "    &lt;heading&gt;Welcome to Hello World.&lt;/heading&gt;<br>"
      "    &lt;text&gt;Thank you for buying <br>"
      "          Hello World v1.0.&lt;/text&gt;<br>"
      "  &lt;/chapter&gt;</pre><p><pre>"
      "  &lt;chapter chname=\"Reference\"&gt;<br>"
      "    &lt;heading&gt;Command reference&lt;/heading&gt;<br>"
      "    &lt;text&gt;Enter &lt;tt&gt;HelloWorld&lt;/tt&gt; to<br>"
      "          run the program.&lt;/text&gt;<br>"
      "  &lt;/chapter&gt;</pre><p><pre>"
      "&lt;/template&gt;</pre>"
      "</font>"
      "</td></tr>"
      "</table>",
      
      0,
      1 });
}


void create()
{
  //  Administration pages
  defvar("tmpl_config_dir",
	 "/template_admin/",
	 "Path to administration pages",
	 TYPE_LOCATION,
	 "Location in the namespace of the virtual file system where "
	 "administration pages for templates are found. Leave empty to "
	 "disable the administration interface entirely.");

  //  Template repository
  defvar("tmpl_storage_path",
	 "/templates/",
	 "Template storage directory",
	 TYPE_LOCATION,
	 "Location in the file system where templates are stored.");

  //  Default template name
  defvar("tmpl_default_name",
	 "default.tmpl",
	 "Default layout template",
	 TYPE_FILE,
	 "Default layout template file used whenever no other template "
	 "file name is provided in a content file.");

#if COMMENT_STRIPPING
  //  Stripping of template comments
  defvar("tmpl_strip_comments",
         0,
         "Strip template comments?",
         TYPE_FLAG,
         "Flag specifying whether &lt;!-- --&gt; comments will be "
	 "removed from the layout template file.");
#endif
  
  //  Error logging
  defvar("tmpl_log_errors",
	 1,
	 "Log errors?",
	 TYPE_FLAG,
	 "Enable to log Templates error messages in the server "
	 "log file.");
}



void start()
{
}


void stop()
{
}


string query_location()
{
  return sizeof (QUERY (tmpl_config_dir)) && QUERY(tmpl_config_dir);
}




//  ===================================================================
//
//  Utility functions


void log_err(string err, void|object req_id)
{
  //  Write string to server log if user configured it this way.
  if (QUERY(tmpl_log_errors)) {
    //  If available we append the name of the current data file to ease
    //  tracking down the problem. If possible we also extract the path.
    if (req_id) {
      err += " [Content file URL \"" + req_id->not_query + "\"]";
    }
    werror(err + "\n");
  }
}


string replace_nocase(string data, string from, string to)
{
  string  data_lc, result;
  int     from_len, pos, search_pos;

  //  Make from string lower-case for easier finding, and make a
  //  lower-case copy of data.
  from = lower_case(from);
  data_lc = lower_case(data);
  from_len = strlen(from);
  search_pos = 0;
  
  //  Find all positions
  result = "";
  while ((pos = search(data_lc, from, search_pos)) > -1) {
        result += data[search_pos..(pos - 1)] + to;
        search_pos = pos + from_len;
  }
  result += data[search_pos..];
  return result;
}


string strip_surrounding_ws(string data)
{
  int  startpos = 0,
       endpos = strlen(data) - 1,
       last_was_lf = 0;
  
  //  Move starting and ending pointers until all whitespace is removed.
  //  If the last character is a linefeed we remember to append one to
  //  the final result.
  if ((strlen(data) > 0) && (data[-1] == '\n'))
    last_was_lf = 1;
  while ((startpos < endpos)
	 && ((data[startpos] == ' ') || (data[startpos] == '\n')
	      || (data[startpos] == '\t')))
    startpos++;
  while ((endpos > startpos)
	 && ((data[endpos] == ' ') || (data[endpos] == '\n')
	  || (data[startpos] == '\t')))
    endpos--;
  return data[startpos..endpos] + (last_was_lf ? "\n" : "");
}


#if COMMENT_STRIPPING
//  This routine is too slow to be practical...
string strip_comments(string data)
{
  string  result = "";
  int     lastpos, pos, len;

  //  Run through the string and remove all comments starting with
  //  <!-- and ending with -->.
  len = strlen(data);
  lastpos = 0;
  for (pos = 0; pos < len - 4; pos++) {
    if (data[pos..(pos+3)] == "<!--") {
      //  Output all text up until now and look for end of comment
      result += data[lastpos..(pos-1)];
      while (pos < len - 3) {
        if (data[pos..(pos+2)] == "-->") {
          pos += 3;
          lastpos = pos;
          break;
        }
        pos++;
      }
    }
  }
  
  //  Output final part
  result += data[lastpos..len];
  return result;
}
#endif



//  ===================================================================
//
//  Template administration


mapping get_custom_tags(string data)
{
  function callback_container_fn = 
             lambda(string tag, mapping args, string contents, array result)
             {
	       string  a_tag;
	       array   tag_list;
	       
	       //  Verify that the tag attribute is included. If found
	       //  we check for a parameter list in block definitions
	       //  as well.
	       if (args->tag) {
		 if (tag == "tmplblock") {
		   array   params;

		   //  Return parameter list as an array
		   if (zero_type(args->params))
		     params = ({});
		   else
		     params = (((lower_case(args->params) - " ") / ",")
			       - ({ "" }) );
		   result[0] += ([ lower_case(args->tag) : params ]);
		 } else {
		   //  Extract tag names from comma-separated list. Only
		   //  add if no earlier definition since parameters
		   //  otherwise might be lost.
		   tag_list = ((lower_case(args->tag) - " ") / ",") - ({ "" });
		   foreach (tag_list, a_tag)
		     if (zero_type(result[0][a_tag]))
		       result[0] += ([ a_tag : ({}) ]);
		 }
	       }
	       return "";
	     };
  function callback_tag_fn =
             lambda(string tag, mapping args, array result, function fn)
	     {
	       //  Just reuse the code for containers...
	       fn(tag, args, "", result);
	       return "";
	     };

  //  To have the callbacks accumulate the result without global variables
  //  we put the result inside an array that we pass to parse_html.
  array result = ({ ([]) });
  
  //  Extract names of custom tags by parsing the template. Only names
  //  passed to <tmplinsert> and <tmplblock> are interesting since all
  //  tags in <tmplinsertblock> will be found through the block definitions.
  //  We pass callback_container_fn as a parameter to the handler function
  //  so the tag handler can call the container handler directly.
  parse_html(data, ([ "tmplinsert" : callback_tag_fn ]),
	           ([ "tmplblock"  : callback_container_fn ]),
	     result, callback_container_fn);
  
  //  Remove tmpldatafile and tmpldatafile_raw since these are defined by
  //  WebLayout Templates.
  result[0] -= ([ "tmpldatafile" : "", "tmpldatafile_raw" : "" ]);
  return result[0];
}


//  This function returns help texts in a mapping formed as:
//  ([ "__foo"  : "Global documentation for this template",
//     "mytag1" : ([ "__foo" : "Help text for mytag1",
//                   "arg1"  : "Help text for arg1 of mytag1",
//                   "arg2"  : "Help text for arg2 of mytag1" ]),
//     "mytag2" : ([ "__foo" : "Help text for mytag2", ... ]), ...
//  ])
mapping get_help_texts(string data)
{
  function callback_container_fn = 
             lambda(string tag, mapping args, string contents, array result)
             {
	       string  args_tag, param_index;
	       
	       //  If the "tag" argument is missing we're handling the
	       //  documentation for the entire template; otherwise
	       //  it's for a given tag. If the "param" argument is
	       //  missing, this is the help text for the complete
	       //  tag; otherwise it's only for the specific argument.
	       if (args->tag) {
		 if (args->param)
		   //  Help text for specific parameter
		   param_index = lower_case(args->param);
		 else
		   //  Help text for entire tag
		   param_index = "__foo";
		 
		 //  Add to result mapping
		 args_tag = lower_case(args->tag);
		 if (zero_type(result[0][args_tag]))
		   result[0][args_tag] = ([ param_index : contents ]);
		 else
		   result[0][args_tag] += ([ param_index : contents ]);
	       } else {
		 //  This is the global documentation for the template
		 result[0]["__foo"] = contents;
	       }
	       return "";
	     };
  
  //  To have the callbacks accumulate the result without global variables
  //  we put the result inside an array that we pass to parse_html.
  array result = ({ ([ ]) });
  
  //  Extract help texts by parsing the template.
  parse_html(data, ([ ]), ([ "tmplhelp"  : callback_container_fn ]),
	     result);
  return result[0];
}


string low_view_template(string data)
{
  string    result, tag_name;
  mapping   custom_tags, help_texts;
  
  //  Get all custom tags
  custom_tags = get_custom_tags(data);

  //  Extract help texts
  help_texts = get_help_texts(data);
  
  //  Replace all '<' and '>' with appropriate codes to prevent the page
  //  from being messed up. Highlight template tags for easier finding in
  //  the code.
  data = replace(data, ({ "<", ">", "&" }), ({ "&lt;", "&gt;", "&amp;" }) );
  data = replace(data, ({ "tmplinsert",
			  "tmplinsertall",
			  "tmplinsertblock",
			  "tmplinsertexcept",
			  "tmpldefaultparam",
			  "tmpldefault",
			  "tmplblock",
                          "tmplhelp" }),
		 ({ "<b><font color=\"#ffffaa\">tmplinsert</font></b>",
		    "<b><font color=\"#ffffaa\">tmplinsertall</font></b>",
		    "<b><font color=\"#ffffaa\">tmplinsertblock</font></b>",
		    "<b><font color=\"#ffffaa\">tmplinsertexcept</font></b>",
		    "<b><font color=\"#ffffaa\">tmpldefaultparam</font></b>",
		    "<b><font color=\"#ffffaa\">tmpldefault</font></b>",
		    "<b><font color=\"#ffffaa\">tmplblock</font></b>",
		    "<b><font color=\"#ffffaa\">tmplhelp</font></b>" }) );
  
  //  Build list of tag names and optional block input parameters. Add
  //  template documentation if needed.
  result = "";
  if (help_texts["__foo"] && strlen(help_texts["__foo"]))
    result += "<h3>Documentation</h3>\n" +
              help_texts["__foo"] + "<p><hr>\n";
  result += "<h3>Custom tags</h3>\n"
            "<ul>\n"
            "<table border=0 cellpadding=0>\n";
  foreach(sort(indices(custom_tags)), tag_name) {
    //  Display bullet and tag name
    result += "<tr><td valign=top><li><b>" + tag_name +
              "&nbsp;&nbsp</b></td><td valign=top>\n";
    
    //  Add help text for tag, if any.
    if (!zero_type(help_texts[tag_name]))
      result += help_texts[tag_name]["__foo"];
    result += "</td></tr>\n";

    //  List all parameters and their corresponding help texts
    if (sizeof(custom_tags[tag_name]) > 0) {
      string  param_name;
      array   postpone_params = ({ });

      result += "<tr><td valign=top>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"
	        "<i>Parameters:</i></td></tr>\n";
      foreach(values(custom_tags[tag_name]), param_name) {
	//  Extract help text and verify that it's non-empty
	if (!zero_type(help_texts[tag_name]) &&
	    !zero_type(help_texts[tag_name][param_name])) {
	  result += "<tr><td valign=top>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;" +
	            param_name + "</td>\n"
	            "<td valign=top>" + help_texts[tag_name][param_name] +
	            "</td></tr>\n";
	} else {
	  //  Save this for the list of parameters for which there are
	  //  no documentation
	  postpone_params += ({ param_name });
	}
      }

      //  Now list all remaining parameters in a comma-separated list
      if (sizeof(postpone_params) > 0) {
	result += "<tr><td valign=top colspan=2>"
	          "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;" +
	          (postpone_params * ", ") + "</td></tr>\n";
      }
    }
  }
  
  result += "</table>\n"
            "</ul><hr>\n"
            "<h3>Template source</h3>\n"
	    "<pre>" + data + "</pre>";
  return result;
}


mapping view_template(string file_name, string file_path, object req_id)
{
  string    data, result;
  
  //  Open the file and read contents
  data = req_id->conf->try_get_file(file_path, req_id);

  if (!data) {
    string  err = "Templates error: Unable to load template "
                  "file \"" + file_name + "\".";
    log_err(err, req_id);
    return http_string_answer(err, "text/plain");
  }
  
  result = "<html><head><title>View Template</title></head>\n"
           "<body bgcolor=\"#000630\" text=\"#ffffff\""
	   "link=\"#ffffaa\" vlink=\"#ffffaa\">\n"
           "<h2>Layout template <i>" + file_name + "</i></h2>\n"
           "<hr>\n"+low_view_template(data)+"<hr>\n</body></html>";
  return http_string_answer(result, "text/html");
}


//  Someone is accessing our /template_admin/ namespace so we examine
//  the path to see what should be returned. We build these pages on
//  the fly.
object|mapping find_file(string file_name, object req_id)
{
  if (!sizeof (QUERY (tmpl_config_dir))) return 0;

  //  Generate index of available templates?
  //  werror("Filename: "+file_name+"\n");
  if (file_name == "" || file_name == "index.html") {
    string  result, a_file;
    array   files;

    //  Find all files in the template storage directory and sort
    //  the array. Remove all directories by filtering file-stats.
    files = req_id->conf->find_dir((req_id->variables->base||"")+
				   QUERY(tmpl_storage_path), req_id);
    if (files) {
      files = Array.filter(files,
		 lambda(string elem, object req_id) {
		   return (req_id->conf->
			   stat_file((req_id->variables->base||"")
				     +QUERY(tmpl_storage_path)
				     + elem, req_id)[1] != -2);
		 }, req_id);
      sort(files);
    }
    
    //  Output list of templates
    result = "<html><head><title>Templates Administration</title>\n"
             "</head><body bgcolor=\"#000630\" text=\"#ffffff\""
             "link=\"#ffffaa\" vlink=\"#ffffaa\">\n"
             "<h2>Templates Administration</h2>\n"
             "<hr>\n"
             "<h3>Available templates</h3>\n"
             "<ul>\n";
    if (files) {
      result += "<table border=0 cellpadding=0>\n";
      foreach(files, a_file) {
	result += "<tr><td>\n"
	          "<li><b>" + a_file + "</b></td><td>\n"
	          "&nbsp;&nbsp;<a href=\"" + a_file + "?view\">View</a>\n"
                  "</td></tr>\n";
      }
      result += "</table>\n";
    } else
      result += "(None found.)\n";
    result += "</ul>\n"
              "<hr></body></html>";
    return http_string_answer(result, "text/html");
  } else {
    string   file_path, cmd;

    //  Handle various commands that are appended to the file name. We also
    //  build a complete path name that we can pass to our helper functions.
    //  We strip any ".." characters to avoid accessing paths outside of the
    //  template directory tree.
    file_name -= "..";

    //  Fix for e.g. cvs filesystems. We must use try_get_file later instead.
    file_path = ((req_id->variables->base||"")+
		 QUERY(tmpl_storage_path) + file_name);
    
    cmd = req_id->query;
    if (!cmd)
      //  Command needed to perform operation
      return http_string_answer("Unknown command.", "text/plain");
    
    switch (cmd) {
      //  Viewing template tags?
      case "view":
        return view_template(file_name, file_path, req_id);
    }
    
    //  Unknown request
    return http_string_answer("Unknown command.", "text/plain");
  }
}




//  ===================================================================
//
//  Template parsing code

#if OLD_GLOBALS
mapping  gBlockDefs,   //  ([ "block1" : ({ paramname_mapping, def }),
                       //     "block2" : ({ paramname_mapping, def })  ])
         gDefaults;    //  ([ "tag1" : "defaultvalue1",
                       //     "tag2" : "defaultvalue2"  ])
string   gEntireSource;
string   gCurrSource;
object   gSourceStack;
#endif

#if OLD_GLOBALS
#define GLB(x,id)  (x)
#define GLB2(x)    (x)
#define GET_GLB(id)
#define GLB_INIT(id)
#define GLB_CLEAR(id)
#else
//  Access thread-safe globals in req_id->misc->weblayout->variable. To
//  speed up repeated access we store a reference to the mapping in a
//  local variable called _glb.
#define GET_GLB(id)     mapping _glb = id->misc->weblayout
#define GLB(x,id)       (id->misc->weblayout->x)
#define GLB2(x)         (_glb->x)
#define GLB_INIT(id)    id->misc->weblayout = ([])
#define GLB_CLEAR(id)   id->misc->weblayout = ([])
#endif



array get_custom_data(string src, array want_tags)
{
  mapping   want_containers;
  string    a_tag;
  function  callback_fn =
    lambda(string tag, mapping args, string contents, array result)
    {
      //  Add entry to result array
      result[0] += ({ ([ "tag"  : tag,
			 "args" : args,
			 "data" : contents ]) });
      return "";
    };

  //  Accumulate callback result as an array of data we extract from the
  //  source string. Each array element is formatted as
  //  ([ "tag" : "tag_name", "args" : ([ "a1" : "v1"]), "data" : "data_text" ])
  array result = ({ ({}) });
  
  //  Parse the src string for all tags that are listed in want_tags
  //  and return an array of mappings where each mapping has the format
  //  ([ "tag" : name_of_tag, "args" : mapping_of_args, "data" : tag_data ])
  want_containers = ([]);
  foreach (want_tags, a_tag)
    want_containers += ([ lower_case(a_tag) : callback_fn ]);
  parse_html(src, ([]), want_containers, result);
  return result[0];
}


//  Forward declarations
string tmplinsert_tag(string tag, mapping args, multiset recursion_chain,
		      object req_id);
string process_block(string data, string tag, mapping args,
		     multiset recursion_chain, object req_id);
string tmpldefaultparam_container(string tag, mapping args, string contents,
				  multiset rec_chain, object req_id,
				  string current_block);


//  Parse the given source data for custom block tags
string handle_content_insert(string data, multiset recursion_chain,
			     object req_id)
{
  mapping    custom_containers = ([]);
  string     block_name;
  function   callback_fn =
    //  Callback handler for custom tags. We re-use the code in
    //  tmplinsert_tag by wrapping the contents in a <XXX>..</XXX>
    //  container.
    lambda(string tag, mapping args, string contents,
	   multiset rec_chain, object req_id)
    {
      string  saveSource, res;
	      
      //  We need to save the current source and process the new source
      //  container. We remove any "_" prefix from the tag name as well.
      if (tag[0] == '_')
	tag = tag[1..];
      GET_GLB(req_id);
      saveSource = GLB2(gCurrSource);
      GLB2(gCurrSource) = contents;
      res = process_block(contents, tag, args, rec_chain, req_id);
      GLB2(gCurrSource) = saveSource;
      return res;
    };
  
  //  Small optimization; if no "<" character is found in the block we
  //  don't bother with looking for enclosed tags. This saves a lot
  //  of processing time!
  if (search(data, "<") == -1)
    return data;
  
  //  Need mapping for handling all known block definitions using
  //  the lambda function declared above. We also need to avoid the
  //  situation where the output of the block is reinterpreted by
  //  parse_html forever (or up to the current limit); we do this
  //  by a hack(TM): append an "_" prefix to all tag names in the
  //  source block, and also in the block names. When we're done
  //  we revert back to the original tag names. Whew!
  data = replace(data, "<", "<_");
  data = replace(data, "<_/", "</_");
  foreach (indices(GLB(gBlockDefs, req_id)), block_name)
    custom_containers += ([ "_" + block_name : callback_fn ]);
  data =  parse_html(data, ([]), custom_containers,
		       recursion_chain, req_id);
  return replace(data, ({ "<_", "</_" }), ({ "<", "</" }) );
}


string process_block(string data, string tag, mapping args,
		     multiset recursion_chain, object req_id)
{
  string   var_args_str, temp_block_def;
  string   param_value, param_name, content;
  mapping  dup_args, block_params;

  //  Convert tag to lower-case and get block definition text.
  GET_GLB(req_id);
  tag = lower_case(tag);
  temp_block_def = GLB2(gBlockDefs)[tag][1];
  
  //  Set current source container to the data found in this tag
  //  and process all repeated template insertion requests. This is
  //  performed as the first step since we need to find all instances
  //  of default parameter values before substitution takes place.
  GLB2(gCurrSource) = data;
  temp_block_def = parse_html(temp_block_def,
		       //  Send error message instead of "" for illegal tags
		       ([
			 "tmplinsert"      : tmplinsert_tag,
		         "tmplinsertall"   : tmplinsert_tag,
			 "tmplinsertexcept": tmplinsert_tag,
			 "tmplinsertblock" : tmplinsert_tag ]),
		       ([ "tmpldefault"    :
			    lambda(string tag, mapping args, string contents,
				   multiset rec_chain, object req_id)
			    {
			      string err = "Templates error: "
				           "Template defaults cannot be set "
				           "inside block.";
			      log_err(err, req_id);
			      return err;
			    },
			  "tmplblock"      :
			     lambda(string tag, mapping args, string contents,
				    multiset rec_chain, object req_id)
			     {
			       string err = "Templates error: "
				            "Nested block declarations not "
				            "allowed.";
			       log_err(err, req_id);
			       return err;
			     },
		          "tmpldefaultparam" :
		             tmpldefaultparam_container ]),

		      //  We pass tag as last parameter for tmpldefaultparam
		      //  only since it needs to know the current block
		      recursion_chain + (< tag >), req_id, tag);
  
  //  Loop over all parameters that the block defines and replace them
  //  with values from the custom tag. We keep a copy of the tag
  //  arguments mapping and shorten it as we go along. If no value is
  //  found we get the default value for that parameter. Note that
  //  args is sent as lower-case by Roxen.
  block_params = GLB2(gBlockDefs)[tag][0];
  dup_args = args;
  foreach (indices(block_params), param_name) {
    //  If this parameter name is present in dup_args, process it and
    //  remove it from the argument list.
    if (!zero_type(param_value = dup_args[param_name])) {
      dup_args -= ([ param_name : 0 ]);
    } else {
      //  Substitute default value for parameter
      param_value = block_params[param_name];
    }

    //  Perform parameter string replacement
    temp_block_def = replace_nocase(temp_block_def, "$" + param_name + "$",
				    param_value);
  }
  
  //  Look for $varargs$ (non-case-sensitive) within the template block
  //  and replace it with the remaining arguments in our duplicate list.
  var_args_str = "";
  foreach (indices(dup_args), param_name) {
    if (strlen(var_args_str) > 0)
      var_args_str += " ";
    var_args_str += param_name + "=\"" + dup_args[param_name] + "\"";
  }
  temp_block_def = replace_nocase(temp_block_def, "$varargs$", var_args_str);
      
  //  Look for $tag_name$ (non-case-sensitive) within the template
  //  block and insert the content data.
  GLB2(gCurrSource) = data;
  content = handle_content_insert(data, recursion_chain + (< tag >), req_id);
  temp_block_def = replace_nocase(temp_block_def, "$" + tag + "$", content);

  //  Return block
  return temp_block_def;
}


string handle_insert_except(string src, array avoid_tags,
			    multiset recursion_chain, object req_id)
{
  //  We need to parse the src string and remove all tags that are
  //  listed in the array avoid_tags. To accomplish this we call
  //  parse_html with a callback which clears the unwanted parts.
  //  After that we use the content insertion handler to process
  //  remaining tags in the source string.
  mapping  avoid_containers;
  string   a_tag;
  function callback_fn = 
    lambda(string tag, mapping args, string contents) { return ""; };
  
  avoid_containers = ([]);
  foreach (avoid_tags, a_tag)
    avoid_containers += ([ lower_case(a_tag) : callback_fn ]);
  return handle_content_insert(parse_html(src, ([]), avoid_containers),
			       recursion_chain, req_id);
}


string tmplinsert_tag(string tag, mapping args, multiset recursion_chain,
		      object req_id)
{
  array    tag_names;
  array    ins_data;
  string   a_name, result = "";
  int      pos;
  mapping  ins_data_item;

  //  A special tag is "tmplinsertall" which may only be used when the
  //  recursion chain is empty.
  GET_GLB(req_id);
  if (tag == "tmplinsertall") {
    if (!sizeof(recursion_chain)) {
      //  Insert all content data at once
      return handle_content_insert(GLB2(gEntireSource),
				   recursion_chain + (< "__dummy__" >),
				   req_id);
    } else {
      string   err = "Templates error: Invalid use of \""
	             "tmplinsertall\". It may only appear outside of "
	             "template blocks.";
      log_err(err);
      return err;
    }
  }
  
  //  Get name from args. There may be a comma-separated list which we need
  //  to split first.
  if (!args->tag)
    return "";
  tag_names = ((args->tag - " ") / ",") - ({ "" });

  //  We then check whether this is a request to insert a regular tag or
  //  a template block tag. For non-block insertions we only allow a
  //  single tag name.
  if (tag == "tmplinsert") {
    if ((sizeof(tag_names) != 1)
	|| !zero_type(GLB2(gBlockDefs)[tag_names[0]])) {
      string  err = "Templates error: Names of template blocks "
	            "or multiple tags may not be used in \"tmplinsert\".";
      log_err(err, req_id);
      return err;
    }
    
    //  Conventional insertion of first occurrence of tag name in data.
    //  Find data to insert in entire source container (i.e., disregard
    //  current source container). Since there is no block there can't
    //  be any macro subsitution.
    ins_data = get_custom_data(GLB2(gEntireSource), ({ tag_names[0] }) );
    if (sizeof(ins_data) >= 1) {
      //  Pick first occurrence and skip macro parameters and extract
      //  replacement text.
      result = ins_data[0]->data;

      //  We need to parse the insertion data for tags that we have block
      //  definitions available for.
      result = handle_content_insert(result, recursion_chain, req_id);
    } else
      //  Test if there is a default value we can use
      if (!zero_type(GLB2(gDefaults)[tag_names[0]]))
	result = GLB2(gDefaults)[tag_names[0]];
    
    //  Output resulting string
    return result;
  }
  
  //  This is a tag which is rather odd, but nevertheless good to have
  //  in certain situations. It looks in the current source container and
  //  inserts everything found there EXCEPT those blocks whose names we get
  //  as an input parameter.
  if (tag == "tmplinsertexcept") {
    return (handle_insert_except(GLB2(gCurrSource), tag_names,
				 recursion_chain, req_id));
  }
  
  //  At this place only "tmplinsertblock" tag is allowed. Also verify
  //  that no non-block names are included in the tag name list.
  if (tag != "tmplinsertblock") {
    string  err = "Templates error: Unknown tag \"" + tag + "\".";
    log_err(err);
    return err;
  }
  foreach (tag_names, a_name)
    if (zero_type(GLB2(gBlockDefs)[a_name])) {
      string  err = "Templates error: Only block tag names "
	            "allowed in \"tmplinsertblock\". Offending tag name \""
	            + a_name + "\".";
      log_err(err, req_id);
      return err;
    }

  //  Also check for infinite recursion. We get this if a block is in
  //  the recursion chain and we get another request for it.
  foreach (tag_names, a_name)
    if (recursion_chain[a_name]) {
      string  err = "Templates error: Infinite recursion detected "
	            "when inserting template block \"" + a_name + "\".";
      log_err(err, req_id);
      return err;
    }
  
  //  Search through the current source container for all tags that
  //  are of interest.
  ins_data = get_custom_data(GLB2(gCurrSource), tag_names);

  //  For any tag whose contents wasn't found we check for default
  //  values.
  multiset  want_tag_set = (< >);
  foreach (tag_names, a_name)
    want_tag_set += (< a_name >);
  foreach (ins_data, ins_data_item)
    want_tag_set -= (< ins_data_item->tag >);
  foreach (tag_names, a_name)
    if (want_tag_set[a_name] && !zero_type(GLB2(gDefaults)[a_name])) {
      ins_data += ({ ([ "tag"  : a_name,
			"args" : ([]),
			"data" : GLB2(gDefaults)[a_name] ]) });
    }
  
  //  Loop over all found data elements
  if (sizeof(ins_data) >= 1) {
    //  Save current source block on stack so we can retrieve it
    //  later. Only needed once outside the loop since gCurrSource
    //  is set in each iteration.
    GLB2(gSourceStack)->push(GLB2(gCurrSource));
    
    //  For each found data item we get the matching block definition
    //  and perform the macro expansion.
    foreach (ins_data, ins_data_item) {
      result += process_block(ins_data_item->data, ins_data_item->tag,
			      ins_data_item->args, recursion_chain, req_id);
    }
    
    //  Restore gCurrSource from stack
    GLB2(gCurrSource) = GLB2(gSourceStack)->pop();
  }

  //  Output resulting string
  return result;
}



string tmpldefaultparam_container(string tag, mapping args, string contents,
				  multiset rec_chain, object req_id,
				  string current_block)
{
  //  Only allow setting of default parameter values within a template
  //  block definition.
  if (sizeof(rec_chain) < 1) {
    string err = "Templates error: Default parameter values can "
                 "only be specified within a block definition.";
    log_err(err, req_id);
    return err;
  }
  
  //  Check that the parameter name is included and that we got the
  //  correct block name from the caller.
  GET_GLB(req_id);
  if (args->param && !zero_type(GLB2(gBlockDefs)[current_block])) {
    //  Verify that the parameter name is valid for this block
    if (!zero_type(GLB2(gBlockDefs)[current_block][0]
		                              [lower_case(args->param)])) {
      //  Update default value for this parameter in this block
      GLB2(gBlockDefs)[current_block][0][lower_case(args->param)] = contents;
    } else {
      string err = "Templates error: Setting of default value "
	           "for unknown parameter \"" + args->param + "\".\n";
      log_err(err, req_id);
      return err;
    }
  }
  return "";
}


string tmpldefault_container(string tag, mapping args, string contents,
			     multiset rec_chain, object req_id)
{
  //  Add value to mapping of defaults. We strip leading and trailing
  //  white-space.
  if (!args->tag)
    return "";
  contents = strip_surrounding_ws(contents);
  GLB(gDefaults, req_id) += ([ lower_case(args->tag) : contents ]);
  return "";
}


string tmplblock_container(string tag, mapping args, string contents,
				  multiset rec_chain, object req_id)
{
  array    want_params;
  mapping  want_params_mapping = ([]);
  int      i;
  
  //  Strip leading and trailing whitespace from block definition
  contents = strip_surrounding_ws(contents);
  
  //  Add this definition to the mapping of block definitions. If
  //  there is an existing entry we will replace it with the new
  //  value, which is exactly what we want. We convert the parameter
  //  array to a mapping where each index is in lower-case for easy
  //  matching at insertion time.
  if (args->tag) {
    if (args->params) {
      //  Clean up comma-spearated string and create mapping. The value
      //  in the mapping is considered the default value for the parameter
      //  in case a real value is missing from the custom tag.
      want_params = ((lower_case(args->params) - " ") / ",") - ({ "" });
      foreach (indices(want_params), i)
	want_params_mapping += ([ want_params[i] : "" ]);
    }
    GLB(gBlockDefs, req_id) += ([ lower_case(args->tag) :
				  ({ want_params_mapping, contents }) ]);
  }
  return "";
}


string tmplhelp_container(string tag, mapping args, string contents,
			     multiset rec_chain, object req_id)
{
  //  In normal processing mode all help containers are simply removed
  return "";
}


string tmploptions_tag(string tag, mapping args, multiset recursion_chain,
		      object req_id)
{
  //  When <tmploptions> is seen by parse_rxml it's already too late to
  //  perform any options handling. Therefore it's handled separately
  //  before, and now we simply ignore it.
  return "";
}


string convert_tags_to_containers(string tmpl_data, string source,
				  object req_id)
{
  string  a_tag;
  
  //  Find all block definitions and remember whether they define tags
  //  or containers.
  mapping result = ([ "taglist" : ([]), "contlist" : ([]), "err" : "" ]);

  parse_html(tmpl_data, ([]), ([ "tmplblock" :
                 lambda(string tag, mapping args, string contents, mapping res)
                 {
		   if (!args->tag)
		     return 0;
		   if (args->singletag) {
		     //  Add to list of single tags
		     if (!strlen(res->err) && (res->contlist[args->tag]))
		       res->err = args->tag;
		     res->taglist += ([ lower_case(args->tag) : 1]);
		   } else {
		     //  Add to list of container tags
		     if (!strlen(res->err) && res->taglist[args->tag])
		       res->err = args->tag;
		     res->contlist += ([ lower_case(args->tag) : 1 ]);
		   }
		   return 0;
		 } ]), result);
  if (strlen(result->err)) {
    string  err = "Templates error: Inconsistent use of "
                  "singletag attribute in redefinition of tag/container \""
                  + result->err + "\".";
    log_err(err, req_id);
    return err;
  }
  
  //  If we get this far we can replace all tags with containers in
  //  the source data. To avoid never-ending parsing of the data we
  //  output we prefix each new tag name with "_".
  if (sizeof(result->taglist)) {
    foreach (indices(result->taglist), a_tag)
      result->taglist[a_tag] = lambda(string tag, mapping args)
                        {
			  string  arg_str = "", an_arg;
			
			  foreach (indices(args), an_arg)
			    arg_str += " " + an_arg + "=\""
                                       + args[an_arg] + "\"";
			  return "<_" + tag + arg_str + "></_" + tag + ">";
			};
    source = parse_html(source, result->taglist, ([]) );
    source = replace(source, ({ "<_", "</_" }), ({ "<", "</" }) );
  }
  
  return source;
}


string handle_tmpl_options(string tmpl_data, string source, object req_id)
{
  function callback_tag_fn =
            lambda(string tag, mapping args, array result)
	    {
	      //  Convert \n\n -> <p>
	      if ((args->option) &&
		  (lower_case(args->option) == "mungedoublelf"))
		result[0] += ([ "mungedoublelf" : "" ]);
	    };
  
  //  Find all tmploptions tags in the template
  array result = ({ ([]) });
  parse_html(tmpl_data, ([ "tmploptions" : callback_tag_fn ]), ([ ]), result);
  
  //  Handle all options found in the template
  if (result[0]->mungedoublelf) {
    //  Convert double linefeeds into <p>
    source = replace(source, ({ "\n\n" }), ({ "\n\n<p>" }) );
  }
  
  //  Return source back to caller in case we've updated it
  return source;
}


string do_processing(string tag, mapping args, string contents,
		     object req_id, mapping defines, string tmpl_path)
{
  string   result, tmpl_file_data,
           document_name, raw_name;

  //  Load template file
  tmpl_file_data = req_id->conf->try_get_file(tmpl_path, req_id, 0, 1);
  
  if (!tmpl_file_data) {
    string  err = "Templates error: Template file \"" + tmpl_path
                  + "\" could not be opened.";
    log_err(err, req_id);
    return err;
  }

#if COMMENT_STRIPPING
  //  Strip comments from template if necessary
  if (QUERY(tmpl_strip_comments))
    tmpl_file_data = strip_comments(tmpl_file_data);
#endif

  //  Before any processing takes place we read all tmploptions tags from
  //  the template file. We pass this function the content as well in case
  //  it wants to modify it.
  contents = handle_tmpl_options(tmpl_file_data, contents, req_id);
  
  //  To support single tags (i.e. non-container tags) we have a simple
  //  solution; look through the template file to find all such definitions
  //  and then modify the source data by adding closing tags immediately
  //  following the opening tag. By doing this we can treat it like a
  //  regular container from now on. We won't allow redefinitions of blocks
  //  from tags to containers (or vice versa) anywere.
  contents = convert_tags_to_containers(tmpl_file_data, contents, req_id);
  
  //  Parse the template file from top to bottom, handling each of our
  //  template tags on the fly. We initialize the globals to hold empty
  //  block definitions and default values, and we set the current source
  //  container to the entire <template></template> container.
  //
  //  We add "tmpldatafile" and "tmpldatafile_raw" values to the defaults
  //  holding the name of the document file with any optional template
  //  override value removed, and the raw document file respectively.
  //  These are useful e.g. when creating links between frames.
  document_name = req_id->not_query;
  raw_name = document_name;
  if (req_id->query)
    raw_name += "?" + req_id->query;
  GET_GLB(req_id);
  GLB2(gDefaults) = ([ "tmpldatafile"     : document_name,
		       "tmpldatafile_raw" : raw_name ]);
  GLB2(gBlockDefs) = ([]);
  GLB2(gSourceStack) = Stack.stack();
  GLB2(gEntireSource) = contents;
  GLB2(gCurrSource) = contents;
  result = parse_html(tmpl_file_data,
		      ([ "tmplinsert"      : tmplinsert_tag,
                         "tmplinsertall"   : tmplinsert_tag,
			 "tmplinsertblock" : tmplinsert_tag,
			 "tmplinsertexcept": tmplinsert_tag,   // *** HACK ***
		         "tmploptions"     : tmploptions_tag ]),
                      ([ "tmpldefault"     : tmpldefault_container,
			 "tmpldefaultparam": tmpldefaultparam_container,
                         "tmplblock"       : tmplblock_container,
		         "tmplhelp"        : tmplhelp_container ]),

		      //  The last parameter is only used by tmpldefaultparam
		      (< >), req_id, "");
  
  //  Clear global data that we don't need to save across calls
#if OLD_GLOBALS
  gBlockDefs = ([]);
  gDefaults = ([]);
  gSourceStack->reset();
  gCurrSource = "";
  gEntireSource = "";
#else
  GLB_CLEAR(req_id);
#endif

  return result;
}


string container_template(string tag, mapping args, string contents,object req_id)
{
  string    tmpl_path;
  string    result;
  int       cached = 0;

  mapping defines = req_id->misc->defines;

  //  Init globals
  GLB_INIT(req_id);
  
  //  Extract name of template file. If there is another template name
  //  passed as a query parameter in the URL it will override these
  //  settings. Strip any ".." characters to prevent access to
  //  templates outside of the template directory tree.
  tmpl_path = (args->base||"")+QUERY(tmpl_storage_path);
  m_delete(args, "base");

  if (req_id->variables->tmpl)
    tmpl_path += (req_id->variables->tmpl - "..");
  else {
    if (sizeof(args) < 1)
      tmpl_path += QUERY(tmpl_default_name);
    else {
      //  Try to use "name" argument if possible (new behavior) before
      //  reverting to grabbing first argument (old behavior)
      if (args->name)
	if (args->name[..0] == "/")
	  tmpl_path = args->name;
	else
	  tmpl_path += args->name;
      else
	tmpl_path += indices(args)[0];
    }
  }

#if !DISABLE_CACHE
  string  hash_str_md5 = "";
  //  Hash template tag contents + template name + template change
  //  date to see if we have the result in the cache. We append the
  //  name of the virtual server to the cache key so we get a unique
  //  table.
  array   change_date = req_id->conf->stat_file(tmpl_path, req_id);
  string  hash_str = "";
  object  md5 = Crypto.md5();
  if (change_date && md5) {
	hash_str = contents + tmpl_path + change_date[3];
	md5->update(hash_str);
	hash_str_md5 = md5->digest();
  }
  if (!req_id->pragma[ "no-cache" ]) {
    if (strlen(hash_str_md5))
    {
      if ((result = cache_lookup("weblayout:" + req_id->conf->name,
				 hash_str_md5)) && result)
	cached = 1;
    }
  }
#endif
  
  //  If not cached we perform the processing as usual
  if (!cached) {
    //  Process template and content data. This is moved to a separate
    //  function to avoid dependencies on caching when running in batch
    //  mode.
    result = do_processing(tag, args, contents, req_id, defines, tmpl_path);
  
#if !DISABLE_CACHE
    //  Put result in cache
    if (strlen(hash_str_md5))
      cache_set("weblayout:" + req_id->conf->name, hash_str_md5, result);
#endif
  }
  
  return result;
}


#if GAUGE_PERF
//  Wrapper function for container_template that times the processing
//  time and writes the result to the log file.
string container_template_gauge(string tag, mapping args, string contents,
				object req_id, mapping defines)
{
  string res;
  werror(sprintf("Templates: gauge = %d msec. [Content file "
		 "URL \"%s\"]\n",
		 gauge { res = container_template(tag, args, contents,
						  req_id, defines); },
		 req_id->not_query));
  return res;
}
#endif



//  ===================================================================
//
//  Tag and container callback registration


multiset query_provides()
{
  return (< "weblayout_templates" >);
}

mapping query_tag_callers()
{
  return ([]);
}


mapping query_container_callers()
{
  return ([ "template" :
#if GAUGE_PERF
	    container_template_gauge
#else
	    container_template
#endif
	 ]);
}

