// Root class for a SiteBuilder site.

// $Id: Site.pike,v 1.79 1999/10/07 00:02:26 leif Exp $

// Given a path to a SiteBuilder storage directory, this class
// handles initialization and work areas with AC.

#if constant (roxen)
#include <config.h>
#include <module.h>

inherit "roxenlib";
#endif

import Sitebuilder;
import VC.CVS;


// Global data.

#define CVS_VERSION "1.10-idonex"
// Check that cvs --version contains this string.

constant handlers_dir = combine_path(__FILE__,"../../../handlers/");

string storage;			// The storage directory. Absolute and ends with '/'.
object/*(ModuleAC)*/ mac;

mapping(string:mixed) settings = ([]);
// Site global variables read from and written to the sbconfig file.

int admin_userid;
// The identity ID of the admin user (always exists).

mapping(string:mapping) content_types;
mapping languages;
mapping editors;
mapping editor_vars;
mapping local_editors;
string default_workarea;
mapping name_to_type, extension_to_type;

private static mapping(string:object(Workarea)) was;
// Work areas indexed by tag. When this is a mapping, we're up and running.

#ifdef THREADS
object(Thread.Mutex) was_mutex = Thread.Mutex();
#define WAS_LOCK object lock; catch (lock = was_mutex->lock())
#else
#define WAS_LOCK
#endif
// Lock for changing was.


function commit_hook;
//
// The commit_hook function is intended for additional commit
// logging, and similar purposes. It gets the following arguments:
//
//      void commit_hook(string type,
//                       string path,
//                       object id,
//                       string commit_log_message,
//                       object vcfile)
//
// The `type' argument is a string which indicates the kind of action
// that resulted in this committing of the file. 0 means no real commit
// was done, "commit" means the operation was a plain commit, and other
// string values indicate other actions resulting in a commit.


// Debugging tools.

#ifdef SBDEBUG
#define SBDEBUGMSG(msg) sbdebug (msg)
#else
#define SBDEBUGMSG(msg)
#endif

void sbdebug (string msg, mixed ... args)
{
#ifdef SBDEBUG
  if (sizeof (args)) msg = sprintf (msg, @args);
  werror ("SB: " + msg);
#endif
}


// AC.

private static object ac_module;
private static object/*(ModuleAC)*/ ac_mac;

object get_ac_module()
{
  if (!ac_module) {
#if constant (roxen)
    mapping acvar = roxen->query_var ("AC");
    if (!acvar || !acvar->loaders ||
	!(ac_module = acvar->loaders[mac->config()]))
      report_error ("SB: No AC module enabled in this configuration\n");
#endif
  }
  return ac_module;
}

#define AC_MAC (ac_mac || (ac_mac = get_ac_module() && ac_module->mac))

// All protection points.

#define PP_CONTENT_EDITOR "content editor"
// The content editor UI.

#define PP_WORKAREA_ADMIN "workarea admin"
// Work area administration. Does not use chained permission.

#define PP_PURGE "purge"
// Write access for nonreversable deletion of files/dirs. Does not use
// chained permission; write access on PP_FILE(path) must be checked
// too.

string pp_workarea (string tag) {return "workarea:" + tag;}
// Work area access for each work area. Chaining.

#define PP_FILE(path) ("file:" + (path))
// File or dir with leading but not trailing '/'. Uses chained
// permission.

#define PP_RXML(pp) ("rxml:" + (pp))
// Protection points for SiteBuilder tags.

#define PP_PLUGIN(pp) ("plugin:" + (pp))
// Protection points for SiteBuilder plugins.


private static int ac_init()
// Returns the admin user ID. :P
{
  int first_time = !mac->pp_get_id (PP_CONTENT_EDITOR);
  mac->pp_create (PP_CONTENT_EDITOR, "Content Editor", 0, "SiteBuilder");
  if (!mac->pp_get_id (PP_CONTENT_EDITOR)) throw (({"Couldn't setup the AC system\n"}));

  if (!admin_userid || !mac->id_get_id (admin_userid)) {
    // FIXME: If the admin user was removed, it should ideally be
    // recreated with the same ID again (which isn't even known at
    // this point).
    admin_userid = mac->id_create (
      "Administrator", mac->ID_TYPE_USER,
#if constant (roxen)
      roxen->query ("ConfigurationUser"), 0, roxen->query ("ConfigurationPassword")
#endif
    );
    if (!admin_userid)
      throw (({"Couldn't create an administrator user in Access Control\n",
	       backtrace()}));
    report_notice (
      "SB: An administrator user was added to the Access Control database. "
#if constant (roxen)
      "It's set to the same user and password as in the configuration interface. "
#endif
      "This user is used internally by SiteBuilder and should not be removed.\n"
    );
  }

  if (first_time) {
    mapping perms = ([PP_CONTENT_EDITOR: PERM_WRITE]);
    if (AC_MAC) {
      perms[({ac_module->PP_ROOT, ac_mac})] = PERM_WRITE;
      ac_mac->pc_set_realm_class (ac_module->PP_ROOT,
				  mac->pc_get_realm_class (PP_CONTENT_EDITOR));
    }
    mac->id_create ("Administrators", mac->ID_TYPE_GROUP, 0, perms, 0, (<admin_userid>));
  }

  mac->pp_clone (PP_CONTENT_EDITOR, PP_WORKAREA_ADMIN, "Work area administration");
  //mac->pp_clone (PP_CONTENT_EDITOR, PP_PURGE, "Purge files/directories");
  mac->pp_clone (PP_CONTENT_EDITOR, PP_FILE ("/"), "Dir: /");

  return admin_userid;
}

int ce_access_try (object id)
{
  id->misc->ce_access_checked = 1;
  // This flag tells whether Workarea.access_try should chain or not;
  // it only chains from this ppoint if this function has been run
  // earlier on the request id. This is getting way too icky.. :(

  return mac->access_try (PP_CONTENT_EDITOR, id);
}
mixed ce_access_throw (object id)
  {return mac->access_throw (PP_CONTENT_EDITOR, id);}

int ac_access_try (object id)
  {return AC_MAC && ac_mac->access_try (ac_module->PP_ROOT, id);}
mixed ac_access_throw (object id)
  {return AC_MAC && ac_mac->access_throw (ac_module->PP_ROOT, id);}

int ppoint_edit_access (object id)
  {return AC_MAC && ac_mac->access_perm (ac_module->PP_ROOT, id) == PERM_WRITE;}
int ppoint_edit_ppid()
{
  // Lying here; it gets too cumbersome to meddle with the different
  // macs all the way..
  return mac->pp_get_id (PP_CONTENT_EDITOR);
}

int wa_admin_access (object id)
  {return mac->access_perm (PP_WORKAREA_ADMIN, id) == PERM_WRITE;}
mixed wa_admin_access_throw (object id)
  {return mac->access_throw (PP_WORKAREA_ADMIN, id);}

int userprefs_edit_access (object id)
  {return mac->access_perm (PP_CONTENT_EDITOR, id) == PERM_WRITE;}

// Purge ppoint disabled until it's implemented.
int purge_access (object id)
  {return 0 /*mac->access_perm (PP_PURGE, id) == PERM_WRITE*/;}
int purge_ppid()
  {return mac->pp_get_id (PP_CONTENT_EDITOR /*PP_PURGE*/);}

// File ppoint handling.

// These functions assume absolute paths without leading or trailing '/'.

private static mapping(string:int) file_pps = ([]);
// Cache that maps paths to file ppoint IDs (for all accessed files and dirs).

int ac_find_file_pp (string|array(string) file)
{
  array(string) path;
  if (arrayp (file)) path = file, file *= "/";
  int ppid = file_pps[file];
  if (ppid) return ppid;

  if (!path) path = file / "/";
  for (int i = sizeof (path) - 1; i >= -1; i--) {
    string pstr = path[..i] * "/";
    ppid = file_pps[pstr] || mac->pp_get_id (PP_FILE ("/" + pstr));
    if (ppid) return file_pps[file] = file_pps[pstr] = ppid;
  }
  return 0;
}

array(string) find_dirhier_pp_paths (string dir)
// Returns the paths of all ppoints at or below dir in the hierarchy.
{
  dir = PP_FILE (sizeof (dir) ? "/" + dir : "");
  array(string) handles = mac->pp_get_handles();
  if (!handles) return ({});
  array(string) paths = Array.filter (
    handles,
    lambda (string path, string dir) {
      return path[..sizeof (dir) - 1] == dir &&
	(sizeof (path) == sizeof (dir) || path[sizeof (dir)] == '/');
    }, dir);
  for (int i = 0; i < sizeof (paths); i++) paths[i] = paths[i][6..];
  return paths;
}

int ac_check_file_pp (string file)
// If there is a ppoint (exactly) on file its ID is returned, 0
// otherwise.
{
  return mac->pp_get_id (PP_FILE ("/" + file));
}

int ac_add_file_pp (string file)
// Returns 0 on error.
{
  SBDEBUGMSG ("Trying to add file ppoint " + PP_FILE ("/" + file) + "\n");
  int ppid = mac->pp_clone (ac_find_file_pp (file), PP_FILE ("/" + file),
			    "File: /" + file);
  mac->pp_set_deletable (ppid, 1);
  if (ppid) file_pps = ([]);
  return ppid;
}

int ac_add_dir_pp (string dir)
// Returns 0 on error.
{
  SBDEBUGMSG ("Trying to add dir ppoint " + PP_FILE ("/" + dir) + "\n");
  int ppid = mac->pp_clone (ac_find_file_pp (dir), PP_FILE ("/" + dir),
			    "Dir: /" + dir);
  mac->pp_set_deletable (ppid, 1);
  if (ppid) file_pps = ([]);
  return ppid;
}

int ac_remove_file_pp (string file)
// Returns 0 on error.
{
  SBDEBUGMSG ("Trying to remove ppoint " + PP_FILE ("/" + file) + "\n");
  int res = sizeof (file) && mac->pp_remove (PP_FILE ("/" + file));
  if (res) file_pps = ([]);
  return res;
}

int ac_copy_dir_pp (string from, string to)
// Copies all ppoints under from. Returns 0 on error.
{
  from = PP_FILE (sizeof (from) ? "/" + from : "");
  array(string) handles = mac->pp_get_handles();
  if (!handles) return 0;
  array(string) paths = Array.filter (
    handles,
    lambda (string path, string from) {
      return path[..sizeof (from) - 1] == from &&
	(sizeof (path) == sizeof (from) || path[sizeof (from)] == '/');
    }, from);

  int ok = 1;
  foreach (paths, string path) {
    string newpath = to + path[sizeof (from)..];
    string name = mac->pp_get_name (path);
    if (name == "File: " + path[5..]) name = "File: /" + newpath;
    else if (name == "Dir: " + path[5..]) name = "Dir: /" + newpath;
    else if (!name) ok = 0;
    SBDEBUGMSG ("Trying to copy ppoint " + path + " to " +
		PP_FILE ("/" + newpath) + "\n");
    ok = mac->pp_clone (path, PP_FILE ("/" + newpath), name) && ok;
    if (ok) file_pps = ([]);
  }

  return ok;
}

int fileperm (string|array(string) file, object id, void|int ppid)
// Note: The perm() method in VCFile/VCDir caches the permission; prefer that one.
{
  if (!ppid) ppid = ac_find_file_pp (file);
  if (!ppid) throw (({"No file/dir protection point found for /" +
		      (stringp (file) ? file : file * "/") + "\n", backtrace()}));
  return mac->access_chain_perm (ppid, id);
}


// ppoints for SiteBuilder tags

array (string) rxml_pp_list( object id )
{
  array (string) handles = mac->pp_get_handles();
  array (string) res = ({ });

  foreach (glob( "rxml:*", handles ), string handle)
    res += ({ handle[5..] });
  return sort( res );
}

int rxml_pp_exists( string handle, object id )
{
  return mac->pp_get_id( PP_RXML( handle ) );
}

int rxml_pp_create( string handle, object id )
{
  if (!wa_admin_access( id ))
    return 0;
  return mac->pp_create( PP_RXML( handle ), "RXML: " + handle, 0,
			 "SiteBuilder" );
}

int rxml_pp_remove( string handle, object id )
{
  if (!wa_admin_access( id ))
    return 0;
  return mac->pp_remove( PP_RXML( handle ) );
}

int rxml_pp_change( string old, string new, object id )
{
  if (!wa_admin_access( id ))
    return 0;
  return mac->pp_set_handle( PP_RXML( old ), PP_RXML( new ) )
    && mac->pp_set_name( PP_RXML( new ), "RXML " + new );
}

int rxml_pp_try( string handle, object id )
{
  return mac->access_try( PP_RXML( handle ), id );
}


// ppoints for plugins:

int plugin_pp_exists( string handle, object id )
{
  return mac->pp_get_id( handle );
}

int plugin_pp_create( string handle, object id )
{
  return mac->pp_clone(PP_CONTENT_EDITOR, PP_PLUGIN( handle ), "Plugin: " + handle);
}

int plugin_pp_remove( string handle, object id )
{
  return mac->pp_remove( PP_PLUGIN( handle ) );
}

int plugin_pp_try( string handle, object id )
{
  return mac->access_perm( PP_PLUGIN( handle ), id );
}

int plugin_get_num_pp( object id )
{
  array (string) handles = mac->pp_get_handles();
  array (string) res = ({ });
  int num=0;
  foreach (glob( "plugin:*", handles ), string handle)
    num++;
  return num;
}


// Loading and saving settings.

void sbconfig_write (mapping(string:mixed) settings)
{
  string sbconfig = storage + "sbconfig";
  object(Stdio.File) fd = Stdio.File();

  if (!fd->open (sbconfig + ".w", "wct"))
    throw (({"Error opening " + sbconfig + ".w for write: " +
	     strerror (fd->errno()) + "\n", backtrace()}));

  fd->write ("/* Site configuration for Roxen SiteBuilder. -*-pike-*- *\n"
	     " * This file is recreated when the site config changes. *\n"
	     " * It's read when the SiteBuilder module is (re)loaded. *\n"
	     " * You cannot add or remove work areas by editing here. */\n"
	     "\n");
  foreach (sort (indices (settings)), mixed var)
    fd->write (pike_encode (var) + ": " + pike_encode (settings[var]) + ",\n");
  fd->close();

  if (fd->errno()) {
    rm (sbconfig + ".w");
    throw (({"Error writing " + sbconfig + ".w: " +
	     strerror (fd->errno()) + "\n", backtrace()}));
  }

  mv (sbconfig, sbconfig + "~");
  if (!mv (sbconfig + ".w", sbconfig)) {
    rm (sbconfig + ".w");
    mv (sbconfig + "~", sbconfig);
    throw (({"Error renaming " + sbconfig + ".w: " +
	     strerror (errno()) + "\n", backtrace()}));
  }
  rm (sbconfig + "~");
}

mapping sbconfig_read()
{
  string sbconfig = storage + "sbconfig";
  object(Stdio.File) fd = Stdio.File();

  if (!fd->open (sbconfig, "r"))
    throw (({"Error opening " + sbconfig + " for read: " +
	     strerror (fd->errno()) + "\n", backtrace()}));

  string contents = fd->read();
  fd->close();

  if (fd->errno()) throw (({"Error reading " + sbconfig + ": " +
			    strerror (fd->errno()) + "\n", backtrace()}));

  mapping(string:mixed) settings;
  program prg;			// Having this inside the catch bonks 0.5..
  array err = catch {
    // This is more Turing complete than we'd like, but it's
    // reasonable to assume that the config file isn't hostile.
    master()->set_inhibit_compile_errors ("");
    prg = compile_string ("constant settings = ([" + contents + "]);");
    master()->set_inhibit_compile_errors (0);
    settings = prg()->settings;
  };

  if (err) throw (({"Syntax error in " + sbconfig + ": " + err[0] + "\n", err[1]}));
  return settings;
}

void type_check (mixed var, function(mixed:int) proper_typep, string name, string type)
{
  if (var && !proper_typep (var))
    throw (({sprintf ("Error in sbconfig file: %s is not a %s: %O\n",
		      name, type, var), backtrace()}));
}

#if !constant (cpp)
#define TYPE_CHECK(var, name, typep) \
  type_check (var, typep, name, "typep"[..sizeof ("typep") - 2])
#else
#define TYPE_CHECK(var, name, typep) \
  type_check (var, typep, name, #typep[..sizeof (#typep) - 2])
#endif

void load (void|int parseonly)
{
  WAS_LOCK;

  if (!parseonly) settings = sbconfig_read();
  foreach (indices (settings), string var)
    TYPE_CHECK (var, "variable name", stringp);

  TYPE_CHECK (settings->admin_userid, "'admin_userid'", intp);
  admin_userid = settings->admin_userid;

  TYPE_CHECK (settings->workareas, "'workareas'", mappingp);
  foreach (indices (settings->workareas || ([])), string tag) {
    TYPE_CHECK (tag, "'workareas' index", stringp);
    TYPE_CHECK (settings->workareas[tag], "'workareas' value", mappingp);
    if (was[tag]) was[tag]->set_settings (settings->workareas[tag]);
    else {
      was[tag] = Workarea (this_object(), tag, settings->workareas[tag], !sizeof (was));
      // Fallback if the AC database is messed up.
      mac->pp_clone (PP_CONTENT_EDITOR, pp_workarea (tag), "$wa_name$");
    }
  }
  foreach (indices (was), string tag)
    if (!settings->workareas[tag]) destruct (was[tag]), m_delete (was, tag);

  TYPE_CHECK (settings->content_types, "'content_types'", mappingp);
  foreach (indices (settings->content_types || ([])), string type) {
    TYPE_CHECK (type, "'content_types' index", stringp);
    TYPE_CHECK (settings->content_types[type], "'content_types' value", mappingp);
    foreach (indices (settings->content_types[type] || ([])), string var) {
      TYPE_CHECK (var, "'content_types[" + type + "]' index", stringp);
      // Don't bother to check value..
    }
  }
  content_types = settings->content_types || default_content_types;

  TYPE_CHECK (settings->languages, "'languages'", mappingp);
  foreach (indices (settings->languages || ([])), string lang) {
    TYPE_CHECK (lang, "'languages' index", stringp);
    TYPE_CHECK (settings->languages[lang], "'languages' value", mappingp);
    foreach (indices (settings->languages[lang] || ([])), string var) {
      TYPE_CHECK (var, "'languages[" + lang + "]' index", stringp);
      // Don't bother to check value..
    }
  }
  languages = settings->languages || default_languages;

  // FIX ME some TYPE_CHECK code necessary? /Wing
  editors = settings->editors
    || ([ "Local"
	  : ([ "valid" : "both",
	       "defs" : ({ ([ "valid" : "named",
			      "valid_content_types" : (< "text/*" >),
			      "content_type" : "application/x-roxen-launcher",
			      "contents" : "nfs=#nfs#\n"
		                           "smb=#smb#\n"
		                           "appleshare=#appleshare#\n"
		                           "content-type=#content-type#\n"
	       ]) }) ]) ]);
  editor_vars = ([ "anonymous" : ([ ]), "named" : ([ ]) ])
    + (settings->editor_vars || ([ ]));
  editor_vars->anonymous->dir = editor_vars->anonymous->dir || "";
  editor_vars->anonymous->smb = editor_vars->anonymous->smb || "";
  editor_vars->anonymous->nfs = editor_vars->anonymous->nfs || "";
  editor_vars->anonymous->appleshare
    = editor_vars->anonymous->appleshare || "";
  editor_vars->named->dir = editor_vars->named->dir || "";
  editor_vars->named->smb = editor_vars->named->smb || "";
  editor_vars->named->nfs = editor_vars->named->nfs || "";
  editor_vars->named->appleshare = editor_vars->named->appleshare || "";

  local_editors = settings->local_editors
    || ([ "Default" : ({ "text/* .txt", "application/vnd.ms-excel .xls",
			 "application/msword .doc" }) ]);

  TYPE_CHECK (settings->default_workarea, "'default_workarea'", stringp);
  default_workarea = settings->default_workarea || "";
}

void save()
{
  settings->admin_userid = admin_userid;
  settings->workareas = ([]);
  foreach (indices (was), string tag)
    settings->workareas[tag] = was[tag]->get_settings();
  settings->content_types = content_types;
  settings->languages = languages;
  settings->editors = editors;
  settings->editor_vars = editor_vars;
  settings->local_editors = local_editors;
  settings->default_workarea = default_workarea;
  
  sbconfig_write (settings);
}

mixed sbconfig_get(string what)
{
  return settings[what];
}


// Handlers

mapping handlers;

void init_handlers()
{
  handlers = ([ ]);

  foreach(get_dir( handlers_dir ) - ({ "CVS" }), string dir)
  {
    if (dir[0..0] == ".")
      continue;
    object h=0;
    catch {
      h=compile_file(handlers_dir+dir+"/handler.pike")(this_object());
    };
    if(h)
    {
      handlers[dir]=h;
#ifdef SBDEBUG
      werror(" Loaded handler "+dir+" successfully.\n");
#endif
    }
  }
}

constant default_content_types = ([
  "text/html" : ([ "name" : "HTML",
		   "handler" : "html",
		   "downloadp" : 1,
		   "parsep" : 1,
		   "extensions" : (< "html", "htm" >),
		   "img" : "html.gif" ]),
  "text/plain" : ([ "name" : "Raw text",
		    "handler" : "text",
		    "downloadp" : 1,
		    "extensions" : (< "txt" >),
		    "img" : "text.gif" ]),
  "text/rtf" : ([ "name" : "Rich Text Format text",
		  "handler" : "text",
		  "downloadp" : 1,
		  "extensions" : (< "rtf" >),
		  "img" : "text.gif" ]),
  "sitebuilder/menu" : ([ "name" : "Menu",
			  "handler" : "menu",
			  "downloadp" : 0,
			  "extensions" : (< "menu" >),
			  "internalp" : 1,
			  "img" : "menu.gif" ]),
  "sitebuilder/pmenu" : ([ "name" : "Parsed menu",
			   "handler" : "pmenu",
			   "downloadp" : 1,
			   "parsep" : 0, // Parsed before menu parsing, not after.
			   "extensions" : (< "pmenu" >),
			   "internalp" : 1,
			   "img" : "menu.gif" ]),
  "sitebuilder/template" : ([ "name" : "Template",
			      "handler" : "template",
			      "downloadp" : 1,
			      "extensions" : (< "tmpl" >),
			      "internalp" : 1,
			      "img" : "tmpl.gif" ]),
  "image/gif" : ([ "name" : "GIF Image",
		   "handler" : "image",
		   "downloadp" : 1,
		   "extensions" : (< "gif" >),
		   "img" : "gif.gif" ]),
  "image/jpeg" : ([ "name" : "JPEG Image",
		    "handler" : "image",
		    "downloadp" : 1,
		    "extensions" : (< "jpg", "jpeg" >),
		    "img" : "jpg.gif" ]),
  "image/png" : ([ "name" : "PNG Image",
		   "handler" : "image",
		   "downloadp" : 1,
		   "extensions" : (< "png" >),
		   "img" : "png.gif" ]),
  "image/tiff" : ([ "name" : "TIFF Image",
		    "handler" : "image",
		    "downloadp" : 1,
		    "extensions" : (< "tiff", "tif" >),
		    "img" : "tiff.gif" ]),
  "video/mpeg" : ([ "name" : "MPEG movie",
		    "handler" : "default",
		    "downloadp" : 1,
		    "extensions" : (< "mpeg", "mpg" >),
		    "img" : "movie.gif" ]),
  "video/quicktime" : ([ "name" : "QuickTime movie",
			 "handler" : "default",
			 "downloadp" : 1,
			 "extensions" : (< "mov", "qt" >),
			 "img" : "movie.gif" ]),
  "model/vrml" : ([ "name" : "VRML model",
		    "handler" : "default",
		    "downloadp" : 1,
		    "extensions" : (< "vrml", "vrm" >),
		    "img" : "image.gif" ]),
  "application/vnd.ms-excel" : ([ "name" : "Microsoft Excel Spreadsheet",
				  "handler" : "default",
				  "downloadp" : 1,
				  "extensions" : (< "xls" >),
				  "img" : "xls.gif" ]),
  "application/msword" : ([ "name" : "Microsoft Word Document",
			    "handler" : "default",
			    "downloadp" : 1,
			    "extensions" : (< "doc" >),
			    "img" : "doc.gif" ]),
  "application/pdf" : ([ "name" : "PDF Document",
			 "handler" : "default",
			 "downloadp" : 1,
			 "extensions" : (< "pdf" >),
			 "img" : "pdf.gif" ]),
  "application/postscript" : ([ "name" : "PostScript Document",
				"handler" : "default",
				"downloadp" : 1,
				"extensions" : (< "ps", "eps" >),
				"img" : "ps.gif" ]),
  "application/mac-binhex40" : ([ "name" : "Macintosh BinHex archive",
				  "handler" : "default",
				  "downloadp" : 1,
				  "extensions" : (< "hqx" >),
				  "img" : "compressed.gif" ]),
  "application/zip" : ([ "name" : "ZIP archive",
			 "handler" : "default",
			 "downloadp" : 1,
			 "extensions" : (< "zip" >),
			 "img" : "compressed.gif" ]),
  "application/octet-stream" : ([ "name" : "Unknown",
				  "handler" : "default",
				  "downloadp" : 1,
				  "extensions" : (< >),
				  "img" : "binary.gif" ]),
  "roxen/pike" : ([ "name" : "Roxen handled/Pike",
		    "handler" : "text",
		    "downloadp" : 1,
		    "extensions" : (< "pike" >),
		    "img" : "pike.gif" ]),
  "roxen/cgi" : ([ "name" : "Roxen handled/CGI",
		   "handler" : "text",
		   "downloadp" : 1,
		   "extensions" : (< "cgi" >),
		   "img" : "cgi.gif" ]),
  "roxen/octet-stream" : ([ "name" : "Roxen handled/Other",
			    "handler" : "default",
			    "downloadp" : 1,
			    "extensions" : (< >),
			    "img" : "unknown.gif" ]),
]);

constant default_languages = ([
  "sv" : ([ "name" : "Swedish",
	    "encoding" : "ISO-8859-1" ]),
  "en" : ([ "name" : "English",
	    "encoding" : "ISO-8859-1" ]),
  "nl" : ([ "name": "Dutch",
	    "encoding": "ISO-8859-1" ]),
  "fr" : ([ "name": "French",
	    "encoding": "ISO-8859-1" ]),
  "de" : ([ "name": "German",
	    "encoding": "ISO-8859-1" ]),
  "it" : ([ "name": "Italian",
	    "encoding": "ISO-8859-1" ]),
  "pt" : ([ "name": "Portugese",
	    "encoding": "ISO-8859-1" ]),
  "es" : ([ "name": "Spanish",
	    "encoding": "ISO-8859-1" ]),
  "ru" : ([ "name": "Russian",
 	    "encoding": "ISO-8859-5" ]),
  "ar" : ([ "name": "Arabic",
	    "encoding": "ISO-8859-6" ]),
  "ja" : ([ "name": "Japanese",
	    "encoding": "ISO-2022-JP" ]),
  "zh" : ([ "name": "Chinese",
	    "encoding": "Big5" ]),
  "" :   ([ "name" : "None",
	    "encoding" : "ISO-8859-1" ]) ]);

mapping get_metadata(object vcfile, object id, void|int undelete_mode)
{
  mapping|object vcdata = vcfile &&
    vcfile->metadata(id, undelete_mode, !id);
  return objectp (vcdata) && vcdata->md;
}

string content_type_for(object vcfile, object id, void|int undelete_mode)
{
  mapping md = get_metadata(vcfile,id,undelete_mode);
  return (md && (md[ "http-content-type" ] || md["http_content_type"]))
    || "application/octet-stream";
}

object handler_for(string|object ct_or_vcf, void|object id);

static object low_handler_for(string|object ct_or_vcf, object id)
{
  if(stringp(ct_or_vcf))
    return handlers[
      content_types[ct_or_vcf] &&
      content_types[ct_or_vcf]->handler ];
  else if(objectp(ct_or_vcf))
    return handler_for(content_type_for(ct_or_vcf,id));
  else
    return 0;
}

object handler_for(string|object ct_or_vcf, void|object id)
{
  return low_handler_for(ct_or_vcf, id) || handlers["default"];
}

string charset_for(string|object lang_or_vcf, void|object id)
{
  if(stringp(lang_or_vcf)) {
    return (languages[lang_or_vcf] || ([ "encoding":"ISO-8859-1" ]))->encoding;
  } else if(objectp(lang_or_vcf))
    return charset_for(get_metadata(lang_or_vcf,id)[ "language" ]);
  else
    return 0;
}

mapping ct_data_for (string|object lang_or_vcf, void|object id)
// Looks up an entry in content_types. Returns an empty mapping on errors.
{
  if (objectp (lang_or_vcf))
    lang_or_vcf = lang_or_vcf->isfile && content_type_for (lang_or_vcf, id);
  if (stringp (lang_or_vcf) && content_types[lang_or_vcf])
    return content_types[lang_or_vcf];
  return ([]);
}

array(string) list_templates(object id)
{
  object vcdir = id->misc->wa->vcobj("/templates", id, 0);
  if(!vcdir)
    return ({});
  array|mapping list = vcdir->list(id);
  if(mappingp(list))
    return ({});
  array templates = Array.filter (list, "isfilep");
  return sort(templates->name(id));
}

array(string) list_languages()
{
  return sort( ({ "English", "Svenska" }) );
}

mapping(string:string) list_content_type_names()
{
  mapping names=([]);
  foreach(indices(content_types), string content_type)
    names[content_types[content_type]->name]=content_type;
  return names;
}

array(string) get_handlers()
{
  return sort(indices(handlers));
}

string find_content_type_from_filename(string filename)
{
  if(search(filename,".")==-1)
    return name_to_type["Unknown"]||"application/octet-stream";
  string ext=(filename/".")[-1];
  return extension_to_type[ext]||find_content_type_from_filename("");
}

void init_content_types()
{
  name_to_type = ([ ]);
  extension_to_type = ([ ]);
  foreach(indices(content_types ), string ct)
  {
    foreach(indices(content_types[ct]->extensions), string extension)
      extension_to_type[extension]=ct;
    name_to_type[ content_types[ ct ]->name ] = ct;
  }
}

mapping get_content_types()
{
  return copy_value(content_types);
}

mapping set_content_types(mapping in)
{
  content_types=copy_value(in);
  save();
  init_content_types();
  return in;
}

mapping get_languages()
{
  return copy_value(languages);
}

mapping set_languages(mapping in)
{
  languages=copy_value(in);
  save();
  return in;
}

mapping get_editors()
{
  return copy_value( editors );
}

array (string) list_editors()
{
  return sort( indices( editors ) );
}

mapping set_editors( mapping in )
{
  editors = copy_value( in );
  save();
  return in;
}

int editor_for( string editor, object|string vcfile, object id )
{
  string ct;
  mapping edef;
  
  if (!editors[ editor ])
    return 0;
  if (stringp( vcfile ))
    ct = vcfile;
  else
    ct = content_type_for( vcfile, id );
  foreach (editors[ editor ]->defs, mapping e)
    if (e->valid_content_types[ ct ])
      edef = e;
  if (!edef)
    foreach (editors[ editor ]->defs, mapping e)
      if (e->valid_content_types[ (ct / "/")[0] + "/*" ])
	edef = e;
  if (!edef)
    foreach (editors[ editor ]->defs, mapping e)
      if (e->valid_content_types[ "*/*" ])
	edef = e;
  if (edef)
  {
    if ((editors[ editor ]->valid == "named"
	 || editors[ editor ]->valid == "both")
	&& editor_vars->named->enabled)
    {
      array uids = ({ });
      object acdb, acid;
      
      acdb = get_ac_module()->online_acdb;
      acid = acdb->identities[ mac->id_get_id( id ) ];
      if (acid)
	uids = acid->active_auths()->auth_handle( mac->AUTH_FS_UID ) - ({0});
      // FIX ME Throw or log here?
      if (sizeof( uids ) > 1)
	return 0;
      if (sizeof( uids ))
	return 1;
    }
    if ((editors[ editor ]->valid == "anonymous"
	 || editors[ editor ]->valid == "both")
	&& editor_vars->anonymous->enabled)
      return 1;
  }
  return 0;
}

mapping get_editor_for( string editor, object vcfile, object id )
{
  string ct;
  mapping edef;
  
  if (!editors[ editor ])
    return 0;
  ct = content_type_for( vcfile, id );
  foreach (editors[ editor ]->defs, mapping e)
    if (e->valid_content_types[ ct ])
      edef = e;
  if (!edef)
    foreach (editors[ editor ]->defs, mapping e)
      if (e->valid_content_types[ (ct / "/")[0] + "/*" ])
	edef = e;
  if (!edef)
    foreach (editors[ editor ]->defs, mapping e)
      if (e->valid_content_types[ "*/*" ])
	edef = e;
  if (edef)
  {
    if ((editors[ editor ]->valid == "named"
	 || editors[ editor ]->valid == "both")
	&& editor_vars->named->enabled)
    {
      array uids = ({ });
      object acdb, acid;

      acdb = get_ac_module()->online_acdb;
      acid = acdb->identities[ mac->id_get_id( id ) ];
      if (acid)
	uids = acid->active_auths()->auth_handle( mac->AUTH_FS_UID ) - ({0});
      // FIX ME Throw or log here?
      if (sizeof( uids ) > 1)
	return 0;
      if (sizeof( uids ))
	if (editors[ editor ]->valid == "named"
	    || editors[ editor ]->valid == "both")
	{
	  string path, smb, nfs, appleshare, contents;
	  
	  path = vcfile->get_named_editarea( id, uids[0] );
	  smb = editor_vars->named->smb + replace( path, "/", "\\" );
	  nfs = editor_vars->named->nfs + path;
	  appleshare
	    = editor_vars->named->appleshare + replace( path, "/", ":" );
	  contents = do_output_tag( ([ ]), ({ ([ "smb" : smb,
						 "nfs" : nfs,
						 "appleshare" : appleshare,
						 "content-type" : ct, ]) }),
				    edef->contents, id );
	  return http_string_answer( contents, edef->content_type );
	}
    }
    if ((editors[ editor ]->valid == "anonymous"
	  || editors[ editor ]->valid == "both")
	&& editor_vars->anonymous->enabled)
    {
      string path, smb, nfs, appleshare, contents;
      
      path = vcfile->get_anonymous_editarea( id );
      smb = editor_vars->anonymous->smb + replace( path, "/", "\\" );
      nfs = editor_vars->anonymous->nfs + path;
      appleshare
	= editor_vars->anonymous->appleshare + replace( path, "/", ":" );
      contents = do_output_tag( ([ ]), ({ ([ "smb" : smb,
					     "nfs" : nfs,
					     "appleshare" : appleshare,
					     "content-type" : ct, ]) }),
				edef->contents, id );
      return http_string_answer( contents, edef->content_type );
    }
  }
  return 0;
}

mapping get_editor_vars()
{
  return copy_value( editor_vars );
}

mapping set_editor_vars( mapping in )
{
  editor_vars = copy_value( in );
  save();
  return in;
}

string get_default_workarea()
{
  return default_workarea;
}

void set_default_workarea( string dw )
{
  default_workarea = dw;
  save();
}

mapping get_local_editors()
{
  return copy_value( local_editors );
}

void set_local_editors( mapping in )
{
  local_editors = copy_value( in );
  save();
}

array (object) get_default_file( string ct, object id )
{
  string default_file;
  object vcobj;
  array list;
  array res;

  default_file = content_types[ ct ] && content_types[ ct ]->default_file;
  if (!default_file)
    return ({ });

  vcobj = id->misc->wa->vcobj( default_file, id, 0);
  if (!vcobj)
    return ({ });
  
  if (vcobj->isdir)
    list = vcobj->list( id );
  else if (vcobj->isfile)
    list = ({ vcobj });
  if (!arrayp( list ))
    return ({ });

  res = ({ });
  foreach (list, vcobj)
    if (vcobj->isfile && content_type_for( vcobj, id ) == ct
	&& vcobj->name( id ) != "index.html")
      res += ({ vcobj });
  return res;
}

// Work area handling.

array(object(Workarea)) wa_list (void|object id)
{
  return id ? Array.filter (values (was), "access_perm", id) : values (was);
}

object(Workarea) wa_lookup (string tag, void|object id)
{
  object(Workarea) wa = was[tag];
  return wa && (!id || wa->access_perm (id)) ? wa : 0;
}

// These return 0 when AC denies access. wa_admin_access_throw() gives
// the right AC response then. Call without id to disable AC checking.

object(Workarea) wa_new (mapping(string:mixed) settings, void|object(Workarea) from,
			 void|object id)
// If from isn't given, a work area for the main CVS branch is created.
{
  if (id && !wa_admin_access (id)) return 0;
  WAS_LOCK;

  string tag = "";
  if (from) {
    tag = sillycaps_mangle (settings->name);
    if (was[tag])
      for (int i = 2; ; i++) if (!was[tag + "_" + i]) {tag += "_" + i; break;}
  }
  else if (was[""]) throw (({"Work area for main branch already exists\n", backtrace()}));

  SBDEBUGMSG ("Adding work area tagged '" + tag + "'\n");

  if (sizeof (tag))
    rtag (storage + "cvsroot", tag, "site", sizeof(from->tag) && from->tag);
  object(Workarea) wa = Workarea (this_object(), tag, settings, !sizeof (was));

  mac->pp_clone (from ? pp_workarea (from->tag) : PP_CONTENT_EDITOR,
		 pp_workarea (tag), "$wa_name$");

  was[tag] = wa;
  save();
  return wa;
}

void wa_delete (object(Workarea) wa, void|object id)
// This will remove the CVS tag too, so it's really undoable.
{
  if (id && (!wa_admin_access (id) )) return;
  WAS_LOCK;
  if (!wa || !was[wa->tag]) return;

  SBDEBUGMSG ("Deleting work area tagged '" + wa->tag + "'\n");

  m_delete (was, wa->tag);
  mixed err = catch (save());
  if (!err) {
    mac->pp_remove (pp_workarea (wa->tag));
    rtag_delete (storage + "cvsroot", wa->tag, "site");
    rm_rf (storage + "/wa" + wa->tag);
    destruct (wa);
  }
  else {
    was[wa->tag] = wa;
    throw (err);
  }
}

void wa_change_settings (object(Workarea) wa, mapping(string:mixed) setting_changes,
			 void|object id)
// Settings not in setting_changes are kept.
{
  if (id && !wa_admin_access (id)) return;
  WAS_LOCK;
  if (!wa) return;

  wa->set_settings (wa->get_settings() | setting_changes);
  save();
}


// Initialization.

void destroy()
{
  if (was) foreach (values (was), object(Workarea) wa) destruct (wa);
  if (mac) destruct (mac);
#if constant (roxen)
  mapping sbs = roxen->query_var ("SiteBuilders");
  object sbconf = sbs && search (sbs, this_object());
  if (sbconf) m_delete (sbs, sbconf);
#endif
}

int is_initialized() {return !!was;}

#define INIT_ERROR(msg) \
  do { \
    was = 0; \
    string errmsg = msg; \
    report_error ("SB: " + errmsg); \
    return errmsg; \
  } while (0)

#define INIT_WARNING(msg) \
  do { \
    was = 0; \
    string errmsg = msg; \
    report_warning ("SB: " + errmsg); \
    return errmsg; \
  } while (0)

void|string initialize (function(string:object/*(ModuleAC)*/) get_mac,
			string storagedir, void|int create_storage)
// Initializes a site from a SiteBuilder storage directory. Returns 0
// on success, an error message otherwise (catches most errors).
{
  WAS_LOCK;			// To be on the safe side.
  if (was) throw (({"initialize() has already run successfully\n", backtrace()}));
  array err;

  // Check CVS binary.

  err = catch {
    object cvs = CVS (".", ".");
#ifndef __NT__    
    if (search (cvs->command ("--version") || "", CVS_VERSION) < 0)
      INIT_ERROR ("Couldn't find required CVS version: " + CVS_VERSION + "\n");
#endif    
  };
  if (err) {
#ifdef SBDEBUG
    werror (describe_backtrace (err));
#endif
    INIT_ERROR ("Failed to run cvs binary: " + err[0]);
  }

  // Storage directory creation and sanity checking.

  storage = combine_path (getcwd(), storagedir);
  if (storage[-1] != '/') storage += "/";
  report_notice ("SB: Initializing in storage " + storage + "\n");

  array(int) stat = file_stat (storage);
  if (!stat)
    if (create_storage) {
      report_notice ("SB: Initializing a new storage: " + storage + "\n");
      if (!mkdir (storage))
	INIT_ERROR ("Couldn't create storage directory " + storage + ": " +
		    strerror (errno()) + "\n");
      stat = file_stat (storage);
    }
    else INIT_WARNING ("Directory does not exist: " + storage + "\n");

  if (stat[1] != -2) INIT_WARNING ("Not a directory: " + storage + "\n");

  string sbconfig = storage + "sbconfig";
  int newconfig = !file_stat (sbconfig);
  if (newconfig)
    if (create_storage) {
      object fd = Stdio.File();
      if (!fd->open (sbconfig, "wc"))
	INIT_ERROR ("Couldn't initialize storage directory " + storage + ": " +
		    strerror (fd->errno()) + "\n");
      fd->close();
    }
    else INIT_WARNING ("Not a storage directory: " + storage + "\n");
  else {
    err = catch (settings = sbconfig_read());
    if (intp (settings->admin_userid))
      // Need this before the AC init.
      admin_userid = settings->admin_userid;
  }
  if (err) {
#ifdef SBDEBUG
    werror (describe_backtrace (err));
#endif
    INIT_ERROR ("Error reading config file " + sbconfig + ": " + err[0]);
  }

  // FIXME: Safeguard against two sitebuilders using the same storage.

  // Init AC.

  int new_admin_userid;
  if (!create_storage && !file_stat (storage + "acdb"))
    INIT_ERROR ("AC database not found: " + storage + "acdb\n");
  err = catch {
    mac = get_mac (storage + "acdb");
    new_admin_userid = ac_init();
  };
  if (err) {
#ifdef SBDEBUG
    werror (describe_backtrace (err));
#endif
    INIT_ERROR (err[0]);
  }

  // Init CVS.

  if (!file_stat (storage + "cvsroot")) {
    if (!create_storage)
      INIT_ERROR ("CVS repository not found: " + storage + "cvsroot\n");

    report_notice ("SB: Creating new CVS repository\n");
    err = catch {
      if (!init (storage + "cvsroot") ||
	  !new_repository (storage + "cvsroot", "site"))
	throw (({strerror (errno()), backtrace()}));
    };
    if (err) {
#ifdef SBDEBUG
      werror (describe_backtrace (err));
#endif
      INIT_ERROR ("Failed to create CVS repository: " + err[0]);
    }
    newconfig = 1;
  }

  // Initialize or read sbconfig.

  was = ([]);
  if (newconfig) {
    content_types = default_content_types;
    languages = default_languages;

    report_notice ("SB: Creating an initial work area\n");
    err = catch {wa_new ((["name": "Main"])) -> create_default_files();};
  }
  else err = catch (load (1));
  if (err) {
#ifdef SBDEBUG
    werror (describe_backtrace (err));
#endif
    INIT_ERROR (err[0]);
  }
  if (newconfig || admin_userid != new_admin_userid) {
    admin_userid = new_admin_userid;
    save();
  }

  // Initialize handlers

  init_handlers();

  // Initialize content types

  init_content_types();

#ifdef SBDEBUG
  report_notice ("SB: Init done\n");
#endif
}
