
// Handles a directory with all the state and AC stuff.

// $Id: VCDir.pike,v 1.58 1999/11/10 01:48:39 mast Exp $

// Created by vcobj() or vcdir() in Sitebuilder.Workarea. There should
// be one unique object for the same work area, file and user. Invalid
// objects returned by vcdir() are not unique, though.
//
// The file status is cached and only resampled when update_state() is
// called, which is done in vcobj(), vcdir(), and all the destructive
// operations. Note that it's not safe to make any assumptions about
// this cache being constant for any amount of time since these
// objects are shared between requests.
//
// All the functions in this class (besides create() and
// reset_caches()) should be safe to call at any time; they do all
// needed sanity checking and return appropriate errors. The interface
// is as similar as possible to VCFile.
//
// Common arguments for functions:
//
// object id		The almighty and ever present id object. It's
//			only used for permission checks and may be zero
//			if the superuser flag is given to the function.
// int undelete_mode	As it says.
// string comment	Comment for the cvs log. Since all cvs operations
//			are synchronous, this is required on a lot of
//			functions, as opposed to VCFile.
// int warning		If an operation that should give a warning is
//			about to be taken, return an error describing
//			the warning instead of do the operation.
// int superuser	Disables all permission checks when nonzero.
//
// The directory operation functions return a mapping(string:mixed)
// when errors occurs. The mapping contains at least the binding
// "type": "errortype", where errortype is a magic string which is
// either standard or documented for each function.
//
// Standard error types:
//
// "access denied":	Access blocked by AC. Additional bindings:
//			"ppoint": The ID of the file access ppoint in use.
//			"reason": If used, a message about the reason.
//			"vcobj": The VCDir object.
// "invalid directory":	As it says. This error is possible when the VCDir
//			object was created by vcdir() or when adding or
//			undeleting a directory outside the dir tree.
// "invalid operation":	The operation is not defined or not allowed in
//			the current state (which may well have changed
//			in the last ns). Additional bindings:
//			"operation": The name of the function.
//			"state": The current state on the form "X/Y".
//			"reason": A short explanation for relevant cases.
//			"vcobj": The VCDir object.
// "invalid name":	Returned by add when it tries to create a new
//			file, but the name doesn't pass
//			Sitebuilder.valid_name(). Bindings:
//			"name": The offending name.
// "warning":		A warning of the "are you sure you want to
//			continue"-type. Only returned when the warning
//			flag was set. The additional bindings are the
//			same as for "invalid operation".

#if constant (roxen)
#include <config.h>
#endif

import Sitebuilder;
import VC.CVS;

constant isdir = 1;
object isfilep() {return 0;}
object isdirp() {return this_object();}

object/*(Workarea)*/ wa;


// Debugging tools.

#ifdef SBDEBUG
#define OPMSG(_msg) sbdebug ("VCDir  " + userid + ":" + (wa && wa->tag) + \
			     ":/" + path * "/" + " " + _msg + "\n")
#else
#define OPMSG(_msg)
#endif


// Identity.

private static array(string) path; // A "clean" path, i.e. "" elements removed etc.
private static int userid;	// The AC identity, 0 if we're in a viewarea.

private static object(cachedir) updir; // The parent directory. Created only when needed.
private static object(cachedir) dir; // May be 0 when created by vcdir().
private static object(cachefile) infofile; // 0 iff dir is 0. Always fake.

private static object(cachedir) getupdir()
{
  if (updir) return updir;
  object(cachedir) topdir = wa->topdir (userid);
  return updir = topdir->find_dir (path[..sizeof (path) - 2]);
}

/*private static*/ object(cachefile) getinfofile (void|int fake)
// For internal use only.
{
  infofile = dir->find_file ("__info", fake);
  if (!infofile)
    throw (({"Directory /" + path * "/" + " has no __info\n", backtrace()}));
  else if (fake) infofile->isbinary (1);
  return infofile;
}

private static object(cachedir) getdir()
{
  if (dir) return dir;
  object(cachedir) topdir = wa->topdir (userid);
  dir = topdir->find_dir (path);
  if (dir) getinfofile();
  return dir;
}


// Error types.

#define ERR_ACCESS_DENIED(_ppoint, _msg) \
  (["type": "access denied", "ppoint": (_ppoint), \
    "reason": (_msg), "vcobj": this_object()])
constant err_invalid_directory = (["type": "invalid directory"]);
#define ERR_INVALID_DIRECTORY err_invalid_directory
#define ERR_LOW_INVALID_OP(_stat, _op, _msg) \
  (["type": "invalid operation", "operation": (_op), "state": (_stat), \
    "reason": (_msg), "vcobj": this_object()])
#define ERR_INVALID_OPERATION(_stat, _op) \
  ERR_LOW_INVALID_OP (_stat, _op, vcdir_invalid_op_msgs[_stat][_op])
#define ERR_INVALID_NAME(_name) (["type": "invalid name", "name": (_name)])
#define ERR_WARNING(_stat, _op) \
  (["type": "warning", "operation": (_op), "state": (_stat), \
    "reason": vcdir_invalid_op_msgs[_stat][_op], "vcobj": this_object()])


// Caches.

private static string state;
private static object(Perm) pobj;
private static mapping(string:mixed) md; // 0 if this dir isn't created or is a file.
private static object/*(VCDir)*/ parent;
private static mapping(string:object/*(VCFile|VCDir)*/) contents = ([]);

void create (object/*(Workarea)*/ _wa, array(string) _path, object(cachedir) _dir,
	     int _userid, void|object(cachedir) _updir, void|object(Perm) _pobj)
{
  wa = _wa, path = _path, dir = _dir, userid = _userid, updir = _updir, pobj = _pobj;

  if (dir) {
    infofile = dir->find_file ("__info");
    if (!infofile) {
      infofile = dir->find_file ("__info", 1);
      infofile->isbinary (1);
      if (!infofile->new())
	throw (({"Couldn't add __info file in directory\n", backtrace()}));
      infofile->write_metadata (([]));
      if (!infofile->commit ("create by " + userid + ", , \n"))
	throw (({"Couldn't commit __info file\n", backtrace()}));
    }
  }

  OPMSG ("create" + (dir ? "" : " (unreal)"));
}

void reset_caches (void|object(Perm) _pobj)
{
  pobj = _pobj;
  state = 0;
}

void reset_contents()
{
  contents = ([]);
}

/*private static*/ string update_state()
// For internal use only.
{
  if (!(dir || getdir())) {
    object(cachefile) thisfile;
    if (getupdir() && (thisfile = updir->find_file (path[-1]))) {
      if (thisfile && thisfile->metadata()->__removed) state = "R";
      else state = "E";
    }
    else state = "N";
    md = 0;
  }
  else {
#ifdef SBDEBUG
    if (!infofile) throw (({"Hasn't got cachefile for __info\n", backtrace()}));
#endif
    if (infofile->status()->status & UNKNOWN)
      // Dir exists in another work area but isn't created here yet.
      state = "N";
    else {
      md = infofile->metadata();
      state = md->__removed ? "R" : "E";
    }
  }

#ifdef SBDEBUG_STATE
  OPMSG ("update_state resolved state " + state);
#endif

  return state;
}

private static object/*(VCDir)*/ update_parent()
{
  if (parent) return parent;
  return sizeof (path) && (parent = wa->_vcobj (userid, path[..sizeof (path) - 2]));
}

/*private static*/ nomask mapping(string:object) update_contents()
// For internal use only.
{
  // This assumes that contents is emptied when a name disappears.
  if (!dir) return ([]);

  // Maybe this will counter the indexing of 0 that sometimes occurs
  // in list(). It shouldn't be needed, though.
  string name;
  while (name = search (contents, 0)) m_delete (contents, name);

  array(string) list = dir->get_dir() - ({"__info"});
  if (sizeof (list) != sizeof (contents))
    foreach (list, string name)
      if (!contents[name])
	if (!(contents[name] = wa->_vcobj (userid, path + ({name}))))
	  throw (({"Failed to create vcobj /" + path * "/" + "/" + name +
		   " from dir list\n", backtrace()}));

  return contents;
}


// Permission handling.

#define GET_POBJ(_id) \
  if (!pobj) { \
    int ppid = wa->ac_find_file_pp (path); \
    pobj = Perm (wa->fileperm (path, _id, ppid), ppid); \
  }

nomask int perm (object id)
{
  GET_POBJ (id);
  return pobj->perm;
}

nomask object(Perm) permobj (object id)
// Returns the permission bundled with the ppid for race safety.
{
  GET_POBJ (id);
  return pobj;
}

object has_ppoint()
{
  // FIXME: Have to check security ppoint?
  return wa->site->ac_check_file_pp (path * "/") && this_object();
}


// State queries (uses the cached state).

private static constant ustat_normal = (["N": "none", "E": "exists", "R": "none"]);

private static constant ustat_undelete = (["N": "none", "E": "exists", "R": "removed"]);

private static constant valid_ops_normal = ([
  "N": (<"ascend",
	 "add", "join">),
  "E": (<"ascend", "descend", "metadata", "list", "list_recursive", "log",
	 "newfile", "newdir", "set_metadata",
	 "discard", "update", "commit", "delete", "copy", "move", "join", "purge">),
  "R": (<"ascend", "descend", "log",
	 "add", "discard", "update", "commit", "join", "purge">),
]);

private static constant valid_ops_undelete = ([
  "N": valid_ops_normal["N"],
  "E": valid_ops_normal["E"],
  "R": valid_ops_normal["R"] +
  (<"metadata", "list", "list_recursive", "log", "undelete">),
]);

private static constant user_ops_normal = ([
  "N": valid_ops_normal["N"],
  "E": valid_ops_normal["E"] - (<"set_metadata">),
  "R": valid_ops_normal["R"],
]);

private static constant user_ops_undelete = ([
  "N": (<"ascend">),
  "E": (<"ascend", "descend", "metadata", "list", "list_recursive">),
  "R": (<"ascend", "descend", "metadata", "list", "list_recursive", "undelete">),
]);

private static constant op_perms = ([
  "ascend":		PERM_READ, // Perm tested on parent.
  "descend":		PERM_READ, // Perm tested on child.
  "metadata":		PERM_READ,
  "list":		PERM_READ,
  "list_recursive":	PERM_READ,
  "log":		PERM_READ,
  "newfile":		PERM_WRITE,
  "newdir":		PERM_WRITE,
  "add":		PERM_WRITE,
  "set_metadata":	PERM_WRITE,
  "discard":		PERM_READ, // Perm tested on each file too.
  "update":		PERM_READ, // Perm tested on each file too.
  "commit":		PERM_READ, // Perm tested on each file too.
  "delete":		PERM_WRITE,
  "undelete":		PERM_WRITE,
  "copy":		PERM_READ,
  "move":		PERM_WRITE,
  "join":		PERM_WRITE, // Perm tested on each file too.
  "purge":		PERM_WRITE,
]);

private static constant op_dsc = ([
  "metadata":		"view metadata",
  "list":		"list directory",
  "list_recursive":	"list directory recursively",
  "log":		"show version log",
  "newfile":		"add a file",
  "newdir":		"add a directory",
  "add":		"create directory",
  "set_metadata":	"change metadata",
  "discard":		"discard local files",
  "update":		"update local files",
  "commit":		"commit local files",
  "delete":		"delete directory",
  "undelete":		"undelete directory",
  "copy":		"copy directory tree",
  "move":		"move directory tree",
  "join":		"join directory tree between work areas",
  "purge":		"purge directory tree",
]);

private static object/*(VCDir)*/ check_op (object id, string operation,
					   int undelete_mode, int superuser, mapping ops)
{
  if (operation == "ascend")
    return update_parent() && parent->exists (id, undelete_mode, superuser) &&
      this_object();
  if (!superuser && sizeof (path) && path[-1][..1] == "__" ||
      min (superuser ? PERM_WRITE : perm (id),
	   userid ? PERM_WRITE : PERM_READ) < op_perms[operation]) return 0;

  string stat = state || update_state();
  if (!md) return 0;

  switch (operation) {
    case "purge":
      return ops[stat]["purge"] && (superuser || wa->site->purge_access (id)) &&
	this_object();
    case "add_ppoint":
      return stat != "N" && (superuser || wa->site->ppoint_edit_access (id)) &&
	!has_ppoint() && this_object();
    case "remove_ppoint":
      return sizeof (path) && (superuser || wa->site->ppoint_edit_access (id)) &&
	has_ppoint() && this_object();
    case "add":
    case "undelete":
      if (update_parent() && !parent->exists (id, 0, superuser)) return 0;
    default:
      return ops[stat][operation] && this_object();
  }
}

nomask string userstat (object id, void|int undelete_mode, void|int superuser)
// Returns the directory status to show to the user; "none", "exists"
// or "removed". Returns 0 if this directory shouldn't be shown at all.
{
  if (!superuser &&
      (sizeof (path) && path[-1][..1] == "__" || perm (id) == PERM_NONE)) return 0;
  return (undelete_mode ? ustat_undelete : ustat_normal)[state || update_state()];
}

nomask string userstat_ea (object id, void|int undelete_mode, void|int superuser)
  {return "none";}
nomask string userstat_rep (object id, void|int undelete_mode, void|int superuser)
  {return userstat (id, undelete_mode, superuser);}
// For function call compatibility with VCFile.

nomask object/*(VCDir)*/ is_valid_op (object id, string operation,
				      void|int undelete_mode, void|int superuser)
// Returns this object if the operation, which is a name of one of the
// directory operation functions, is valid.
{
  return check_op (id, operation, undelete_mode, superuser,
		   undelete_mode ? valid_ops_undelete : valid_ops_normal);
}

nomask object/*(VCDir)*/ is_user_op (object id, string operation,
				     void|int undelete_mode, void|int superuser)
// Returns this object if the operation, which is a name of one of the
// dir operation functions, should be available to the user.
{
  return check_op (id, operation, undelete_mode, superuser,
		   undelete_mode ? user_ops_undelete : user_ops_normal);
}

nomask mapping access_denied (object id, string operation)
// Returns an access denied error mapping using the correct ppoint.
{
  GET_POBJ (id);
  switch (op_perms[operation]) {
    case PERM_READ:
      return ERR_ACCESS_DENIED (
	pobj->ppid, "Read access required to " + op_dsc[operation] + ".");
    case PERM_WRITE:
      return ERR_ACCESS_DENIED (
	pobj->ppid, "Write access required to " + op_dsc[operation] + ".");
    default: throw (({"Bogus permission check\n", backtrace()}));
  }
}

nomask object/*(VCFile)*/ accessible (object id)
// Returns this object if the dir is accessible for non-superusers.
{
  return !(sizeof (path) && path[-1][..1] == "__" || perm (id) == PERM_NONE) &&
    (dir || getdir()) && this_object();
}

nomask object/*(VCDir)*/ exists (object id, void|int undelete_mode, void|int superuser)
// Returns this object if the dir exists.
{
  if (!superuser && (sizeof (path) && path[-1][..1] == "__" || perm (id) == PERM_NONE) ||
      !(dir || getdir())) return 0;
  return (undelete_mode ? ustat_undelete : ustat_normal)
    [state || update_state()] != "none" && this_object();
}

nomask string abspath (object id, void|int undelete_mode, void|int superuser)
// Returns the absolute path without leading and trailing '/'.
{
  return exists (id, undelete_mode, superuser) && path * "/";
}

nomask string real_abspath() {return path * "/";}

nomask string name (object id, void|int undelete_mode, void|int superuser)
// Returns 0 for the root directory.
{
  return exists (id, undelete_mode, superuser) && sizeof (path) && path[-1];
}

nomask string real_name() {return sizeof (path) && path[-1];}

nomask array stat (object id, void|int undelete_mode, void|int superuser)
// Returns the directory stat. For some fields, we don't even pretend
// that they have the proper meaning.
{
  int _perm = min (superuser ? PERM_WRITE : perm (id), userid ? PERM_WRITE : PERM_READ);
  if (!superuser && (sizeof (path) && path[-1][..1] == "__" || _perm == PERM_NONE) ||
      !(dir || getdir()) ||
      (undelete_mode ? ustat_undelete : ustat_normal)[state || update_state()] == "none")
    return 0;

  array(int|string) s = dir->stat();
  if (!s) return 0;
  s[0] = _perm == PERM_WRITE ? 0777 : 0555;
  s[5] = "none";
  s[6] = userstat (id, undelete_mode, 1);
  return s;
}


// Directory operations.

// Internals.

#define PERM_CHECK(_id, _op, _superuser) \
  if (!userid && op_perms[_op] > PERM_READ) \
    /* This is a viewarea; setting the superuser flag doesn't help. */ \
    return ERR_LOW_INVALID_OP ( \
      state, _op, "Write access required to " + op_dsc[_op] + "."); \
  if (!superuser) { \
    GET_POBJ (_id); \
    if (pobj->perm < op_perms[_op]) return access_denied (_id, _op); \
  }

#define EDIT_INIT(_valid_ops, _state, _operation) \
  if (!(dir || getdir())) return ERR_INVALID_DIRECTORY; \
  update_state(); \
  OPMSG (_operation + " called in state " + _state); \
  if (!(_valid_ops)[_state][_operation]) \
    return ERR_INVALID_OPERATION (_state, _operation)

#ifdef THREADS

#define WA_UPDATE_LOCK \
  object wa_update_mutex_lock; catch (wa_update_mutex_lock = wa->wa_update_mutex->lock())
// A lock on the whole work area to ensure consistent state in the
// repository, since all destructive operations commit immediately.
// Use a catch to ignore recursive locks.

#define WA_UPDATE_LOCK_2(_wa_1, _wa_2) \
  array(object) wa_update_locks = wa_update_lock_2 (_wa_1, _wa_2)

#else  // THREADS

#define WA_UPDATE_LOCK
#define WA_UPDATE_LOCK_2

#endif

// Protection point handling.

mapping|void add_ppoint (object id, void|int superuser)
{
  if (!superuser && !wa->site->ppoint_edit_access (id))
    return ERR_ACCESS_DENIED (
      wa->site->ppoint_edit_ppid(),
      "You don't have permission to add or remove protection points.");

  string p = path * "/";
  int ppid = wa->site->ac_check_file_pp (p);
  if (ppid) return ERR_LOW_INVALID_OP (
    state, "add_ppoint", "There already is a protection point on the directory.");

  if (!wa->site->ac_add_dir_pp (p))
    throw (({"Failed to add ppoint on directory /" + p + "\n", backtrace()}));

  OPMSG ("add_ppoint");
}

mapping|void remove_ppoint (object id, void|int superuser)
{
  if (!superuser && !wa->site->ppoint_edit_access (id))
    return ERR_ACCESS_DENIED (
      wa->site->ppoint_edit_ppid(),
      "You don't have permission to add or remove protection points.");

  string p = path * "/";
  int ppid = wa->site->ac_check_file_pp (p);
  if (!ppid) return ERR_LOW_INVALID_OP (
    state, "remove_ppoint", "There is no protection point on the directory.");

  if (!wa->site->ac_remove_file_pp (p))
    throw (({"Failed to remove ppoint from directory /" + p + "\n", backtrace()}));

  OPMSG ("remove_ppoint");
}

// Nondestructive operations (do not lock the VCDir object).

mapping|object/*(VCDir)*/ ascend (object id, void|int undelete_mode,
				  void|int superuser)
// Returs the parent directory or 0 if at root. The permission and
// undelete_mode checks are done on the parent, not this directory.
{
  if (!sizeof (path)) return 0;
  if (!update_parent()) return ERR_INVALID_DIRECTORY;
  if (!superuser && !parent->accessible (id))
    return access_denied (id, "ascend");
  if (!parent->exists (id, undelete_mode, superuser)) return 0;

  parent->reset_caches();
  return parent;
}

mapping|object/*(VCFile|VCDir)*/ descend (object id, string name, void|int undelete_mode,
					  void|int superuser)
// The permission and undelete_mode checks are done on the file/subdir
// name, not this directory. Returns 0 if name doesn't exist.
{
  if (!(dir || getdir())) return ERR_INVALID_DIRECTORY;
  object vcobj = contents[name] || update_contents()[name];
  if (!vcobj) return 0;
  if (!superuser && !vcobj->accessible (id))
    return access_denied (id, "descend");
  if (!vcobj->exists (id, undelete_mode, superuser)) return 0;

  vcobj->reset_caches();
  return vcobj;
}

mapping|object(VCFileData) metadata (object id, void|int undelete_mode,
				     void|int superuser)
// Returns the metadata in a one VCFileData object.
{
  PERM_CHECK (id, "metadata", superuser);
  if (!(dir || getdir())) return ERR_INVALID_DIRECTORY;

  if (!(undelete_mode ? valid_ops_undelete : valid_ops_normal)
      [state || update_state()]["metadata"])
    return ERR_INVALID_OPERATION (state, "metadata");

  mapping(string:mixed) _md = md || (update_state(), md);
  if (_md->__removed) _md -= (["__removed": 1]);
  return VCFileData (_md, 0);
}

mapping|array(object/*(VCFile|VCDir)*/) list (object id, void|int undelete_mode,
					      void|int superuser)
// Returns an array of _all_ files and directories in this directory.
// To get e.g. the names of those files that the current request can
// view and those dirs it can list, do
//
//	vcdir->list(id)->is_user_op(id, "metadata") - ({0})
//
// ("metadata" is handy since its a common operation on both files and
// dirs that has the same requirements on permissions and
// undeletedness as "view" for files and "list" for dirs.)
{
  PERM_CHECK (id, "list", superuser);
  if (!(dir || getdir())) return ERR_INVALID_DIRECTORY;

  if (!(undelete_mode ? valid_ops_undelete : valid_ops_normal)
      [state || update_state()]["list"])
    return ERR_INVALID_OPERATION (state, "list");

  array(object/*(VCFile|VCDir)*/) lst = values (update_contents());
  lst->reset_caches();
  return lst;
}

mapping|array(object/*(VCFile|VCDir)*/) list_recursive (object id, void|int undelete_mode,
							void|int superuser)
// Like list(), but lists all files and directories recursively. The
// returned array is preordered. Recursively generated errors are
// ignored; a simple error mapping is only returned if it applies to
// this object.
{
  PERM_CHECK (id, "list_recursive", superuser);
  if (!(dir || getdir())) return ERR_INVALID_DIRECTORY;

  if (!(undelete_mode ? valid_ops_undelete : valid_ops_normal)
      [state || update_state()]["list_recursive"])
    return ERR_INVALID_OPERATION (state, "list_recursive");

  array(object/*(VCFile|VCDir)*/) lst = values (update_contents());
  lst->reset_caches();

  foreach (lst, object vcobj)
    if (vcobj->isdir) {
      mixed res = vcobj->list_recursive (id, undelete_mode, superuser);
      if (arrayp (res)) lst += res;
    }
  return lst;
}

mapping|array(object(VCLogEntry)) log (object id,
				       void|string revisions, void|string dates,
				       void|int undelete_mode, void|int superuser)
// revisions and dates are in the formats accepted by cvs log -r and
// -d, respectively.
{
  PERM_CHECK (id, "log", superuser);
  if (!(dir || getdir())) return ERR_INVALID_DIRECTORY;

  if (!(undelete_mode ? valid_ops_undelete : valid_ops_normal)
      [state || update_state()]["log"])
    return ERR_INVALID_OPERATION (state, "log");

  return convert_log (infofile->file->log (revisions, dates));
}

// Destructive operations (lock the VCFile object).

mapping|object/*(VCFile)*/ newfile (object id, string name, void|int superuser)
// Returns any error from VCFile.add().
{
  PERM_CHECK (id, "newfile", superuser);
  EDIT_INIT (valid_ops_normal, state, "newfile");

  name = clean_name (name);
  object vcobj = wa->_vcobj (userid, path + ({name}));
  if (vcobj && vcobj->isfile) vcobj->reset_caches();
  else vcobj = VCFile (wa, path + ({name}), 0, userid);
  mapping res = vcobj->add (id, superuser);
  if (mappingp (res)) return res;

  OPMSG ("newfile " + name + " ok");
  return vcobj;
}

mapping|object/*(VCDir)*/ newdir (object id, string name, string comment,
				  void|int superuser)
// Returns any error from VCDir.add().
{
  PERM_CHECK (id, "newdir", superuser);
  EDIT_INIT (valid_ops_normal, state, "newfile");

  name = clean_name (name);
  object vcobj = wa->_vcobj (userid, path + ({name}));
  if (vcobj && vcobj->isdir) vcobj->reset_caches();
  else vcobj = Sitebuilder["VCDir"] (wa, path + ({name}), 0, userid);
  mapping res = vcobj->add (id, comment, superuser);
  if (mappingp (res)) return res;

  OPMSG ("newdir " + name + " ok");
  return vcobj;
}

mapping|void add (object id, string comment, void|int superuser)
// Returns an error of type "invalid name" if the directory name is
// invalid.
{
  WA_UPDATE_LOCK;
  PERM_CHECK (id, "add", superuser);
  update_state();
  if (!valid_ops_normal[state]["add"]) return ERR_INVALID_OPERATION (state, "add");
  if (update_parent() && !parent->exists (id, 0, superuser))
    return ERR_INVALID_DIRECTORY;

  switch (state) {
    case "N":
      if (!getupdir()) return ERR_INVALID_DIRECTORY;
      if (!sizeof (path) || !superuser && !valid_name (path[-1]))
	return ERR_INVALID_NAME (sizeof (path) ? path[-1] : "");
      dir = updir->mkdir (path[-1]);
      if (!dir) throw (({"Couldn't add new directory\n", backtrace()}));
      infofile = dir->find_file ("__info", 1);
      if (!infofile->new())
	throw (({"Couldn't add __info file in new directory\n", backtrace()}));
      wa->_intern_vcobj (userid, path, this_object());
      md = ([]);
      break;

    case "E":
      return ERR_INVALID_OPERATION (state, "add");

    case "R":
      if (!dir) return ERR_INVALID_OPERATION (state, "add");
      md -= (["__removed": 1]);
      infofile->unfake();
  }

  infofile->write_metadata (md);
  if (!infofile->commit (sprintf ("create by %d, %s, %s\n%s",
				  userid,
				  wa->mac->id_get_handle (userid) || "",
				  wa->mac->id_get_name (userid) || "",
				  comment)))
    throw (({"Couldn't commit __info file\n", backtrace()}));

  update_state();
  OPMSG ("add ok");
}

mapping|void set_metadata (object id, mapping metadata, void|int superuser)
{
  WA_UPDATE_LOCK;
  PERM_CHECK (id, "set_metadata", superuser);
  EDIT_INIT (valid_ops_normal, state, "set_metadata");

  infofile->unfake();
  md = metadata;
  infofile->write_metadata (md);
  if (!infofile->commit (sprintf ("edit by %d, %s, %s\n",
				  userid,
				  wa->mac->id_get_handle (userid) || "",
				  wa->mac->id_get_name (userid) || "")))
    throw (({"Couldn't commit __info file\n", backtrace()}));

  update_state();
  OPMSG ("set_metadata ok");
}

mapping|array(mapping) discard (object id, void|int warning, void|int superuser)
// Calls discard() recursively. If warning is nonzero, all generated
// error mappings of type "warning" are returned in an array,
// otherwise an empty array is returned. Recursively generated errors
// are ignored; a simple error mapping is only returned if it applies
// to this object.
{
  PERM_CHECK (id, "discard", superuser);
  EDIT_INIT (valid_ops_normal, state, "discard");

  array(mapping) res = ({});
  if (warning)
    foreach (values (update_contents()), object vcobj) {
      mixed r = vcobj->discard (id, warning, superuser);
      if (mappingp (r) && r->type == "warning") res += ({r});
      else if (arrayp (r)) res += r;
    }
  else values (update_contents())->discard (id, warning, superuser);

  OPMSG ("discard ok");
  return res;
}

mapping|array(object(VCFileConflict)) update (object id, void|int superuser)
// Calls update() recursively on all files. Returns an array of
// VCFileConflict objects for the files where there are differences to
// resolve. Recursively generated errors are ignored; a simple error
// mapping is only returned if it applies to this object.
{
  PERM_CHECK (id, "update", superuser);
  EDIT_INIT (valid_ops_normal, state, "update");

  array(object(VCFileConflict)) res = ({});
  foreach (values (update_contents()), object vcobj) {
    mixed r = vcobj->update (id, superuser);
    if (objectp (r)) res += ({r});
    else if (arrayp (r)) res += r;
  }

  OPMSG ("update ok");
  return res;
}

mapping|array(mapping) commit (object id, string comment, void|int superuser)
// Calls commit() recursively on all files which are locally added,
// modified or removed. Returns an array of all generated error
// mappings. A simple error mapping is only returned if it applies to
// this object.
//
// FIXME: Use locks to make this atomic.
{
  PERM_CHECK (id, "commit", superuser);
  EDIT_INIT (valid_ops_normal, state, "commit");

  array(mapping) res = ({});
  foreach (values (update_contents()), object vcobj)
    if (vcobj->isdir ||
	(<"added", "modified", "removed">)[vcobj->userstat_ea (id, 0, superuser)]) {
      mixed r = vcobj->commit (id, comment, superuser);
      if (mappingp (r)) res += ({r});
      else if (arrayp (r)) res += r;
    }

  OPMSG ("commit ok");
  return res;
}

int can_delete (object id )
{
  foreach (wa->site->find_dirhier_pp_paths (path * "/"), string pp_path)
  {
    object vcobj = wa->_vcobj (userid, pp_path / "/");
    if (vcobj && vcobj->exists (0, 0, 1))
    {
      object(Perm) permobj = vcobj->permobj (id);

      if (permobj->perm < PERM_WRITE)
	return 0;
    }
  }
  return 1;
}

mapping|void delete (object id, string comment, void|int warning, void|int superuser)
{
  WA_UPDATE_LOCK;
  PERM_CHECK (id, "delete", superuser);
  EDIT_INIT (valid_ops_normal, state, "delete");

  if (!superuser && !can_delete (id))
    return ERR_ACCESS_DENIED (
      permobj (id)->ppid, "Write access to the whole directory hierarchy is required.");
  if (warning && sizeof (values (update_contents())->exists (id, 0, 1) - ({0})))
    return ERR_WARNING (state, "delete");

  // There's a race on the permission check here, but that's harmless.

  mixed res = list (0, 0, 1);
  if (mappingp (res)) throw (({"Unexpected error from list(): " +
			       format_error (res) + "\n", backtrace()}));
  res = res->exists (0, 0, 1) - ({0});

  array(object) vcdirs =
    Array.filter (res, lambda (object vcobj) {return vcobj->isdir;});
  array(object) vcfiles =
    Array.filter (res, lambda (object vcobj) {return vcobj->isfile;});

  res = vcdirs->delete (0, comment, 0, 1) - ({0});
  if (sizeof (res)) throw (({"Unexpected error from VCDir.delete():\n" +
			     Array.map (res, format_error) * "\n" + "\n", backtrace()}));
  res = vcfiles->delete (0, 0, 1) - ({0});
  if (sizeof (res)) throw (({"Unexpected error from VCFile.delete():\n" +
			     Array.map (res, format_error) * "\n" + "\n", backtrace()}));
  vcfiles = vcfiles->exists (0, 0, 1) - ({0}); // May have files in N/N state here.
  res = vcfiles->commit (0, comment, 1) - ({0});
  if (sizeof (res)) throw (({"Unexpected error from VCFile.commit():\n" +
			     Array.map (res, format_error) * "\n" + "\n", backtrace()}));

  md->__removed = 1;
  infofile->unfake();		// Shouldn't change the metadata.
  infofile->write_metadata (md);
  if (!infofile->commit (sprintf ("delete by %d, %s, %s\n%s",
				  userid,
				  wa->mac->id_get_handle (userid) || "",
				  wa->mac->id_get_name (userid) || "",
				  comment)))
    throw (({"Couldn't commit __info file\n", backtrace()}));

  update_state();
  OPMSG ("delete ok");
}

mapping|void undelete (object id, string comment, void|int superuser)
// Does not undelete recursively.
{
  WA_UPDATE_LOCK;
  PERM_CHECK (id, "undelete", superuser);
  EDIT_INIT (valid_ops_undelete, state, "undelete");
  if (update_parent() && !parent->exists (id, 0, superuser))
    return ERR_INVALID_DIRECTORY;

  m_delete (md, "__removed");
  infofile->unfake();		// Shouldn't change the metadata.
  infofile->write_metadata (md);
  if (!infofile->commit (sprintf ("undelete by %d, %s, %s\n%s",
				  userid,
				  wa->mac->id_get_handle (userid) || "",
				  wa->mac->id_get_name (userid) || "",
				  comment)))
    throw (({"Couldn't commit __info file\n", backtrace()}));

  update_state();
  OPMSG ("undelete ok");
}

// FIXME?
// mapping|void revert (object id, int warning, int superuser, object(VCOldDir) vcodir)

mapping|void copy (object id, object/*(VCDir)*/ to, string comment,
		   void|int superuser,
		   /*void|object(VCOldDir) vcodir FIXME? */)
{
  WA_UPDATE_LOCK;
  PERM_CHECK (id, "copy", superuser);
  EDIT_INIT (valid_ops_normal, state, "copy");

  mapping|array(object) l = list(id, 0, superuser);
  if(mappingp(l))
    return l;

  foreach(l, object vco)
    if(vco->is_valid_op (id, (vco->isdir ? "list" : "view"), 0, superuser)) {
      string name = vco->name(id, 0, superuser);
      mapping|object dvco;
      if(vco->isfile)
	dvco = to->newfile(id, name, superuser);
      else
	dvco = to->newdir(id, name, comment, superuser);
      if(mappingp(dvco))
	return dvco;
      mapping err;
      if(vco->isfile)
	err = vco->copy(id, dvco, superuser);
      else
	err = vco->copy(id, dvco, comment, superuser);
      if(!mappingp(err) && dvco->isfile)
	err = dvco->commit(id, comment, superuser, 1);
      if(mappingp(err))
	return err;
    }
  
  OPMSG ("copy to " + to->real_abspath() + " ok");
}

mapping|void move (object id, object/*(VCDir)*/ to, string comment,
		   void|int superuser)
{
  WA_UPDATE_LOCK;
  PERM_CHECK (id, "move", superuser);
  EDIT_INIT (valid_ops_normal, state, "move");

  if (!superuser && !can_delete (id))
    return ERR_ACCESS_DENIED (
      permobj (id)->ppid, "Write access to the whole directory hierarchy is required.");

  mapping|array(object) l = list(id, 0, superuser);
  if(mappingp(l))
    return l;

  foreach(l, object vco)
    if(vco->is_valid_op (id, (vco->isdir ? "list" : "view"), 0, superuser)) {
      string name = vco->name(id, 0, superuser);
      mapping|object dvco;
      if(vco->isfile)
	dvco = to->newfile(id, name, superuser);
      else
	dvco = to->newdir(id, name, comment, superuser);
      if(mappingp(dvco))
	return dvco;
      mapping err;
      if(vco->isfile)
	err = vco->move(id, dvco, superuser);
      else
	err = vco->move(id, dvco, comment, superuser);
      if(!mappingp(err) && dvco->isfile)
	err = dvco->commit(id, comment, superuser, 1);
      if(mappingp(err))
	return err;
    }

  mapping|void derr = delete(id, comment, 0, superuser);

  if(mappingp(derr))
    return derr;
  
  update_state();
  OPMSG ("move to " + to->real_abspath() + " ok");
}

mapping|array(object(VCFileConflict)) join (object id, object/*(Workarea)*/ other,
					    void|int superuser)
// Calls join() recursively on all files. Returns an array of
// VCFileConflict objects for the files where there are differences to
// resolve. Recursively generated errors are collapsed to a simple
// error mapping.
//
// Note: Workarea.access_try() should be called on both work areas
// before this function to get the appropriately lowered permission.
{
  object/*(VCDir)*/ vcother = other->_vcobj (userid, path);
#ifdef SBDEBUG
  if (!vcother || !vcother->isdir)
    throw (({"No VCDir object in the other work area for /" + path * "/" + "\n",
	     backtrace()}));
#endif

  WA_UPDATE_LOCK_2 (wa, other);
  PERM_CHECK (id, "join", superuser);
  update_state();
  vcother->update_state();
  // No reason to check validness; join is always valid on dirs.

  OPMSG ("join called in state " + state);

  array(object) vcobjs = ({});
  mapping(string:object) vcother_contents = vcother->update_contents();
  foreach (indices (vcother_contents + update_contents()), string name) {
    object t = contents[name], o = vcother_contents[name];
    if (!t || !o)
      // t might be zero if the the other workarea contains a new file.
      return ERR_LOW_INVALID_OP (state, "join",
				 "There are local copies in one or both work areas; "
				 "use discard or update/commit first.");
    if (t->exists (0, 0, 1) || o->exists (0, 0, 1)) vcobjs += ({t});
  }

  array(object(VCFileConflict)) res = ({});
  foreach (vcobjs, object vcobj) {
    mixed r = vcobj->join (id, other, superuser);
    if (mappingp (r))
      switch (r->type) {
	case "invalid operation":
	  return ERR_LOW_INVALID_OP (state, "join",
				     "There are local copies in one or both work areas; "
				     "use discard or update/commit first.");
	case "access denied":
	  r->reason = "Write access to the whole directory hierarchy is required.";
	  return r;
	default:
	  throw (({"Unexpected error from join(): " +
		   format_error (r) + "\n", backtrace()}));
      }
    else if (objectp (r)) res += ({r});
    else if (arrayp (r)) res += r;
  }

  OPMSG ("join with '" + other->tag + "' ok");
  return res;
}

mapping|void join_commit (object id, object/*(Workarea)*/ other, string comment,
			  void|int superuser)
// Finishes and commits a join on all files in the directory tree as
// one atomic operation. In every (old) VCFileConflict object,
// state->md is used for the metadata and state->data for the
// contents. Returns the first error encountered. Note: Uses "join" as
// op string.
{
  object/*(VCDir)*/ vcother = other->_vcobj (userid, path);
#ifdef SBDEBUG
  if (!vcother || !vcother->isdir)
    throw (({"No VCDir object in the other work area for /" + path * "/" + "\n",
	     backtrace()}));
#endif

  WA_UPDATE_LOCK_2 (wa, other);
  OPMSG ("join_commit called");

  mapping|array(object(VCFileConflict)) vcfcs = join (id, other, superuser);
  if (mappingp (vcfcs)) return vcfcs;

  array(string) filenames = vcfcs->this_vcfile->real_abspath();
  sort (filenames, vcfcs);	// For canonical locking order of every EDIT_LOCK_2.
#ifdef THREADS
  array(array(object)) editlocks = allocate (sizeof (vcfcs));
#endif

  for (int i = 0; i < sizeof (vcfcs); i++) {
    object(VCFileConflict) vcfc = vcfcs[i];

#ifdef THREADS
    // Do EDIT_LOCK_2 for every file.
    editlocks[i] = edit_lock_2 (vcfc->this_vcfile, vcfc->other_vcfile);
#endif

    if (vcfc->this_vcfile->operation_in_progress != vcfc ||
	!vcfc->other_vcfile->operation_in_progress ||
	vcfc->other_vcfile->operation_in_progress->ptr != vcfc) {
      // Handle the race from the join() call above this way instead
      // of finding all vcfiles and edit_lock them before it. Since we
      // already got the wa_update_lock, the only possible races are
      // due to local changes, which affects operation_in_progress.
      mapping|object(VCFileConflict) res = vcfc->this_vcfile->join (id, other, superuser);
      if (mappingp (res)) return res; // FIXME: Should mangle this as in join() above..
      vcfc = vcfcs[i] = res;
    }

    if (zero_type (vcfc->state->md)) vcfcs[i] = 0; // Nothing to do.
    else if (vcfc->state->md && vcfc->state->md->__invalid) {
      mapping err = ERR_LOW_INVALID_OP (
	state, "join", "/" + filenames[i] + " has invalid contents.");
      err->vcobj = vcfc->this_vcfile;
      return err;
    }
  }

  // Irreversible things below.

  foreach (vcfcs - ({0}), object(VCFileConflict) vcfc)
    vcfc->this_vcfile->low_join_commit (vcfc->other_vcfile, vcfc, comment);

  OPMSG ("join_commit with '" + other->tag + "' ok");
}

mapping|void purge (object id, void|int superuser)
{
  WA_UPDATE_LOCK;
  PERM_CHECK (id, "purge", superuser);
  if (!superuser && !wa->site->purge_access (id)) return ERR_ACCESS_DENIED (
    wa->site->purge_ppid(), "You don't have permission to purge directories.");

  // Almost but not quite EDIT_INIT.
  if (!(dir || getdir())) return ERR_INVALID_DIRECTORY;
  update_state();
  OPMSG ("purge called in state " + state);

  throw (({"Not implemented\n", backtrace()}));

  update_state();
  OPMSG ("purge ok");
}
