// Handles a file with all the state and AC stuff.

// $Id: VCFile.pike,v 1.73 1999/11/01 19:24:40 mast Exp $

// Created by vcobj() or vcfile() in Sitebuilder.Workarea. There
// should be one unique object for the same work area, file and user.
// Invalid objects returned by vcfile() are not unique, though.
//
// The file status is cached and only resampled when update_state() is
// called, which is done in vcobj(), vcfile(), and all the destructive
// operations. Note that it's not safe to make any assumptions about
// this cache being constant for any amount of time since these
// objects are shared between requests.
//
// All the functions in this class (besides create() and
// reset_caches()) should be safe to call at any time; they do all
// needed sanity checking and return appropriate errors. The interface
// is as similar as possible to VCDir.
//
// Common arguments for functions:
//
// object id		The almighty and ever present id object. It's
//			only used for permission checks and may be zero
//			if the SUPERUSER flag is given to the function.
// int undelete_mode	As it says.
// int warning		If an operation that should give a warning is
//			about to be taken, return an error describing
//			the warning instead of do the operation.
// int superuser	Disables all permission checks when nonzero.
//
// The file operation functions return a mapping(string:mixed) when
// errors occurs. The mapping contains at least the binding "type":
// "errortype", where errortype is a magic string which is either
// standard or documented for each function.
//
// Standard error types:
//
// "access denied":	Access blocked by AC. Additional bindings:
//			"ppoint": The ID of the file access ppoint in use.
//			"reason": If used, a message about the reason.
//			"vcobj": The VCFile object.
// "invalid file":	As it says. This error is possible when the VCFile
//			object was created by vcfile() or when adding or
//			undeleting a file outside the dir tree.
// "invalid operation":	The operation is not defined or not allowed in
//			the current state (which may well have changed
//			in the last ns). Additional bindings:
//			"operation": The name of the function.
//			"state": The current state on the form "X/Y".
//			"reason": A short explanation for relevant cases.
//			"vcobj": The VCFile object.
// "invalid name":	Returned by add, upload and join when they try to
//			create a new file, but the name doesn't pass
//			Sitebuilder.valid_name(). Bindings:
//			"name": The offending name.
// "warning":		A warning of the "are you sure you want to
//			continue"-type. Only returned when the warning
//			flag was set. The additional bindings are the
//			same as for "invalid operation".

#if constant (roxen)
#include <config.h>
#endif

import Sitebuilder;
import VC.CVS;

constant isfile = 1;
object isfilep() {return this_object();}
object isdirp() {return 0;}

object/*(Workarea)*/ wa;


// Debugging tools.

#ifdef SBDEBUG
#define OPMSG(_msg) sbdebug ("VCFile " + userid + ":" + (wa && wa->tag) + \
			     ":/" + path * "/" + " " + _msg + "\n")
#else
#define OPMSG(_msg)
#endif


// Identity.

private static array(string) path; // A "clean" path, i.e. "" elements removed etc.
private static int userid;	// The AC identity, 0 if we're in a viewarea.

private static object(cachedir) dir; // The current directory. Created only when needed.
private static object(cachefile) file; // May be 0 in the N/N state.

private static object(cachedir) getdir()
{
  if (dir) return dir;
  object(cachedir) topdir = wa->topdir (userid);
  return dir = topdir->find_dir (path[..sizeof (path) - 2]);
}

/*private static*/ object(cachefile) getfile (void|int fake)
// For internal use only.
{
  if (file) return file;
  if (!sizeof (path)) return 0;
  if (getdir()) {
    file = dir->find_file (path[-1], fake);
    if (fake && file) file->isbinary (1);
  }
  return file;
}


// Error types.

#define ERR_ACCESS_DENIED(_ppoint, _msg) \
  (["type": "access denied", "ppoint": (_ppoint), \
    "reason": (_msg), "vcobj": this_object()])
constant err_invalid_file = (["type": "invalid file"]);
#define ERR_INVALID_FILE err_invalid_file
#define ERR_LOW_INVALID_OP(_stat, _op, _msg) \
  (["type": "invalid operation", "operation": (_op), "state": (_stat), \
    "reason": (_msg), "vcobj": this_object()])
#define ERR_INVALID_OPERATION(_stat, _op) \
  ERR_LOW_INVALID_OP (_stat, _op, vcfile_invalid_op_msgs[_stat][_op])
#define ERR_INVALID_NAME(_name) (["type": "invalid name", "name": (_name)]);
#define ERR_WARNING(_stat, _op) \
  (["type": "warning", "operation": (_op), "state": (_stat), \
    "reason": vcfile_invalid_op_msgs[_stat][_op], "vcobj": this_object()])
constant err_interrupted = (["type": "interrupted"]);
#define ERR_INTERRUPTED err_interrupted
constant err_repository_changed = (["type": "repository changed"]);
#define ERR_REPOSITORY_CHANGED err_repository_changed
constant err_removed_revision = (["type": "removed revision"]);
#define ERR_REMOVED_REVISION err_removed_revision
#define ERR_IO_ERROR(_err) (["type": "io error", "errno": _err])


// Caches.

private static object/*(file.status_class)*/ status;
private static string state;
private static object(Perm) pobj;
/*private static*/ mapping(string:mixed) md; // 0 if this file isn't created or is a dir.
private static object/*(VCDir)*/ parent;

void create (object/*(Workarea)*/ _wa, array(string) _path, object(cachefile) _file,
	     int _userid, void|object(cachedir) _dir, void|object(Perm) _pobj)
{
  wa = _wa, path = _path, file = _file, userid = _userid, dir = _dir, pobj = _pobj;
  OPMSG ("create" + (file ? "" : " (unreal)"));
}

void reset_caches (void|object(Perm) _pobj)
{
  pobj = _pobj;
  state = 0;
}

/*private static*/ nomask string update_state()
// For internal use only.
{
  object(cachefile) _file = file || getfile();

  object/*(file.status_class)*/ _status;
  mapping(string:mixed) _md = 0;

  if (!_file) {
    object(cachedir) thisdir;
    if (getdir() && (thisdir = dir->find_dir (path[-1]))) {
      object infofile = thisdir->find_file ("__info");
      if (infofile && infofile->metadata()->__removed) state = "N/R";
      else state = "N/E";
    }
    else state = "N/N";
  }
  else {
    // FIXME: Is this the right order? Is there a possible simple
    // order? parent_metadata() is involved too.
    _status = _file->status();
    _md = ([]);

    int flags = _status->status;

    if (flags & (UNKNOWN|FAKE) || !userid) // N/*
      // N/* is also used for all viewarea states.
      if (!_status->repository_version) state = "N/N";
      else if ((_md = _file->metadata())->__removed) state = "N/R";
      else state = "N/E";
    else if ((_md = _file->metadata())->__removed) // R/*
      if (_file->parent_metadata()->__removed) state = "R/R";
      else state = "R/E";
    else if (_md->__ancestor_removed)
      // Ancestor is removed. We lie to ourselves by faking the
      // implicit M/MR-M/R transition on this level; the R/R-N/R
      // implicit transition is handled on entry to every destructive
      // operation without lying.
      if (_file->parent_metadata()->__removed) state = "M/R";
      else state = "M/M";
    else if ((<0, "New">)[_status->working_version]) // A/*
      if (flags & (RMODIFIED|CONFLICT))
	if (_file->parent_metadata()->__removed) state = "A/MR";
	else state = "A/M";
      else state = "A/N";
    else if (flags & LMODIFIED)	// M/*
      if (flags & (RMODIFIED|CONFLICT))	// M/M*
	if (_file->parent_metadata()->__removed) state = "M/MR";
	else state = "M/M";
      else if (_file->parent_metadata()->__removed) state = "M/R";
      else state = "M/E";
    else			// E/*
      if (flags & (RMODIFIED|CONFLICT))	// E/M*
	if (_file->parent_metadata()->__removed) state = "E/MR";
	else state = "E/M";
      else state = "E/E";
  }

#ifdef SBDEBUG_STATE
  OPMSG ("update_state resolved state " + state);
#endif

  status = _status, md = _md;
  return state;
}

private static object/*(VCDir)*/ update_parent()
{
  if (parent) return parent;
  return parent = wa->_vcobj (userid, path[..sizeof (path) - 2]);
}

// Permission handling.

#define GET_POBJ(_id) \
  if (!pobj) { \
    int ppid = wa->ac_find_file_pp (path); \
    pobj = Perm (wa->fileperm (path, _id, ppid), ppid); \
  }

nomask int perm (object id)
{
  GET_POBJ (id);
  return pobj->perm;
}

nomask object(Perm) permobj (object id)
// Returns the permission bundled with the ppid for race safety.
{
  GET_POBJ (id);
  return pobj;
}

object/*(VCFile)*/ has_ppoint()
{
  return wa->site->ac_check_file_pp (path * "/") && this_object();
}


// State queries (uses the cached state).

private static constant ustat_ea = ([
  "N/N":  "none",	"N/E":	"none",		"N/R":	"none",
  "E/E":  "exists",	"E/M":	"exists",	"E/MR":	"exists",
  "A/N":  "added",	"A/M":	"added",	"A/MR":	"added",
  "M/E":  "modified",	"M/M":	"add/mod",	"M/MR":	"add/mod",	"M/R":  "add/mod",
  "R/E":  "removed",	"R/R":	"none",
]);

private static constant ustat_rep_normal = ([
  "N/N":  "none",	"N/E":	"exists",	"N/R":	"none",
  "E/E":  "exists",	"E/M":	"modified",	"E/MR":	"removed",
  "A/N":  "none",	"A/M":	"modified",	"A/MR":	"none",
  "M/E":  "exists",	"M/M":	"modified",	"M/MR":	"non/rem",	"M/R":	"non/rem",
  "R/E":  "exists",	"R/R":	"none",
]);

private static constant ustat_rep_undelete = ([
  "N/N":  "none",	"N/E":	"exists",	"N/R":	"removed",
  "E/E":  "exists",	"E/M":	"modified",	"E/MR":	"removed",
  "A/N":  "none",	"A/M":	"modified",	"A/MR":	"removed",
  "M/E":  "exists",	"M/M":	"modified",	"M/MR":	"removed",	"M/R":  "removed",
  "R/E":  "exists",	"R/R":	"removed",
]);

private static constant valid_ops_normal = ([
  "N/N":  (<"ascend",
	    "add", "upload",
	    "join">),
  "N/E":  (<"ascend", "metadata", "view", "log",
	    "download", "upload",
	    // Allowing upload here introduces races, but users think
	    // it's "annoying" otherwise.. :P
	    "delete", "revert", "copy", "move", "join", "purge">),
  "N/R":  (<"ascend",
	    "add", "upload",
	    "revert", "join", "purge">),
  "E/E":  (<"ascend", "metadata", "view", "log",
	    "download", "set_metadata", "upload",
	    "discard", "delete", "revert", "copy", "move", "purge">),
  "E/M":  (<"ascend", "metadata", "view", "log",
	    "download", "set_metadata", "upload",
	    "discard", "delete", "revert", "copy", "move", "purge">),
  "E/MR": (<"ascend", "metadata", "view", "log",
	    "download", "set_metadata", "upload",
	    "discard", "delete", "revert", "copy", "move", "purge">),
  "A/N":  (<"ascend", "metadata", "view",
	    "download", "set_metadata", "upload",
	    "discard", "delete", "commit", "copy", "move", "purge">),
  "A/M":  (<"ascend", "metadata", "view", "log",
	    "download", "set_metadata", "upload",
	    "discard", "update", "delete", "revert", "copy", "move", "purge">),
  "A/MR": (<"ascend", "metadata", "view", "log",
	    "download", "set_metadata", "upload",
	    "discard", "delete", "revert", "commit", "copy", "move", "purge">),
  "M/E":  (<"ascend", "metadata", "view", "log",
	    "download", "set_metadata", "upload",
	    "discard", "delete", "revert", "commit", "copy", "move", "purge">),
  "M/M":  (<"ascend", "metadata", "view", "log",
	    "download", "set_metadata", "upload",
	    "discard", "update", "delete", "revert", "copy", "move", "purge">),
  "M/MR": (<"ascend", "metadata", "view", "log",
	    "download", "set_metadata", "upload",
	    "discard", "update", "delete", "revert", "copy", "move", "purge">),
  "M/R":  (<"ascend", "metadata", "view", "log",
	    "download", "set_metadata", "upload",
	    "discard", "delete", "revert", "commit", "copy", "move", "purge">),
  "R/E":  (<"ascend", "metadata", "log",
	    "discard", "delete", "undelete", "revert", "commit", "purge">),
  "R/R":  (<"ascend",
	    "add", "upload",
	    "revert", "join", "purge">),
]);

private static constant valid_ops_undelete = ([
  "N/N":  valid_ops_normal["N/N"],
  "N/E":  valid_ops_normal["N/E"],
  "N/R":  valid_ops_normal["N/R"] + (<"metadata", "view", "log", "undelete">),
  "E/E":  valid_ops_normal["E/E"],
  "E/M":  valid_ops_normal["E/M"],
  "E/MR": valid_ops_normal["E/MR"] + (<"undelete">),
  "A/N":  valid_ops_normal["A/N"],
  "A/M":  valid_ops_normal["A/M"],
  "A/MR": valid_ops_normal["A/MR"],
  "M/E":  valid_ops_normal["M/E"],
  "M/M":  valid_ops_normal["M/M"],
  "M/MR": valid_ops_normal["M/MR"],
  "M/R":  valid_ops_normal["M/R"],
  "R/E":  valid_ops_normal["R/E"] + (<"view">),
  "R/R":  valid_ops_normal["R/R"] + (<"metadata", "view", "log", "undelete">),
]);

private static constant user_ops_normal = ([
  "N/N":  valid_ops_normal["N/N"],
  "N/E":  valid_ops_normal["N/E"],
  "N/R":  valid_ops_normal["N/R"],
  "E/E":  valid_ops_normal["E/E"],
  "E/M":  valid_ops_normal["E/M"],
  "E/MR": valid_ops_normal["E/MR"],
  "A/N":  valid_ops_normal["A/N"] - (<"discard">),
  "A/M":  valid_ops_normal["A/M"],
  "A/MR": valid_ops_normal["A/MR"] - (<"discard">),
  "M/E":  valid_ops_normal["M/E"],
  "M/M":  valid_ops_normal["M/M"],
  "M/MR": valid_ops_normal["M/MR"] - (<"discard">),
  "M/R":  valid_ops_normal["M/R"] - (<"discard">),
  "R/E":  valid_ops_normal["R/E"] - (<"metadata", "discard", "delete">),
  "R/R":  valid_ops_normal["R/R"],
]);

private static constant user_ops_undelete = ([
  "N/N":  (<"ascend">),
  "N/E":  (<"ascend", "metadata", "view", "log">),
  "N/R":  (<"ascend", "metadata", "view", "log", "undelete">),
  "E/E":  (<"ascend", "metadata", "view", "log">),
  "E/M":  (<"ascend", "metadata", "view", "log">),
  "E/MR": (<"ascend", "metadata", "view", "log", "undelete">),
  "A/N":  (<"ascend", "metadata", "view">),
  "A/M":  (<"ascend", "metadata", "view", "log">),
  "A/MR": (<"ascend", "metadata", "view", "log">),
  "M/E":  (<"ascend", "metadata", "view", "log">),
  "M/M":  (<"ascend", "metadata", "view", "log">),
  "M/MR": (<"ascend", "metadata", "view", "log">),
  "M/R":  (<"ascend", "metadata", "view", "log", "delete">),
  "R/E":  (<"ascend", "metadata", "view", "log", "undelete">),
  "R/R":  (<"ascend", "metadata", "view", "log", "undelete">),
]);

private static constant warning_ops = ([
  "N/N":  (<>),
  "N/E":  (<>),
  "N/R":  (<>),
  "E/E":  (<>),
  "E/M":  (<>),
  "E/MR": (<"upload">),
  "A/N":  (<"discard", "delete">),
  "A/M":  (<"discard", "delete", "revert">),
  "A/MR": (<"discard", "upload", "delete", "delete">),
  "M/E":  (<"discard", "delete", "revert">),
  "M/M":  (<"discard", "delete", "revert">),
  "M/MR": (<"discard", "upload", "delete", "revert">),
  "M/R":  (<"discard", "delete", "revert">),
  "R/E":  (<"discard", "revert">),
  "R/R":  (<>),
]);

private static constant op_perms = ([
  "ascend":		PERM_NONE, // Perm tested on parent.
  "metadata":		PERM_READ,
  "view":		PERM_READ,
  "log":		PERM_READ,
  "add":		PERM_WRITE,
  "download":		PERM_WRITE,
  "set_metadata":	PERM_WRITE,
  "upload":		PERM_WRITE,
  "discard":		PERM_WRITE,
  "update":		PERM_WRITE,
  "delete":		PERM_WRITE,
  "undelete":		PERM_WRITE,
  "revert":		PERM_WRITE,
  "commit":		PERM_WRITE,
  "copy":		PERM_READ,
  "move":		PERM_WRITE,
  "join":		PERM_WRITE,
  "purge":		PERM_WRITE,
]);

private static constant op_dsc = ([
  "metadata":		"view metadata",
  "view":		"view file",
  "log":		"show revision log",
  "add":		"create file",
  "download":		"download file",
  "set_metadata":	"change metadata",
  "upload":		"upload file",
  "discard":		"discard local file",
  "update":		"update local file",
  "delete":		"delete file",
  "undelete":		"undelete file",
  "revert":		"revert to old revision",
  "commit":		"commit local files",
  "copy":		"copy file",
  "move":		"move file",
  "join":		"join file between work areas",
  "purge":		"purge file",
]);

private static object/*(VCFile)*/ check_op (object id, string operation,
					    int undelete_mode, int superuser, mapping ops)
{
  if (operation == "ascend")
    return update_parent() && parent->exists (id, undelete_mode, superuser) &&
      this_object();
  if (!superuser && path[-1][..1] == "__" ||
      min (superuser ? PERM_WRITE : perm (id),
	   userid ? PERM_WRITE : PERM_READ) < op_perms[operation]) return 0;

  string stat = state || update_state();
  if (!md) return 0;

  switch (operation) {
    case "purge":
      return ops[stat]["purge"] && (superuser || wa->site->purge_access (id)) &&
	this_object();
    case "add_ppoint":
      return stat != "N/N" && (superuser || wa->site->ppoint_edit_access (id)) &&
	!has_ppoint() && this_object();
    case "remove_ppoint":
      return sizeof (path) && (superuser || wa->site->ppoint_edit_access (id)) &&
	has_ppoint() && this_object();
    case "commit":
      return !md->__invalid && ops[stat]["commit"] && this_object();
    case "add":
    case "upload":
    case "undelete":
    case "revert":
      if (!update_parent() || !parent->exists (id, 0, superuser)) return 0;
    default:
      return ops[stat][operation] && this_object();
  }
}

nomask string userstat_ea (object id, void|int undelete_mode, void|int superuser)
// Returns the edit area file status to show to the user; "none",
// "exists", "added", "modified" or "removed". Returns 0 if this file
// shouldn't be shown at all. In viewareas either "none" or 0 is
// always returned.
{
  if (!superuser && (path[-1][..1] == "__" || perm (id) == PERM_NONE)) return 0;
  string eastat = ustat_ea[state || update_state()];
  if (eastat == "add/mod") eastat = md->__added ? "added" : "modified";
  return (eastat != "none" ||
	  (undelete_mode ? ustat_rep_undelete : ustat_rep_normal)[state] != "none"
	 ) && eastat;
}

nomask string userstat_rep (object id, void|int undelete_mode, void|int superuser)
// Returns the repository file status to show to the user; "none",
// "exists", "modified" or "removed" (only in undelete mode). Returns
// 0 if this file shouldn't be shown at all.
{
  if (!superuser && (path[-1][..1] == "__" || perm (id) == PERM_NONE)) return 0;
  string repstat =
    (undelete_mode ? ustat_rep_undelete : ustat_rep_normal)[state || update_state()];
  if (repstat == "non/rem") repstat = md->__added ? "none" : "removed";
  return (ustat_ea[state] != "none" || repstat != "none") && repstat;
}

nomask object/*(VCFile)*/ is_valid_op (object id, string operation,
			   void|int undelete_mode, void|int superuser)
// Returns this object if the operation, which is a name of one of the
// file operation functions, is valid.
{
  return check_op (id, operation, undelete_mode, superuser,
		   undelete_mode ? valid_ops_undelete : valid_ops_normal);
}

nomask object/*(VCFile)*/ is_user_op (object id, string operation,
			  void|int undelete_mode, void|int superuser)
// Returns this object if the operation, which is a name of one of the
// file operation functions, should be available to the user.
{
  return check_op (id, operation, undelete_mode, superuser,
		   undelete_mode ? user_ops_undelete : user_ops_normal);
}

nomask mapping access_denied (object id, string operation)
// Returns an access denied error mapping using the correct ppoint.
{
  GET_POBJ (id);
  switch (op_perms[operation]) {
    case PERM_READ:
      return ERR_ACCESS_DENIED (
	pobj->ppid, "Read access required to " + op_dsc[operation] + ".");
    case PERM_WRITE:
      return ERR_ACCESS_DENIED (
	pobj->ppid, "Write access required to " + op_dsc[operation] + ".");
    default: throw (({"Bogus permission check\n", backtrace()}));
  }
}

nomask object/*(VCFile)*/ accessible (object id)
// Returns this object if the file is accessible for non-superusers.
{
  return path[-1][..1] != "__" && perm (id) != PERM_NONE && this_object();
}

nomask object/*(VCFile)*/ exists (object id, void|int undelete_mode, void|int superuser)
// Returns this object if the file exists.
{
  if (!superuser && (path[-1][..1] == "__" || perm (id) == PERM_NONE)) return 0;
  return (ustat_ea[state || update_state()] != "none" ||
	  (undelete_mode ? ustat_rep_undelete : ustat_rep_normal)[state] != "none") &&
    this_object();
}

nomask string abspath (object id, void|int undelete_mode, void|int superuser)
// Returns the absolute path without leading and trailing '/'.
{
  return exists (id, undelete_mode, superuser) && path * "/";
}

nomask string real_abspath() {return path * "/";}

nomask string name (object id, void|int undelete_mode, void|int superuser)
{
  return exists (id, undelete_mode, superuser) && sizeof (path) && path[-1];
}

nomask string real_name() {return sizeof (path) && path[-1];}

nomask array stat (object id, void|int undelete_mode, void|int superuser)
// Returns the file stat. For some fields, we don't even pretend that
// they have the proper meanings.
{
  int _perm = min (superuser ? PERM_WRITE : perm (id), userid ? PERM_WRITE : PERM_READ);
  if (!superuser && (path[-1][..1] == "__" || _perm == PERM_NONE) ||
      !(file || getfile()) ||
      ustat_ea[state || update_state()] == "none" &&
      (undelete_mode ? ustat_rep_undelete : ustat_rep_normal)[state] == "none")
    return 0;

  array(int|string) s = (<'N', 'R'>)[state[0]] && userid ?
    file->parent->stat() : file->stat();
  if (file->metadata()->__file_location)
  {
    array (int) s2;

    if (file->metadata()->__file_location_type == "anonymous")
      s2 = file_stat( wa->site->editor_vars->anonymous->dir
		      + file->metadata()->__file_location );
    else if (file->metadata()->__file_location_type == "named")
      s2 = file_stat( wa->site->editor_vars->named->dir
		      + file->metadata()->__file_location );

    if (s2)
      s = s2;
  }
  if (!s) return 0;
  s[0] = _perm == PERM_WRITE ? 0666 : 0444;
  s[5] = userstat_ea (id, undelete_mode, 1); // The user is the local status.
  s[6] = userstat_rep (id, undelete_mode, 1); // The group is the repository status.
  return s;
}


// File operations.

// Internals.

#define PERM_CHECK(_id, _op, _superuser) \
  if (!userid && op_perms[_op] > PERM_READ) \
    /* This is a viewarea; setting the superuser flag doesn't help. */ \
    return ERR_LOW_INVALID_OP ( \
      state, _op, "Write access required to " + op_dsc[_op] + "."); \
  if (!_superuser) { \
    GET_POBJ (_id); \
    if (pobj->perm < op_perms[_op]) return access_denied (_id, _op); \
  }

class VCOpPtr {
  object ptr;
  void create (object _ptr) {ptr = _ptr;}
}

/*private static*/ object(VCFileData)|object(VCFileConflict)|object(VCOpPtr)
  operation_in_progress;

#define BREAK_OP_IN_PROGRESS \
  if (operation_in_progress) { \
    if (!zero_type (operation_in_progress->fd)) { \
      operation_in_progress->fd->rollback(); \
    } \
    operation_in_progress = 0; \
  }

#define DO_IMPLICIT_TRANSITIONS \
  if (state == "R/R") \
    if (file->ununfake (1)) update_state(); \
    else throw (({"Couldn't remove local file\n", backtrace()}))

#define EDIT_INIT(_fake, _valid_ops, _state, _operation) \
  BREAK_OP_IN_PROGRESS; \
  if (!(file || getfile (_fake) || _fake)) return ERR_INVALID_FILE; \
  update_state(); \
  DO_IMPLICIT_TRANSITIONS; \
  OPMSG (_operation + " called in state " + _state); \
  if (!(_valid_ops)[_state][_operation]) \
    return ERR_INVALID_OPERATION (_state, _operation)

#define CHECK_WARNING(_warning, _state, _operation) \
  if ((_warning) && warning_ops[_state][_operation]) \
    return ERR_WARNING (_state, _operation)

#ifdef THREADS

/*private static*/ object(Thread.Mutex) edit_mutex = Thread.Mutex();

#define EDIT_LOCK object edit_lock; catch (edit_lock = edit_mutex->lock())
// A lock on all the destructive operations. Ensures consistent state
// in the edit area. Use a catch to ignore recursive locks.

#define WA_UPDATE_LOCK \
  object wa_update_lock; catch (wa_update_lock = wa->wa_update_mutex->lock())
// A lock on the whole work area to ensure consistent state in the
// repository. This can be avoided if cvs could bump the working
// revision to a specified one without setting a sticky tag.

#define EDIT_LOCK_2(_vcfile_1, _vcfile_2) \
  array(object) edit_locks = edit_lock_2 (_vcfile_1, _vcfile_2)
#define WA_UPDATE_LOCK_2(_wa_1, _wa_2) \
  array(object) wa_update_locks = wa_update_lock_2 (_wa_1, _wa_2)

#else  // THREADS

#define EDIT_LOCK
#define EDIT_LOCK_2
#define WA_UPDATE_LOCK
#define WA_UPDATE_LOCK_2

#endif

#define WASH_MD(_md) ((_md) - ([ \
    "__added": 1, \
    "__ancestor_removed": 1, \
    "__jointag": 1, \
    "__create_rev": 1, \
    "__join_revs": 1, \
    "__removed": 1, \
    "__file_location" : 1, \
    "__file_location_type" : 1, \
  ]))

private static int ensure_existing_unfaked()
// Ensure that the file is unfaked if it exists. Assumes EDIT_LOCK.
// Returns 0 on state failure.
{
  if (state[0] == 'N') {
    file->unfake();
    md = file->metadata();
    if (md->__removed) {
      if (!file->ununfake())
	throw (({"Couldn't refake recently unfaked file\n", backtrace()}));
      return 0;
    }
    update_state();
  }
  return 1;
}

/*private static*/ mapping ensure_unfaked()
// Ensure that the file is unfaked, creating it if necessary. Assumes
// WA_UPDATE_LOCK and EDIT_LOCK. Returns the new metadata if it was
// changed, but does not write it. For internal use only.
{
  mapping new_md = 0;
  if (state[0] == 'N') {
    if (state[-1] == 'N') {
      if (!file->new()) throw (({"Couldn't add file\n", backtrace()}));
      wa->_intern_vcobj (userid, path, this_object());
    }
    else {			// The file is removed.
      file->unfake();
      md = file->metadata();
      if (md->__removed)
	new_md = (["__added": 1,
		   "__ancestor_removed": 1,
		   "__create_rev": "next"]);
    }
    update_state();
  }
  return new_md;
}

private static object(VCFileConflict) make_vcfc (string b_rev, string old_rev)
{
  object(VCFileConflict) vcfc =
    VCFileConflict (this_object(), status->repository_version);

  int|string a_create_rev = -1, b_create_rev = -1;

  if (state != "N/N" && !md->__removed) {
    a_create_rev = md->__create_rev;
    vcfc->a_md = WASH_MD (md);
    vcfc->a_data = file->getfile();
#ifdef SBDEBUG
    if (!vcfc->a_data) throw (({"No data from getfile()\n", backtrace()}));
#endif
  }

  if (b_rev) {
    // versionedfile doesn't always work when there are modified local
    // files (e.g. in the A/M state), so try to use the parent
    // (view area) file if we have one.
    object(versionedfile) vfile =
      versionedfile (file->parent ? file->parent->file : file->file, b_rev);
    vcfc->b_md = vfile->metadata();
    if (vcfc->b_md->__removed) vcfc->b_md = 0;
    else {
      b_create_rev = vcfc->b_md->__create_rev;
      vcfc->b_md = WASH_MD (vcfc->b_md);
      vcfc->b_data = vfile->getfile();
#ifdef SBDEBUG
      if (!vcfc->b_md || !vcfc->b_data)
	throw (({"No data from versionedfile for b_rev\n", backtrace()}));
#endif
    }
  }

  if (old_rev) {		// Note: old_rev might not be in this work area.
    object(versionedfile) vfile =
      versionedfile (file->parent ? file->parent->file : file->file, old_rev);
    vcfc->old_md = vfile->metadata();
    if (vcfc->old_md->__removed ||
	a_create_rev != -1 && vcfc->old_md->__create_rev != a_create_rev ||
	b_create_rev != -1 && vcfc->old_md->__create_rev != b_create_rev)
      vcfc->old_md = 0;
    else {
      vcfc->old_rev = old_rev;
      vcfc->old_data = vfile->getfile();
      vcfc->old_md = WASH_MD (vcfc->old_md);
#ifdef SBDEBUG
      if (!vcfc->old_data)
	throw (({"No data from versionedfile for old_rev\n", backtrace()}));
#endif
    }
  }

  return vcfc;
}

/*private static*/ mapping|void make_path (object id, void|string comment,
					   void|int superuser)
// Tries to ensure that the path to this file exists. The given
// comment is used when committing directories. If no comment is
// given, only the checks are done to see if the path can be created.
// Internal use only; assumes that wa->wa_update_mutex is already
// locked when directories are created.
{
  if (!update_parent() || !parent->exists (0, 0, 1)) {
    mapping err;

    for (int i = 0; i < sizeof (path) - 1; i++) {
      if (!superuser && !valid_name (path[i])) { // Kludgy double work.
	err = ERR_INVALID_NAME (path[i]);
	err->make_path = path[..i] * "/";
	break;
      }

      object/*(VCDir)*/ dir =
	wa->_vcobj (userid, path[..i]) || Sitebuilder["VCDir"] (wa, path[..i], 0, userid);
      if (!dir->exists (0, 0, 1)) {
	// Kludgy double work here too.
	if (!superuser) {
	  GET_POBJ (id);
	  if (pobj->perm < PERM_WRITE) {
	    err = ERR_ACCESS_DENIED (
	      pobj->ppid, "Write access required to create directory.");
	    err->make_path = path[..i] * "/";
	    break;
	  }
	}

	if (comment) {
	  err = dir->add (id, comment, superuser);
	  if (err) {
	    err = copy_value (err);
	    err->make_path = path[..i] * "/";
	    break;
	  }
	}
      }
    }

    if (err) {
      err->vcobj = this_object();
      err->reason = "Failed on /" + err->make_path +
	" when trying to create path to /" + path * "/" +
	(err->reason ? ": " + err->reason : ".");
      return err;
    }
  }
}

mapping|void commit (object id, string comment, void|int superuser);

void low_join_commit (object/*(VCFile)*/ vcother, object(VCFileConflict) vcfc,
		      string comment)
// Used internally by VCFile.join_commit and VCDir.join_commit.
{
  mapping this_jr = get_join_revs (md), other_jr = get_join_revs (vcother->md);
  foreach (indices (this_jr + other_jr), string tag)
    if (!this_jr[tag]) this_jr[tag] = other_jr[tag];
    else if (!other_jr[tag]) other_jr[tag] = this_jr[tag];
    else this_jr[tag] = other_jr[tag] =
	   this_jr[tag][1] > other_jr[tag][1] ? this_jr[tag] : other_jr[tag];

  mapping err = vcfc->join_finish (vcfc->state->md, vcfc->state->data, comment);
  if (err) throw (({"Unexpected error from join_finish(): " +
		    format_error (err) + "\n", backtrace()}));

  if (!md->__removed) {
    m_delete (this_jr, wa->tag);
    this_jr[vcother->wa->tag] = ({next_revision (vcfc->other_rev), time()});
    md->__join_revs = this_jr;
    md->__jointag = vcother->wa->tag;
    file->write_metadata (encode_md (md));
  }

  if (!vcother->md->__removed) {
    m_delete (other_jr, vcother->wa->tag);
    other_jr[wa->tag] = ({next_revision (vcfc->this_rev), time (1)});
    vcother->md->__join_revs = other_jr;
    vcother->md->__jointag = wa->tag;
    vcother->getfile()->write_metadata (encode_md (vcother->md));
  }

  err = update_state()[0] != 'N' && commit (0, comment, 1);
  if (err) throw (({"Unexpected error from commit(): " +
		    format_error (err) + "\n", backtrace()}));

  err = vcother->update_state()[0] != 'N' && vcother->commit (0, comment, 1);
  if (err) throw (({"Unexpected error from vcother->commit(): " +
		    format_error (err) + "\n", backtrace()}));
}


// Protection point handling.

mapping|void add_ppoint (object id, void|int superuser)
{
  if (!superuser && !wa->site->ppoint_edit_access (id))
    return ERR_ACCESS_DENIED (
      wa->site->ppoint_edit_ppid(),
      "You don't have permission to add or remove protection points.");

  string p = path * "/";
  int ppid = wa->site->ac_check_file_pp (p);
  if (ppid) return ERR_LOW_INVALID_OP (
    state, "add_ppoint", "There already is a protection point on the file.");

  if (!wa->site->ac_add_file_pp (p))
    throw (({"Failed to add ppoint on file /" + p + "\n", backtrace()}));

  OPMSG ("add_ppoint");
}

mapping|void remove_ppoint (object id, void|int superuser)
{
  if (!superuser && !wa->site->ppoint_edit_access (id))
    return ERR_ACCESS_DENIED (
      wa->site->ppoint_edit_ppid(),
      "You don't have permission to add or remove protection points.");

  string p = path * "/";
  int ppid = wa->site->ac_check_file_pp (p);
  if (!ppid) return ERR_LOW_INVALID_OP (
    state, "remove_ppoint", "There is no protection point on the file.");

  if (!wa->site->ac_remove_file_pp (p))
    throw (({"Failed to remove ppoint from file /" + p + "\n", backtrace()}));

  OPMSG ("remove_ppoint");
}

// Nondestructive operations (do not lock the VCFile object).

mapping|object/*(VCDir)*/ ascend (object id, void|int undelete_mode,
				  void|int superuser)
// Returs the directory of the file. The permission and undelete_mode
// checks are done on the directory, not this file.
{
  if (!update_parent()) return ERR_INVALID_FILE;
  if (!parent->exists (id, undelete_mode, superuser))
    return ERR_INVALID_OPERATION (state || update_state(), "ascend");

  parent->reset_caches();
  return parent;
}

mapping|object(VCFileData) metadata (object id, void|int undelete_mode,
				     void|int superuser)
// Returns the metadata; the fd in the returned value is 0. This
// operation is not as lightweight as one might believe.
{
  PERM_CHECK (id, "metadata", superuser);
  if (!(file || getfile())) return ERR_INVALID_FILE;

  if (!(undelete_mode ? valid_ops_undelete : valid_ops_normal)
      [state || update_state()]["metadata"])
    return ERR_INVALID_OPERATION (state, "metadata");

  mapping(string:mixed) _md;
  if ((<'N', 'R'>)[state[0]] && userid)
    // The race here is harmless; the only async transition would
    // change the __removed flag, but that's ignored anyway.
    _md = file->parent->metadata();
  else _md = md;
  return VCFileData (WASH_MD (_md), 0);
}

mapping|object(VCFileData) view (object id, void|int undelete_mode,
				 void|int superuser)
// Returns the metadata and a file descriptor opened for reading after
// the metadata. This does not unfake the file; download() does.
{
  PERM_CHECK (id, "view", superuser);
  if (!(file || getfile())) return ERR_INVALID_FILE;

  if (!(undelete_mode ? valid_ops_undelete : valid_ops_normal)
      [state || update_state()]["view"])
    return ERR_INVALID_OPERATION (state, "view");

  array(object|mapping) data;
  if ((<'N', 'R'>)[state[0]] && userid)
    // The race here is harmless; the only async transition would
    // change the __removed flag, but that's ignored anyway.
    data = file->parent->openfile_with_metadata ("r");
  else
  {
    ensure_editarea_integrity( id );
    data = file->openfile_with_metadata ("r");
  }
  if (!data) throw (({"Couldn't open file for reading\n", backtrace()}));

  return VCFileData (WASH_MD (data[1]), data[0]);
}

mapping|array(object(VCLogEntry)) log (object id,
				       void|string revisions, void|string dates,
				       void|int undelete_mode, void|int superuser)
// revisions and dates are in the formats accepted by cvs log -r and
// -d, respectively. The branch tag for this work area is used if
// revisions is left out.
{
  PERM_CHECK (id, "log", superuser);
  if (!(file || getfile())) return ERR_INVALID_FILE;

  if (!(undelete_mode ? valid_ops_undelete : valid_ops_normal)
      [state || update_state()]["log"])
    return ERR_INVALID_OPERATION (state, "log");

  return convert_log (file->file->log (revisions || wa->tag, dates));
}

mapping|array(mapping|array(object(Annotation))) annotate(object id,
							  void|int undel_mode,
							  void|int superuser)
{
  mapping|array(object(VCLogEntry)) l =
    log(id, 0, 0, undel_mode, superuser);

  if (mappingp(l)) {
    return(l);
  }

  mapping(string:string) user_lookup = ([]);
  foreach(l, object(VCLogEntry) e) {
    user_lookup[e->revision] = e->user;
  }

  array(mapping|array(object(Annotation))) ann = file->annotate();
  ann[0] = WASH_MD (ann[0]);

  foreach(values(ann[0]) + ann[1], object(Annotation) o) {
    o->user = user_lookup[o->revision];
  }
  return(ann);
}

// Destructive operations (lock the VCFile object).

mapping|void add (object id, void|int superuser)
{
  WA_UPDATE_LOCK; // We get a race to the choice new() <-> unfake() otherwise.
  EDIT_LOCK;
  PERM_CHECK (id, "add", superuser);
  EDIT_INIT (1, valid_ops_normal, state, "add");
  if (!file) return ERR_INVALID_OPERATION (state, "add");
  if (!update_parent() || !parent->exists (id, 0, superuser)) return ERR_INVALID_FILE;
  if (!sizeof (path) || !superuser && !valid_name (path[-1]))
    return ERR_INVALID_NAME (sizeof (path) ? path[-1] : "");

  file->openfile_temp_w (encode_md (ensure_unfaked()))->commit();

  update_state();
  OPMSG ("add ok");
}

mapping|object(VCFileData) download (object id, void|int superuser)
// This unfakes the file; view() doesn't.
{
  EDIT_LOCK;
  PERM_CHECK (id, "download", superuser);
  EDIT_INIT (0, valid_ops_normal, state, "download");

  if (!ensure_existing_unfaked()) return ERR_INVALID_OPERATION (state, "download");

  ensure_editarea_integrity( id );

  array(object|mapping) data = file->openfile_with_metadata ("r");
  if (!data) throw (({"Couldn't open file for reading\n", backtrace()}));

  OPMSG ("download ok");
  return VCFileData (WASH_MD (data[1]), data[0]);
}

mapping|void set_metadata (object id, mapping metadata, void|int superuser)
{
  EDIT_LOCK;
  PERM_CHECK (id, "set_metadata", superuser);
  EDIT_INIT (0, valid_ops_normal, state, "set_metadata");

  if (!ensure_existing_unfaked()) return ERR_INVALID_OPERATION (state, "set_metadata");

  file->write_metadata (encode_md (set_md (md, metadata)));

  update_state();
  OPMSG ("set_metadata ok");
}

static class TempFile
{
  inherit Stdio.File;

  static int odate;
  static string path;
  static int active;

  int reopen (string flags)
  {
    werror( "[TempFile::reopen] " + path + "\n" );
    object file;
    
    if(!active)
      throw(({"tried to reopen file already comitted or rolled back.\n",
	      backtrace()}));
    catch (close());
    if (!(file = Stdio.File( path, "r" )))
      throw( ({"couldn't read " + path + ".\n", backtrace() }) );

    array (int) s = file_stat( path );

    if (!open( path + "~0~", flags, s[0] & 0777 ))
      throw( ({"couldn't create tempfile " + path + "~O~.\n", backtrace() }) );
    werror( "opened " + path + "~0~ flags: " + flags + "\n" );
    write( file->read( 0x7fffffff ) );
    seek( 0 );

    return 1;
  }

#define SCRAP(file) rm ((file) + "~"); mv ((file), (file) + "~")

  void commit()
  {
    werror( "[TempFile::commint] " + path + "\n" );
    if(!active)
      throw(({"tried to commit file already comitted or rolled back.\n",
	      backtrace()}));
    close();
    
    array(int) s = file_stat(path);
    if(s && s[3]>odate)
      throw(({"file "+ path + " changed under my feet.\n", backtrace()}));

    SCRAP (path);
    if(!mv(path + "~0~", path))
      throw(({"couldn't replace "+ path + " with tempfile!\n", backtrace()}));
    active = 0;
  }

  void rollback()
  {
    werror( "[TempFile::rollback] " + path + "\n" );
    if(active) {
      close();
      rm( path + "~0~" );
      active = 0;
    }
  }

  void destroy()
  {
    werror( "[TempFile::destroy] " + path + "\n" );
    if(active)
      rollback();
  }

  void create(string p)
  {
    ::create();
    path = p;
    odate = file_stat( p )[3];
    active = 1;
    werror( "[TempFile::create] " + path + "\n" );
  }

};

// Damn handler API!
static class FakeTempFile
{
  inherit Stdio.File;
  string path;

  int reopen (string flags)
  {
    return open( path, flags );
  }

  void commit()
  { }

  void rollback()
  { }

  void create( string p )
  {
    ::create();
    path = p;
  }
};

mapping|object(VCFileData) upload (object id, void|mapping metadata,
				   void|int warning, void|int superuser)
// Returns a file descriptor to write the data to. Call upload_end()
// with the VCFileData structure when the upload ends.
//
// To avoid potential deadlocks, the VCFile object isn't locked during
// the upload. If destructive operations are done then, they will run
// as if the upload hasn't happened, and the upload will fail
// (upload_end() will return an "interrupted" error). The upload file
// descriptor is closed when this happens.
{
  EDIT_LOCK;
  PERM_CHECK (id, "upload", superuser);
  EDIT_INIT (1, valid_ops_normal, state, "upload");
  if (!file) return ERR_INVALID_OPERATION (state, "upload");
  if (!update_parent() || !parent->exists (id, 0, superuser)) return ERR_INVALID_FILE;
  if (!sizeof (path) || !superuser && !valid_name (path[-1]))
    return ERR_INVALID_NAME (sizeof (path) ? path[-1] : "");
  CHECK_WARNING (warning, state, "upload");

  mapping new_md = 0;
  if (state[0] == 'N') {
#ifdef THREADS
    // Have to unlock and relock edit_lock after wa_update_lock.. P
    if (edit_lock) destruct (edit_lock);
    WA_UPDATE_LOCK; // We get a race to the choice new() <-> unfake() otherwise.
    catch (edit_lock = edit_mutex->lock());
    // Must redo parts of EDIT_INIT.
    update_state();
    DO_IMPLICIT_TRANSITIONS;
    if (!valid_ops_normal[state]["upload"])
      return ERR_INVALID_OPERATION (state, "upload");
#endif
    new_md = ensure_unfaked();
  }

  if (metadata) new_md = set_md (new_md || md, metadata);

  object fd;
  if (file->metadata()->__file_location)
    if (file->metadata()->__file_location_type == "anonymous"
	&& file_stat( wa->site->editor_vars->anonymous->dir
		      + file->metadata()->__file_location ))
      {
	fd = TempFile( wa->site->editor_vars->anonymous->dir
		       + file->metadata()->__file_location );
	fd->reopen( "cwt" );
      }
    else if (file->metadata()->__file_location_type == "named"
	     && file_stat( wa->site->editor_vars->named->dir
			   + file->metadata()->__file_location ) )
    {
      fd = TempFile( wa->site->editor_vars->named->dir
		     + file->metadata()->__file_location );
      fd->reopen( "cwt" );
    }
  if (!fd)
    fd = file->openfile_temp_w (encode_md (new_md));

  operation_in_progress = VCFileData ((["vcfile": this_object()]), fd);

  OPMSG ("upload ok");
  return operation_in_progress;
}

mapping|void upload_end (object(VCFileData) vcfd, int failed)
// See VCFileData.upload_end().
{
  EDIT_LOCK;
  OPMSG ("upload_end called in state " + state);

  if (operation_in_progress != vcfd) {
    ((object) vcfd->fd)->rollback();
    return ERR_INTERRUPTED;
  }
  operation_in_progress = 0;

  if (failed) ((object) vcfd->fd)->rollback();
  else ((object) vcfd->fd)->commit();

  update_state();
  OPMSG ("upload_end ok" + (failed ? " (failed)" : ""));
}

mapping|void discard (object id, void|int warning, void|int superuser)
{
  EDIT_LOCK;
  PERM_CHECK (id, "discard", superuser);
  EDIT_INIT (0, valid_ops_normal, state, "discard");
  CHECK_WARNING (warning, state, "discard");

  ensure_editarea_integrity( id, 1, 1 );
  if (!file->ununfake (1)) throw (({"Couldn't discard local file\n", backtrace()}));
  if (state[-1] == 'N') {
    wa->_unintern_vcobj (userid, path);
    if (update_parent()) parent->reset_contents();
  }

  update_state();
  OPMSG ("discard ok");
}

mapping|object(VCFileConflict) update (object id, void|int dont_read, void|int superuser)
// Returns an object for resolving conflicts between the local file
// and the current repository version, or 0 if there's no changes
// either locally or in the repository. Call update_finish() in that
// object to resolve the conflict. If dont_read is nonzero, no files
// are read (useable when you've already made up you're mind about
// which version should take effect).
//
// To avoid potential deadlocks, the VCFile object isn't locked during
// the update. If destructive operations are done then, they will run
// as normal, and the update will fail (update_finish() will return an
// "interrupted" error). However, if update() is called when there
// already is an update in progress and the repository revision hasn't
// changed, the same VCFileConflict object will be returned instead.
{
  EDIT_LOCK;
  PERM_CHECK (id, "update", superuser);
  if (!(file || getfile())) return ERR_INVALID_FILE;
  update_state();
  DO_IMPLICIT_TRANSITIONS;
  OPMSG ("update called in state " + state);
  if (!valid_ops_normal[state]["update"]) return ERR_INVALID_OPERATION (state, "update");

  ensure_editarea_integrity( id, 1 );
  if (operation_in_progress) {
    object op = operation_in_progress;
    if (!op->other_vcfile &&
	op->this_rev &&
	op->this_rev == status->repository_version &&
	(op->a_md || dont_read)) {
      OPMSG ("update ok (returned old)");
      return op;
    }
    else {
      if (!zero_type (op->fd)) {
	op->fd->rollback();
      }
      operation_in_progress = 0;
    }
  }

  if (state[0] == 'E') {
    if (!file->ununfake (1)) throw (({"Couldn't discard local file\n", backtrace()}));
    update_state();
    OPMSG ("update ok (no differences)");
    return 0;
  }

  object(VCFileConflict) vcfc;
  if (dont_read) vcfc = VCFileConflict (this_object(), status->repository_version);
  else {
    string b_rev, old_rev;

    switch (state) {
      case "M/M": case "M/MR":
	if (!md->__added) old_rev = status->working_version;
	if (state == "M/MR") break;

      case "A/M":
	b_rev = status->repository_version;
    }

    vcfc = make_vcfc (b_rev, old_rev);
  }

  OPMSG ("update ok");
  return operation_in_progress = vcfc;
}

mapping|void update_finish (
  object(VCFileConflict) vcfc, int override, void|mapping metadata, void|string data)
// See VCFileConflict.update_finish().
{
#ifdef SBDEBUG
  if (vcfc->other_vcfile) throw (({"Should call join_finish()?\n", backtrace()}));
#endif

  WA_UPDATE_LOCK;
  EDIT_LOCK;
  // All permission checks are done in update(), so we just ignore
  // those races and use the superuser flag.
  OPMSG ("update_finish called in state " + state);
  update_state();
  if (operation_in_progress != vcfc) return ERR_INTERRUPTED;
  operation_in_progress = 0;

  if (vcfc->this_rev != status->repository_version)
    return ERR_REPOSITORY_CHANGED;
#ifdef SBDEBUG
  if (!valid_ops_normal[state]["update"])
    throw (({"State shift in update_finish without error\n", backtrace()}));
#endif

  if (override) {
    string rv = status->repository_version;
    if (!file->update_soft()) throw (({"Couldn't do soft update\n", backtrace()}));
    update_state();
    if (rv != status->repository_version) // WA_UPDATE_LOCK should avoid this.
      throw (({"Repository revision increased during soft update\n", backtrace()}));

    if (file->parent_metadata()->__removed) md->__ancestor_removed = 1;
    if (metadata) md = set_md (md, metadata);

    // Import some internal metadata settings from the repository revision.
    mapping rep_md = versionedfile (
      file->parent ? file->parent->file : file->file, vcfc->this_rev)->metadata();
    if (rep_md->__create_rev) md->__create_rev = rep_md->__create_rev;
    else m_delete (md, "__create_rev");
    if (rep_md->__join_revs) md->__join_revs = rep_md->__join_revs;
    else m_delete (md, "__join_revs"); // Should not happen.

    if (data) {
      object tf = file->openfile_temp_w (encode_md (md));
      tf->write (data);
      tf->commit();
    }
    else file->write_metadata (encode_md (md));
  }
  else
    if (!file->ununfake (1)) throw (({"Couldn't discard local file\n", backtrace()}));

  update_state();
  OPMSG ("update_finish ok (" + (override ? "local" : "repository") + " override)");
}

mapping|void delete (object id, void|int warning, void|int superuser)
{
  EDIT_LOCK;
  PERM_CHECK (id, "delete", superuser);
  EDIT_INIT (0, valid_ops_normal, state, "delete");
  CHECK_WARNING (warning, state, "delete");

  ensure_editarea_integrity( id, 1, 1 );
  if ((<'N', 'R'>)[state[-1]]) {
    // Conceptually there's no race here; we're just a little late.
    if (!file->ununfake (1)) throw (({"Couldn't discard local file\n", backtrace()}));
    if (state[-1] == 'N') {
      wa->_unintern_vcobj (userid, path);
      if (update_parent()) parent->reset_contents();
    }
  }

  else switch (state[0]) {
    case 'R':
      return;

    case 'N':
      int begtime = time();
      file->fake = 1;
      file->unfake();
      if (time() == begtime) sleep (1);
      update_state();

    default:
      file->openfile_temp_w (encode_md ((["__removed": 1,
					  "__jointag": md->__jointag,
					  "__create_rev": md->__create_rev,
					  "__join_revs": md->__join_revs]))
			    )->commit();
  }

  update_state();
  OPMSG ("delete ok");
}

mapping|void undelete (object id, void|int superuser)
{
  WA_UPDATE_LOCK;
  EDIT_LOCK;
  PERM_CHECK (id, "undelete", superuser);

  EDIT_INIT (0, valid_ops_undelete, state, "undelete");
  if (!update_parent() || !parent->exists (id, 0, superuser)) return ERR_INVALID_FILE;

  string rv = status->repository_version;
  switch (state) {
    case "N/R":
      file->unfake();
      break;

    case "E/MR":
      if (file->update_hard() > 1) throw (({"Couldn't do hard update\n", backtrace()}));
      break;

    case "R/E":
      if (!file->ununfake (1)) throw (({"Couldn't discard local file\n", backtrace()}));
      update_state();
      return;
  }

  update_state();
  if (rv != status->repository_version)	// WA_UPDATE_LOCK should avoid this.
    throw (({"Repository revision increased during undelete\n", backtrace()}));
  file->write_metadata (
    encode_md (md - (["__removed": 1]) + (["__ancestor_removed": 1])));

  update_state();
  OPMSG ("undelete ok");
}

mapping|void revert (object id, int warning, int superuser, object/*(VCOldFile)*/ vcofile)
// See VCOldFile.revert().
{
  EDIT_LOCK;
  PERM_CHECK (id, "revert", superuser);
  EDIT_INIT (0, valid_ops_normal, state, "revert");
  if (!update_parent() || !parent->exists (id, 0, superuser)) return ERR_INVALID_FILE;
  CHECK_WARNING (warning, state, "revert");

  ensure_editarea_integrity( id, 1, 1 );

  mapping|object(VCFileData) vcfd = vcofile->view (0, 0, 1);
  if (mappingp (vcfd)) return ERR_REMOVED_REVISION;

  if (state[0] != 'N') {
    if (!file->ununfake (1)) throw (({"Couldn't discard local file\n", backtrace()}));
    update_state();
  }

  // This is a race, but there's not much to do about it on this
  // level; there'll be one anyway from the info in the UI.
  file->unfake();
  if (md->__removed) vcfd->md += (["__ancestor_removed": 1]);

  object/*(TempFile)*/ tf = file->openfile_temp_w (encode_md (vcfd->md));
  tf->write (vcfd->fd->read());	// vcfd has the whole file in memory anyway.
  tf->commit();

  update_state();
  OPMSG ("revert ok");
}

mapping|void commit (object id, string comment, void|int superuser)
// Besides the normal criteria, this operation is invalid whenever the
// metadata __invalid exists.
{
  WA_UPDATE_LOCK;
  EDIT_LOCK;
  PERM_CHECK (id, "commit", superuser);

  EDIT_INIT (0, valid_ops_normal, state, "commit");
  if (!superuser && md->__invalid)
    return ERR_LOW_INVALID_OP (state, "commit",
			       "The file has invalid contents.");

  ensure_editarea_integrity( id, 1 );

  string orig_state = state;
  int added = md->__added, ancestor_removed = md->__ancestor_removed;
  string jointag = md->__jointag, joinrev = jointag && get_join_revs (md)[jointag][0];

  string rv = status->repository_version;
  switch (state) {
    case "A/MR":
    case "M/R":			// Call the lie that we maybe did in update_state().
      if (!file->update_soft()) throw (({"Couldn't do soft update\n", backtrace()}));
      update_state();
      break;

    case "R/E":
      mixed join_revs = md->__join_revs;
      if (file->update_hard() > 1) throw (({"Couldn't do hard update\n", backtrace()}));
      md = file->metadata() + (["__removed": 1]);
      if (join_revs) md->__join_revs = join_revs;
      file->write_metadata (encode_md (md));
      break;
   }
  if (rv != status->repository_version)	// WA_UPDATE_LOCK should avoid this.
    throw (({"Repository revision increased during commit\n", backtrace()}));

  int md_changed = md->__added || md->__ancestor_removed || md->__jointag;
  if (md->__create_rev == "next") {
    md->__create_rev = next_revision (file->status()->working_version);
    md_changed = 1;
  }
  if (md_changed)
    file->write_metadata (
      encode_md (md - (["__added": 1,
			"__ancestor_removed": 1,
			"__jointag": 1,
			"__file_location": 1,
			"__file_location_type": 1,
		      ])));

  string action;
  if (orig_state == "R/E") action = "delete";
  else if (added || (<"A/N", "A/MR">)[orig_state]) action = "create";
  else if (ancestor_removed) action = "undelete";
  else action = "edit";
  comment = sprintf ("%s by %d, %s, %s%s\n%s",
		     action,
		     userid,
		     wa->mac->id_get_handle (userid) || "",
		     wa->mac->id_get_name (userid) || "",
		     jointag ? "; join " + jointag + ", " + joinrev : "",
		     comment);

  if (!file->commit (comment))
    // WA_UPDATE_LOCK should avoid this.
    throw (({"Commit failed\n", backtrace()}));

  if (wa->site->commit_hook && functionp(wa->site->commit_hook))
  { // Hook for additional commit logging etc.
    wa->site->commit_hook("commit",       // commit type
                          real_abspath(), // absolute path to commited file
                          id,             // request id
                          comment,        // commit log message
                          this_object()); // vcfile object
  }

  update_state();
  OPMSG ("commit ok");
}

mapping|void copy (object id, object/*(VCFile)*/ to, void|int superuser,
		   void|object/*(VCOldFile)*/ vcofile)
// Returns an error of type "removed revision" if a vcvfile is
// specified that is removed.
{
  EDIT_LOCK_2 (this_object(), to);
  PERM_CHECK (id, "copy", superuser);
  EDIT_INIT (0, valid_ops_normal, state, "copy");

  ensure_editarea_integrity( id );

  if (vcofile)
    throw (({"Not implemented\n", backtrace()}));

  mapping|object(VCFileData) src = view( id, superuser );

  if (mappingp(src))
    return src;

  mapping|object(VCFileData) dest = to->upload( id, src->md, 0, superuser);

  if (mappingp(dest))
    return dest;

  string s;
  while ((s = src->fd->read( 16384 )) && s!="")
    if(dest->fd->write( s ) != sizeof(s)) {
      int err = dest->fd->errno();
      dest->upload_end(1);
      OPMSG (strerror(err));
      return ERR_IO_ERROR (err);
    }

  dest->upload_end(0);

  // FIXME: Add metadata to get copy info in commit messages.

  OPMSG ("copy to " + to->abspath(id, 0, 1) + " ok");
}

mapping|void move (object id, object/*(VCFile)*/ to, void|int superuser)
{
  EDIT_LOCK_2 (this_object(), to);
  PERM_CHECK (id, "move", superuser);
  EDIT_INIT (0, valid_ops_normal, state, "move");

  ensure_editarea_integrity( id, 1 );

  mapping|void err = copy(id, to, superuser);

  if(mappingp(err))
    return err;

  err = delete(id, 0, superuser);

  if(mappingp(err))
    return err;

  // FIXME: Add metadata to get move info in commit messages.

  OPMSG ("move to " + to->abspath(id, 0, 1) + " ok");
}

mapping|object(VCFileConflict) join (object id, object/*(Workarea)*/ other,
				     void|int dont_read, void|int superuser)
// Works like update(), but does conflict resolution against this file
// in another work area.
//
// Note: Workarea.access_try() should be called on both work areas
// before this function to get the appropriately lowered permission.
{
  object/*(VCFile)*/ vcother = other->vcfile (path * "/", id);
#ifdef SBDEBUG
  if (!vcother || !vcother->isfile)
    throw (({"No VCFile object in the other work area for /" + path * "/" + "\n",
	     backtrace()}));
#endif

  EDIT_LOCK_2 (this_object(), vcother);
  PERM_CHECK (id, "join", superuser);

  {
    // Relevant parts of EDIT_INIT for both files.
    if (!(file || getfile (1)) || !vcother->getfile (1)) return ERR_INVALID_FILE;
    update_state();
    DO_IMPLICIT_TRANSITIONS;
    if (!valid_ops_normal[state]["join"])
      return ERR_LOW_INVALID_OP (state, "join",
				 "There's a local copy in one or both work areas; "
				 "use discard or update/commit first.");

    string otherstate = vcother->update_state();
    // DO_IMPLICIT_TRANSITIONS for vcother.
    if (otherstate == "R/R") {
      if (vcother->getfile()->ununfake (1)) otherstate = update_state();
      else throw (({"Couldn't remove local file\n", backtrace()}));
    }
    if (!valid_ops_normal[otherstate]["join"])
      return ERR_LOW_INVALID_OP (state, "join",
				 "There's a local copy in one or both work areas; "
				 "use discard or update/commit first.");
    if (state[-1] == 'N' && otherstate[-1] == 'N')
      return ERR_LOW_INVALID_OP (state, "join",
				 "No file with that name exists in either work area.");

    mapping err = make_path (id);
    if (err) return err;
    err = vcother->make_path (id);
    if (err) return err;
  }

  OPMSG ("join called in state " + state);

  string this_rev = status->repository_version;
  string other_rev = vcother->getfile(1)->status()->repository_version;

  if (operation_in_progress) {
    object op = operation_in_progress;
    if (op->other_vcfile == vcother &&
	op->other_vcfile->operation_in_progress &&
	op->other_vcfile->operation_in_progress->ptr == operation_in_progress &&
	op->this_rev == this_rev && op->other_rev == other_rev &&
	(op->a_md || op->b_md || dont_read)) {
      OPMSG ("join ok (returned old)");
      return op;
    }
    else {
      if (!zero_type (op->fd)) {
	op->fd->rollback();
      }
      operation_in_progress = 0;
    }
  }
  if (vcother->operation_in_progress) {
    if (!zero_type (vcother->operation_in_progress->fd)) {
      vcother->operation_in_progress->fd->rollback();
    }
    vcother->operation_in_progress = 0;
  }

  string old_rev = 0;
  if (this_rev && other_rev) {
    array o_rev = get_join_revs (vcother->md)[wa->tag];
    array t_rev = get_join_revs (md)[vcother->wa->tag];
    if (o_rev && t_rev)
      if ((o_rev[0] == "new" ? (this_rev / ".")[-1] == "1" : o_rev[0] == this_rev) &&
	  (t_rev[0] == "new" ? (other_rev / ".")[-1] == "1" : t_rev[0] == other_rev)) {
	OPMSG ("join with '" + other->tag + "' ok (no changes)");
	return 0;
      }
      else old_rev = o_rev[1] > t_rev[1] ? o_rev[0] : t_rev[0];
    else if (o_rev) old_rev = o_rev[0];
    else if (t_rev) old_rev = t_rev[0];
    if ((<0, "new">)[old_rev])
      if (this_rev == other_rev) {
	OPMSG ("join with '" + other->tag + "' ok (no changes)");
	return 0;
      }
      else
	// make_vcfc() will check that neither file have been deleted since old_rev.
	old_rev = file->split_version (this_rev, other_rev);
  }

  object(VCFileConflict) vcfc =
    dont_read ? VCFileConflict (this_object(), this_rev) : make_vcfc (other_rev, old_rev);
  vcfc->other_vcfile = vcother;
  vcfc->other_rev = other_rev;

  vcother->operation_in_progress = VCOpPtr (vcfc);
  OPMSG ("join with '" + other->tag + "' ok");
  return operation_in_progress = vcfc;
}

mapping|void join_finish (object(VCFileConflict) vcfc, mapping metadata, string data,
			  void|string comment)
// See VCFileConflict.join_finish().
{
  object/*(VCFile)*/ vcother = vcfc->other_vcfile;
#ifdef SBDEBUG
  if (!vcother) throw (({"Should call update_finish()?\n", backtrace()}));
#endif

  WA_UPDATE_LOCK_2 (wa, vcother->wa);
  EDIT_LOCK_2 (this_object(), vcother);
  // All permission checks are done in join(), so we just ignore those
  // races and use the superuser flag.

  update_state();
  vcother->update_state();

  OPMSG ("join_finish called in state " + state);

  if (operation_in_progress != vcfc ||
      !vcother->operation_in_progress ||
      vcother->operation_in_progress->ptr != vcfc)
    return ERR_INTERRUPTED;
  operation_in_progress = vcother->operation_in_progress = 0;

  object(cachefile) otherfile = vcother->getfile (1);
  if (vcfc->this_rev != status->repository_version ||
      vcfc->other_rev != otherfile->status()->repository_version)
    return ERR_REPOSITORY_CHANGED;
#ifdef SBDEBUG
  if (!valid_ops_normal[state]["join"] || !vcother->is_valid_op (0, "join", 0, 1))
    throw (({"State shift in join_finish without error\n", backtrace()}));
#endif

  if (metadata) {
    mapping res = make_path (
      0, comment || "Automatically created by join operation.", 1);
    if (mappingp (res)) throw (({"Unexpected error from make_path(): " +
				 format_error (res) + "\n", backtrace()}));
    res = vcother->make_path (
      0, comment || "Automatically created by join operation.", 1);
    if (mappingp (res)) throw (({"Unexpected error from vcother->make_path(): " +
				 format_error (res) + "\n", backtrace()}));

    object fd = file->openfile_temp_w (
      encode_md (set_md (ensure_unfaked() || ([]), metadata)));
    fd->write (data);		// FIXME: Handle write error.
    fd->commit();
    update_state();

    fd = otherfile->openfile_temp_w (
      encode_md (set_md (vcother->ensure_unfaked() || ([]), metadata)));
    fd->write (data);		// FIXME: Handle write error.
    fd->commit();
    vcother->update_state();
  }

  else {
    if (vcfc->this_rev && !md->__removed) {
      mapping res = delete (0, 0, 1);
      if (mappingp (res)) throw (({"Unexpected error from delete(): " +
				   format_error (res) + "\n", backtrace()}));
    }
    if (vcfc->other_rev && !vcother->md->__removed) {
      mapping res = vcother->delete (0, 0, 1);
      if (mappingp (res)) throw (({"Unexpected error from vcother->delete(): " +
				   format_error (res) + "\n", backtrace()}));
    }
  }

  OPMSG ("join_finish with '" + vcother->wa->tag + "' ok");
}

mapping|void join_commit (object id, object/*(Workarea)*/ other, string comment,
			  void|int superuser)
// Finishes and commits a join on the file as one atomic operation. In
// the (old) VCFileConflict object, state->md is used for the metadata
// and state->data for the contents. Note: Uses "join" as op string.
{
  object/*(VCFile)*/ vcother = other->vcfile (path * "/", id);
#ifdef SBDEBUG
  if (!vcother || !vcother->isfile)
    throw (({"No VCFile object in the other work area for /" + path * "/" + "\n",
	     backtrace()}));
#endif

  WA_UPDATE_LOCK_2 (wa, other);
  EDIT_LOCK_2 (this_object(), vcother);
  OPMSG ("join_commit called");

  mapping|object(VCFileConflict) vcfc = join (id, other, superuser);
  if (mappingp (vcfc)) return vcfc;
  if (vcfc->state->md && vcfc->state->md->__invalid)
    return ERR_LOW_INVALID_OP (state, "join", "The file has invalid contents.");

  if (zero_type (vcfc->state->md)) {
    OPMSG ("join_commit with '" + other->tag + "' ok (nothing to do)");
    operation_in_progress = vcother->operation_in_progress = 0;
    return;
  }

  low_join_commit (vcother, vcfc, comment);

  OPMSG ("join_commit with '" + other->tag + "' ok");
}

mapping|void purge (object id, void|int superuser)
{
  EDIT_LOCK;
  PERM_CHECK (id, "purge", superuser);
  if (!superuser && !wa->site->purge_access (id)) return ERR_ACCESS_DENIED (
    wa->site->purge_ppid(), "You don't have permission to purge files.");

  // Almost but not quite EDIT_INIT.
  BREAK_OP_IN_PROGRESS;
  if (!(file || getfile())) return ERR_INVALID_FILE;
  update_state();
  OPMSG ("purge called in state " + state);
  DO_IMPLICIT_TRANSITIONS;

  throw (({"Not implemented\n", backtrace()}));

  update_state();
  OPMSG ("purge ok");
}

void ensure_editarea_integrity( object id, void|int deletep,
				void|int only_deletep )
{
  // A check of the dates of the files here would make sense
  // But can we be sure that nothing else will touch the file in
  // the editarea?

  string path;
  
  if (file->metadata()->__file_location)
    if (file->metadata()->__file_location_type == "anonymous")
      path = wa->site->editor_vars->anonymous->dir
	+ file->metadata()->__file_location;
    else if (file->metadata()->__file_location_type == "named")
      path = wa->site->editor_vars->named->dir
	+ file->metadata()->__file_location;

  if (path && file_stat( path ))
  {
    if (!only_deletep)
    {
      object obj = FakeTempFile( path );
      object handler = wa->site->handler_for( this_object(), id );

      handler->upload( this_object(), id,
		       VCFileData( WASH_MD( file->metadata() ), obj ) );
      file->write( Stdio.File( path, "r" )
		   ->read( 0x7fffffff ) );
    }
    if (deletep)
    {
      rm( path );
      if (file->metadata()->__file_location_type == "anonymous")
      {
	array arr = path / "/";

	rm( arr[..sizeof( arr )-2] * "/" );
      }
      else if (file->metadata()->__file_location_type == "named")
      {
	array arr = file->metadata()->__file_location / "/";

	for (int c=sizeof( arr )-2;
	     sizeof( get_dir( wa->site->editor_vars->named->dir
			      + arr[..c] * "/" ) ) == 0
	       && c >= 0; c--)
	  rm( wa->site->editor_vars->named->dir + arr[..c] * "/" );
      }
    }
  }
}

string get_downloaded_contents( object id )
{
  object handler = wa->site->handler_for( this_object(), id );
  mixed res;

  array(object|mapping) data = file->openfile_with_metadata ("r");
  if (!data) throw (({"Couldn't open file for reading\n", backtrace()}));

  res = handler->download( data[0], ([ ]), id, WASH_MD( data[1] ),
			   this_object() );
  if (stringp( res ))
    return res;
  else if (objectp( res ))
    return res->read( 0x7fffffff );
  else
    throw( ({ "Unknown return type from download method in handler\n",
	      backtrace() }) );
}

string get_named_editarea( object id, int uid )
{
  string path, dir, filename;
  array arr;
  object obj;

  if (!wa->site->editor_vars->named->enabled)
    throw( ({ "Named access not enabled.", backtrace() }) );
  if (file->metadata()->__file_location)
    if (file->metadata()->__file_location_type == "named")
      if (file_stat( wa->site->editor_vars->named->dir
		     + file->metadata()->__file_location ))
	return file->metadata()->__file_location;
      else ;
    else
      ensure_editarea_integrity( id, 1 );
  path = wa->site->editor_vars->named->dir;
  dir = "";
  arr = ({ (string )uid }) + real_abspath() / "/";
  arr = arr[..sizeof( arr )-2];
  foreach (arr, string d)
  {
    dir += d + "/";
    if (!file_stat( path + dir ))
    {
      mkdir( path + dir, 0700 );
      // FIX ME Privs?
#ifndef __NT__
      chown( path + dir, uid, -1 );
#endif      
    }
  }
  obj = Stdio.File( path + dir + real_name(), "ctw", 0600 );
  obj->write( get_downloaded_contents( id ) );
  obj->close();
  // FIX ME Privs?
#ifndef __NT__
  chown( path + dir + real_name(), uid, -1 );
#endif
  file->write_metadata( file->metadata()
			+ ([ "__file_location" : dir + real_name(),
			     "__file_location_type" : "named" ]) );
  return dir + real_name();
}

string get_anonymous_editarea( object id )
{
  string path, dir, filename;
  object obj;

  if (!wa->site->editor_vars->anonymous->enabled)
    throw( ({ "Anonymous access not enabled.", backtrace() }) );
  if (file->metadata()->__file_location)
    if (file->metadata()->__file_location_type == "anonymous")
      if (file_stat( wa->site->editor_vars->anonymous->dir
		     + file->metadata()->__file_location ))
	return file->metadata()->__file_location;
      else ;
    else
      ensure_editarea_integrity( id, 1 );
  path = wa->site->editor_vars->anonymous->dir;
  do
  {
    dir = MIME.encode_base64( Crypto.randomness.reasonably_random()
				   ->read( 32 ) ) - "/" - "+" - "=";
  } while (file_stat( path + dir ));
  mkdir( path + dir, 0711 );
  obj = Stdio.File( path + dir + "/" + real_name(), "ctw", 0666 );
  obj->write( get_downloaded_contents( id ) );
  obj->close();
  chmod( path + dir + "/" + real_name(), 0666 );
  file->write_metadata( file->metadata()
			+ ([ "__file_location" : dir + "/" + real_name(),
			     "__file_location_type" : "anonymous" ]) );
  return dir + "/" + real_name();
}

