// Object returned by VCFile.update() to resolve conflicts between a
// local file (a) and a specific repository version (b).

// Also returned by VCFile.join() to resolve conflicts between the
// same file in two work areas. In that case, the b version is the file
// in the other work area.

import Sitebuilder;

object/*(VCFile)*/ this_vcfile;
object/*(VCFile)*/ other_vcfile; // Only used with join().
string this_rev;
string other_rev;		// Only used with join().
string old_rev;

void create (object/*(VCFile)*/ t_vc, string t_rev,
	     void|object/*(VCFile)*/ o_vc, void|string o_rev)
{
  this_vcfile = t_vc, this_rev = t_rev;
  other_vcfile = o_vc, other_rev = o_rev;
}

mapping(string:mixed) a_md, b_md, old_md;
string a_data, b_data, old_data;
// For update(): The metadata and file contents of the local (a),
// repository (b) and common ancestor (old) versions. The repository
// data is zero iff the repository version has been deleted since the
// common ancestor. The common ancestor data is zero iff there is no
// common ancestor version.
//
// For join(): The metadata and file contents of the file in this
// work area (a), the other (b) and the common ancestor (old). Either a
// or b may be zero if there's no such file in that work area. The
// common ancestor is zero iff there is no common ancestor version.

mapping(string:mixed) state;
// Use this to store data about the current state of the conflict
// resolution.

mapping|void update_finish (int override, void|mapping metadata, void|string data)
// Resolves the conflict from update(). The local version is discarded
// if override is zero, otherwise the repository version is overridden
// with the specified metadata and data. The current local version is
// used if either of these are zero. Returned error types:
//
// "interrupted":	Some other destructive operation was done on the
//			VCFile object during the update.
// "repository changed": The repository version has changed.
{
  return this_vcfile->update_finish (this_object(), override, metadata, data);
}

mapping|void join_finish (mapping metadata, string data, void|string comment)
// Resolves the conflict from join(). Overrides the files in both
// work areas with the specified metadata and data. Both files are
// marked for deletion if metadata is zero. If comment is given, it's
// used as commit message if directories need to be created, otherwise
// a default message is used. Returned error types:
//
// "interrupted":	Some other destructive operation was done on
//			some VCFile object during the join.
// "repository changed": Some repository version has changed.
//
// Note: This doesn't set the join info in the metadata since no
// commit is done.
{
  return this_vcfile->join_finish (this_object(), metadata, data, comment);
}


// Diff-based helper.

array(string|array(string)) diff;
// Each element is either a similarity string or a difference array in
// which each element is either a string or 0 (when there's no
// conflicting text. Each difference array has the local version
// first, followed by the repository version, and then the old
// version, if any.

int create_diff (function(string:array(string)) splitfn,
		 function(array(string):string) joinfn)
// Initializes diff. splitfn splits the data into the chunks used for
// diffing. joinfn does the reverse. Returns 0 if there's no
// differences (after splitfn), 1 if there are conflict free
// differences and 2 if there are conflicts.
{
  if (!a_data || !b_data) throw (({"Can't diff without two files\n", backtrace()}));

  array(array(string)) a, b, old;

  if (old_data) {
    array data = diff3 (splitfn (a_data), splitfn (b_data), splitfn (old_data));
    a = data[0], b = data[1], old = data[2];

    // Merge differences where a and b agrees against old with
    // similarities.
    for (int i = 1; i < sizeof (a);)
      if (a[i] == b[i] && a[i - 1] == b[i - 1]) {
	array(string) sim = a[i - 1] + a[i];
	if (b[i] == old[i] && b[i - 1] == old[i - 1])
	  a[i - 1] = b[i - 1] = old[i - 1] = sim;
	else
	  a[i - 1] = b[i - 1] = sim, old[i - 1] += old[i];
	a = a[..i - 1] + a[i + 1..];
	b = b[..i - 1] + b[i + 1..];
	old = old[..i - 1] + old[i + 1..];
      }
      else i++;
  }
  else {
    array data = Array.diff (splitfn (a_data), splitfn (b_data));
    a = data[0], b = data[1];
  }

  int res = 0;
  diff = allocate (sizeof (a));
  for (int i = 0; i < sizeof (diff); i++)
    if (a[i] == b[i])
      if (!sizeof (a[i])) throw (({"Strange data from diff or diff3\n", backtrace()}));
      else diff[i] = joinfn (a[i]);
    else {
      array block = allocate (old ? 3 : 2);
      if (sizeof (a[i])) block[0] = joinfn (a[i]);
      if (sizeof (b[i])) block[1] = joinfn (b[i]);
      if (old)
	if (old[i] == a[i]) block[2] = block[0];
	else if (old[i] == b[i]) block[2] = block[1];
	else {
	  block[2] = joinfn (old[i]);
	  res = 2;
	}
      else if (block[0] && block[1]) res = 2;
      if (res < 1) res = 1;
      diff[i] = block;
    }

  return res;
}
