// Handles one work area.

// $Id: Workarea.pike,v 1.34 1999/04/24 17:40:42 mast Exp $

// AC checking is done for file access, but the user has to handle the
// work area global permission with access_try() and access_throw().

#if constant (roxen)
#include <config.h>
#endif

import Sitebuilder;
import VC.CVS;


// Global data.

constant default_dir = combine_path (__FILE__, "../../../default/");

object/*(Site)*/ site;
object/*(ModuleAC)*/ mac;	// Shortcut to site->mac.

string name, tag;		// Authorative; settings is not.
string tagdir;			// Shortcut. Ends with '/'.

mapping(string:mixed) settings = ([]);
// Work area global variables read from and written to the sbconfig file.

// Internals.

#ifdef THREADS

private static object(Thread.Mutex) wa_internal_mutex = Thread.Mutex();
// Locks the internal data below.

object(Thread.Mutex) wa_update_mutex = Thread.Mutex();
// Used by VCFile/VCDir for work area global locks.

#define WA_INTERNAL_LOCK \
  object wa_internal_lock; catch (wa_internal_lock = wa_internal_mutex->lock())

#else  // THREADS

#define WA_INTERNAL_LOCK

#endif

private static mapping(string:object(VCFile)|object(VCDir)) vcobjs = ([]);
// Indexed by <AC user ID> + ":" + <path without leading and trailing '/'>.

private static object(cachedir) viewtopdir;

private static mapping(int:object(cachedir)) edittopdirs = ([]);
// Indexed on AC ID numbers.

object(cachedir) topdir (void|int userid)
{
  if (!userid) return viewtopdir;

  object(cachedir) topdir = edittopdirs[userid];
  if (topdir) return topdir;

  WA_INTERNAL_LOCK;
  topdir = edittopdirs[userid];
  if (topdir) return topdir;

  string editdir = tagdir + userid;
  if (!file_stat (editdir)) {
#ifdef SBDEBUG
    report_notice ("SB: Checking out edit area " + name + " for ID " + userid + "\n");
#endif
    if (!mkdirhier (editdir + "/"))
      throw (({"Couldn't create directory " + editdir + "\n", backtrace()}));
    topdir = viewtopdir->clone (editdir);
  }
  else topdir = cachedir (editdir, viewtopdir, 0, 0,
			  site->storage + "cvsroot");

  if (!topdir) throw (({"Failed to get VC.CVS.cachedir for " + tagdir + userid + "\n"}));

  return edittopdirs[userid] = topdir;
}

void _intern_vcobj (int userid, array(string) path, object(VCFile)|object(VCDir) vcobj)
{
#ifdef SBDEBUG
  if (!vcobj->isfile && !vcobj->isdir || vcobj->isoldfile || vcobj->isolddir)
    throw (({"Try to put unknown object in VCFile/VCDir cache\n", backtrace()}));
#endif
  string vcobjs_key = userid + ":" + path * "/";
  WA_INTERNAL_LOCK;
  if (vcobjs[vcobjs_key] && vcobjs[vcobjs_key] != vcobj)
    throw (({"A different VCFile|VCDir object is already in vcobjs\n", backtrace()}));
  vcobjs[vcobjs_key] = vcobj;
}

void _unintern_vcobj (int userid, array(string) path)
{
  string vcobjs_key = userid + ":" + path * "/";
  WA_INTERNAL_LOCK;
  m_delete (vcobjs, vcobjs_key);
}

object(VCFile)|object(VCDir) _vcobj (int userid, array(string) path)
{
  string vcobjs_key = userid + ":" + path * "/";
  object(VCFile)|object(VCDir) vcobj = vcobjs[vcobjs_key];

  if (!vcobj) {
    WA_INTERNAL_LOCK;
    vcobj = vcobjs[vcobjs_key];
    if (!vcobj) {
      object(cachedir) topdir = userid ? topdir (userid) : viewtopdir;
      object(cachedir) dir = topdir->find_dir (path);
      if (dir) vcobj = VCDir (this_object(), path, dir, userid);
      else if (path[-1] != "__info") {
	dir = topdir->find_dir (path[..sizeof (path) - 2]);
	object(cachefile) file = dir && dir->find_file (path[-1]);
	if (file) vcobj = VCFile (this_object(), path, file, userid, dir);
      }
      if (vcobj) vcobjs[vcobjs_key] = vcobj;
    }
  }

  return vcobj;
}

object(VCOldFile)|object(VCOldDir) _vcoldobj (object vcobj, string revision, string date)
{
  // FIXME: Cache?
  return (vcobj->isfile ? VCOldFile : VCOldDir) (this_object(), vcobj, revision, date);
}


// AC.

private static string pp_workarea;

private static void ac_init()
{
  pp_workarea = site->pp_workarea (tag);
}

function(string|array(string):int) ac_find_file_pp;
function(string|array(string),object,void|int:int) fileperm;
// Shortcuts to functions in site.

int access_try (object id)
// Resolves the AC identity of the request. Returns the work area
// permission. It should be used iff the files in a specific work area
// are going to be accessed.
{
  // Note: This is chaining from PP_CONTENT_EDITOR if that has been
  // checked, not otherwise.
  return id->misc->ce_access_checked ?
    mac->access_chain_try (pp_workarea, id) : mac->access_try (pp_workarea, id);
}

int access_perm (object id)
// Returns the work area permission.
{
  // Note: This is chaining from PP_CONTENT_EDITOR if that has been
  // checked, not otherwise.
  return id->misc->ce_access_checked ?
    mac->access_chain_perm (pp_workarea, id) : mac->access_perm (pp_workarea, id);
}

mixed access_throw (object id, void|mapping vcerror)
// Returns the AC response for the request as if it has insufficient
// permissions. vcerror may be an error mapping returned from a VCFile
// or VCDir function. If it's an AC error on that object, a file/dir
// specific response is generated, otherwise 0 is returned.
{
  if (vcerror)
    return vcerror->type == "access denied" &&
      mac->access_throw (vcerror->ppoint, id, vcerror->reason);
  else
    return mac->access_throw (pp_workarea, id);
}

int rawview_access (object id)
  {return mac->access_perm (pp_workarea, id) == PERM_WRITE;}
mixed rawview_access_throw (object id)
  {return mac->access_throw (pp_workarea, id);}


// Loading and saving settings.

void set_settings (mapping(string:mixed) _settings)
{
  site->type_check (settings->name, stringp, "work area 'name'", "string");
  if (_settings->name && !valid_name (_settings->name)) return;
  settings = _settings;
  name = settings->name && clean_name (settings->name);
}

mapping(string:mixed) get_settings()
{
  settings->name = name;
  return settings;
}


// Work area handling.

void create (object _site, string _tag, mapping(string:mixed) _settings, void|int first)
{
  site = _site, tag = _tag;
  set_settings (_settings);
  tagdir = site->storage + "wa" + tag + "/";

  mac = site->mac;
  ac_find_file_pp = site->ac_find_file_pp, fileperm = site->fileperm;
  ac_init();

  string viewdir = tagdir + "view";

  if (!file_stat (viewdir)) {
#ifdef SBDEBUG
    report_notice ("SB: Checking out view area " + name + "\n");
#endif
    if (!mkdirhier (viewdir + "/"))
      throw (({"Couldn't create directory " + viewdir + "\n", backtrace()}));
    checkout_repository (site->storage + "cvsroot", "site", viewdir, tag == "" ? 0 : tag);
  }

  viewtopdir = cachedir (viewdir,
			 0,	// no parent
			 1,	// automatic update on viewarea
			 0, site->storage + "cvsroot");

  if (first) {
    array(string) stray = ({});
    foreach (viewtopdir->dir->cleanup_locks(), array(string) lock)
      if (sizeof (lock) > 2) stray += ({lock[0] + ": " + lock[2]});
    if (sizeof (stray))
      report_warning ("SB: Failed to removed stray CVS lock files:\n" +
		      stray * "\n" + "\n");
  }
}


// Accessing files and directories.

object vcobj (string loc, object id, void|int undelete_mode_not_used,
	      void|string revision, void|string date)
// Returns a VC object for the dir or file in the given location, or 0
// if one doesn't exist (the VC object that is; the dir or file may
// still not exist, depending on what one means with "exist"). If id
// is 0, the admin user edit area is used. If id isn't 0, it's assumed
// that the request has been fed through access_try().
{
#ifdef SBDEBUG
  if (id && !mac->id_is_known (id)) throw (({"Unidentified request\n", backtrace()}));
#endif

  // Thorough path wash.
  array(string) path = combine_path ("/", loc)[1..] / "/";
  if (path[-1] == "") path = path[..sizeof (path) - 2];
  if (sizeof (path)) path[-1] = clean_name (path[-1]);

  int ppid = ac_find_file_pp (path), perm, userid;
  if (id) {
    perm = fileperm (path, id, ppid);
    userid = perm == PERM_WRITE && mac->id_get_id (id);
  }
  else perm = PERM_WRITE, userid = site->admin_userid;

  object vcobj = _vcobj (userid, path);

  if (!vcobj) return 0;
  vcobj->reset_caches (Perm (perm, ppid));

  if (revision || date) vcobj = _vcoldobj (vcobj, revision, date);

  return vcobj;
}

object(VCFile) vcfile (string loc, object id)
// Returns a VCFile for the given location, regardless whether such a
// file exists or not and if it can be created or not. If id is 0, the
// admin user edit area is used. If id isn't 0, it's assumed that the
// request has been fed through access_try().
{
#ifdef SBDEBUG
  if (id && !mac->id_is_known (id)) throw (({"Unidentified request\n", backtrace()}));
#endif

  // Thorough path wash.
  array(string) path = combine_path ("/", loc)[1..] / "/";
  if (path[-1] == "") path = path[..sizeof (path) - 2];
  if (sizeof (path)) path[-1] = clean_name (path[-1]);

  int ppid = ac_find_file_pp (path), perm, userid;
  if (id) {
    perm = fileperm (path, id, ppid);
    userid = perm == PERM_WRITE && mac->id_get_id (id);
  }
  else perm = PERM_WRITE, userid = site->admin_userid;

  object(VCFile)|object(VCDir) vcobj = _vcobj (userid, path);

  if (vcobj && vcobj->isfile) {
    vcobj->reset_caches (Perm (perm, ppid));
    return vcobj;
  }
  return VCFile (this_object(), path, 0, userid, 0, Perm (perm, ppid));
}

object(VCDir) vcdir (string loc, object id)
// Returns a VCDir for the given location, regardless whether such a
// directory exists or not and if it can be created or not. If id is
// 0, the admin user edit area is used. If id isn't 0, it's assumed
// that the request has been fed through access_try().
{
#ifdef SBDEBUG
  if (id && !mac->id_is_known (id)) throw (({"Unidentified request\n", backtrace()}));
#endif

  // Thorough path wash.
  array(string) path = combine_path ("/", loc)[1..] / "/";
  if (path[-1] == "") path = path[..sizeof (path) - 2];
  if (sizeof (path)) path[-1] = clean_name (path[-1]);

  int ppid = ac_find_file_pp (path), perm, userid;
  if (id) {
    perm = fileperm (path, id, ppid);
    userid = perm == PERM_WRITE && mac->id_get_id (id);
  }
  else perm = PERM_WRITE, userid = site->admin_userid;

  object(VCFile)|object(VCDir) vcobj = _vcobj (userid, path);

  if (vcobj && vcobj->isdir) {
    vcobj->reset_caches (Perm (perm, ppid));
    return vcobj;
  }
  return VCDir (this_object(), path, 0, userid, 0, Perm (perm, ppid));
}


// Creating default files

void create_default_files (void|string from, void|object(VCDir) to)
{
  if(!from)
    from = default_dir;
  if(!to)
    to = vcdir ("/", 0);

  array(string) files = get_dir (from);

  if(!files)
    throw (({"Can't access directory " + from + "\n", backtrace ()}));

  foreach(files, string file)
    if(valid_name (file[..1] == "__" ? file[2..] + "a" : file)) {
      string afile = combine_path (from, file);
      array(int) st = file_stat (afile);
      if(!st)
	throw (({"Can't stat " + afile + "\n", backtrace ()}));
      if(st[1]!=-2) {
	mapping|object(VCFile) vf = to->newfile (0, file, 1);
	if(!objectp (vf))
	  throw (({"Can't create file " + to->abspath (0, 0, 1) + "/" +
		   file + "\n", backtrace ()}));
	mapping|object(VCFileData) u = vf->upload (0, 0, 0, 1);
	if(!objectp (u))
	  throw (({"Can't upload file " + to->abspath (0, 0, 1) + "/" +
		   file + "\n", backtrace ()}));
	object f = u->fd;
	// Let's cheat a bit...  :-)
	f->seek (0);
	string fdat = Stdio.read_bytes (afile);
	if(!fdat) {
	  u->upload_end (1);
	  throw (({"Can't read file " + afile + "\n", backtrace ()}));
	}
	f->write (fdat);
	fdat = 0;
	mapping|void ue = u->upload_end ();
	if(ue)
	  throw (({"Can't upload_end file " + to->abspath (0, 0, 1) + "/" +
		   file + "\n", backtrace ()}));

	// Fill out missing metadata.
	mapping|object md = vf->metadata (0, 0, 1);
	if (mappingp (md)) throw (({"Couldn't get metadata for " +
				    vf->real_abspath() + "\n", backtrace()}));
	md = MetaData.complete_metadata (site, md->md + ([]));
	if (mappingp (vf->set_metadata (0, md, 1)))
	  throw (({"Couldn't set metadata for " + vf->real_abspath() + "\n",
		   backtrace()}));

	mapping|void c =
	  vf->commit (0, "Default file created by SiteBuilder", 1);
	if(c)
	  throw (({"Can't commit file " + to->abspath (0, 0, 1) + "/" +
		   file + "\n", backtrace ()}));	
      } else {
	mapping|object(VCDir) vd =
	  to->newdir (0, file, "Default dir created by SiteBuilder", 1);
	if(!objectp (vd))
	  throw (({"Can't create directory " + to->abspath(0, 0, 1) + "/" +
		   file + "\n", backtrace ()}));
	create_default_files (afile, vd);
      }
    }
}
