// Module to handle a SiteBuilder site with cvs files under AC control.

// $Id: module.pmod,v 1.42 1999/08/29 18:22:23 mast Exp $

#if constant (roxen)
#include <config.h>
#endif


// Debugging tools.

void sbdebug (string msg, mixed ... args)
{
#ifdef SBDEBUG
  if (sizeof (args)) msg = sprintf (msg, @args);
  werror ("SB: " + msg);
#endif
}

string format_error (mapping error)
{
  if (!mappingp (error) || !error->type) return sprintf ("%O", error);
  array(string) msg = ({});
  foreach (({"operation", "ppoint", "state", "reason"}), string s)
    if (error[s]) msg += ({s + " " + error[s]});
  return sizeof (msg) ? error->type + ": " + msg * ", " : error->type;
}


// Misc functions.

object invalid_name_regexp = Regexp ("^$|[#\\/:<>|!&'`\"*?\1-\037\177-\237]|^\\.|\\.$|"
				     "^(CVS|Attic|#cvs\\.lock)$|"
				     ".~|^__");

int valid_name (string name)
// Returns nonzero if name is valid for a file or directory.
{
  // Unfortunately Regexp() doesn't like \0.
  return (search(name, "\0") == -1) && !invalid_name_regexp->match (name);
}

string clean_name (string name)
// Cleans a name for use as a file or directory name.
{
  // Trim any leading and trailing spaces, and compact multiple spaces
  // in the middle to one space.
  return (name / " " - ({""})) * " ";
}

string error_msg (mapping error)
{
  string msg = 0;
  switch (error->type) {
    case "access denied":
      if (!error->reason) msg = "Permission denied."; break;
    case "invalid file":
      msg = "This is not a valid file."; break;
    case "invalid directory":
      msg = "This is not a valid directory."; break;
    case "invalid operation":
      if (!error->reason) msg = "Invalid operation."; break;
    case "invalid name":
      msg = "The name '" + error->name + "' is invalid."; break;
    case "warning":
      if (!error->reason) msg = "You might not want to do that."; break;
    case "interrupted":
      msg = "The operation was interrupted by another on the same local file."; break;
    case "repository changed":
      msg = "The repository file has changed during the conflict resolution."; break;
    case "removed revision":
      msg = "Can't revert to a deleted revision."; break;
    case "io error":
      msg = "I/O error: "+strerror(error->errno); break;
  }
  if (error->reason) msg = error->reason + (msg ? " " + msg : "");
  return msg;
}

string nice_abspath (object vcobj, void|object id, void|int undelete_mode)
{
  if (!vcobj) return 0;
  string path = id ? vcobj->abspath (id, undelete_mode) : vcobj->real_abspath();
  if (vcobj->isdir) path += "/";
  if (path[0] != '/') path = "/" + path;
  return path;
}


// Internal helpers.

string pike_encode (mixed val)
{
#if __VERSION__ < 0.6
  if (stringp (val))
    return "\"" + replace (val, ({"\\", "\"", "\n"}), ({"\\\\", "\\\"", "\\n"})) + "\"";
  else
#endif
    return sprintf ("%O", val);
}

void rm_rf(string dir)
{
  array a=get_dir(dir);
  if (!a) return;
  a-=({".",".."});
  foreach (a,string file)
  {
    array s=file_stat(dir+"/"+file);
    if (s && s[1]==-2)
      rm_rf(dir+"/"+file);
    else
      rm(dir+"/"+file);
  }
  rm(dir);
}

string sillycaps_mangle (string name)
// Used to get a branch tag.
{
  name = replace (name, "\"`()[]{}" / "", "'''''''''" / "") - "'";
  name = replace (name, "\t\n!@#$%&/=+?\\^~*<>|,;.:-_" / "",
		  "                          " / "");
  array(string) words = name / " " - ({""});
  int wordlen = max (20 / sizeof (words), 4) - 1;
  string mangled = "";
  foreach (words, string word) mangled += String.capitalize (word[..wordlen]);
  return mangled[..19];
}

string filename_mangle (string name)
// Used to get a suitable file name without too strange characters. No
// guarantees about preserving identity or anything, though.
{
  name = replace (name, "\"`()[]{}" / "", "'''''''''" / "") - "'";
  name = replace (name, " \t\n&/?\\*<>|" / "", "___________" / "");
  return (name / "_" - ({""})) * "_";
}

array(string) server_urls (object conf)
{
  if (!conf) return ({});

  array(string) urls = ({});
  foreach (conf->query ("Ports"), array port) {
    string prt;

    switch(port[1][0..2])
    {
      case "ssl":
	prt = "https://";
	break;
      case "ftp":
	prt = "ftp://";
	break;
      
      default:
	prt = port[1]+"://";
    }
    if(port[2] && port[2]!="ANY")
      prt += port[2];
    else
#if efun(gethostname)
      prt += (gethostname()/".")[0] + "." + conf->query ("Domain");
#else
    ;
#endif
    urls += ({prt + ":"+port[0]+"/"});
  }

  return urls;
}

array(array(array)) diff3 (array a, array b, array c)
{
  mixed err = catch {return Array.diff3 (a, b, c);};

  if (arrayp (err) && sizeof (err) && stringp (err[0]))
    // Generate a better bug description in case of errors in Array.diff3().
    err[0] += sprintf ("diff3 input:\na = %O\nb = %O\nc = %O\n", a, b, c);
  throw (err);
}


// Helpers for internal VCFile/VCDir metadata handling.

mapping set_md (mapping base_md, mapping metadata)
// Preserves the internal variables from the old metadata.
{
  mapping md = metadata + ([]);
  if (base_md->__added) md->__added = 1;
  if (base_md->__ancestor_removed) md->__ancestor_removed = 1;
  if (base_md->__jointag) md->__jointag = base_md->__jointag;
  if (base_md->__create_rev) md->__create_rev = base_md->__create_rev;
  if (base_md->__join_revs) md->__join_revs = base_md->__join_revs;
  if (base_md->__removed) md->__removed = 1;
  return md;
}

string next_revision (string rev)
{
  if (!rev || !sizeof (rev)) return "new";
  array(string) parts = rev / ".";
  return parts[..sizeof (parts) - 2] * "." + "." + ((int) parts[-1] + 1);
}

mapping get_join_revs (mapping md)
{
  if (!mappingp (md->__join_revs)) {
    mapping jr = ([]);
    if (stringp (md->__join_revs)) {
      foreach (md->__join_revs / ",", string part) {
	string tag, rev;
	int time;
	if (sscanf (part, "%s:%s:%d", tag, rev, time) == 3) jr[tag] = ({rev, time});
      }
    }
    md->__join_revs = jr;
  }
  return md->__join_revs;
}

mapping encode_md (mapping md)
{
  if (md) {
    if (mappingp (md->__join_revs)) {
      array jr = ({});
      foreach (indices (md->__join_revs), string tag)
	jr += ({sprintf ("%s:%s:%d", tag,
			 md->__join_revs[tag][0], md->__join_revs[tag][1])});
      md->__join_revs = jr * ",";
    }
  }
  return md;
}


// Mutex lock helpers.

#ifdef THREADS

array(object) edit_lock_2 (object/*(VCFile)*/ vcfile_1, object/*(VCFile)*/ vcfile_2)
// Lock two files in a canonical order to avoid deadlocks.
{
  object edit_lock_1, edit_lock_2;
  string cname_1 = vcfile_1->wa->tag + "/" + vcfile_1->real_abspath();
  string cname_2 = vcfile_2->wa->tag + "/" + vcfile_2->real_abspath();
  catch {			// Ignore recursive locks.
    if (cname_1 < cname_2) {
      edit_lock_1 = vcfile_1->edit_mutex->lock();
      edit_lock_2 = vcfile_2->edit_mutex->lock();
    }
    else if (cname_2 < cname_1) {
      edit_lock_2 = vcfile_2->edit_mutex->lock();
      edit_lock_1 = vcfile_1->edit_mutex->lock();
    }
    else			// Same file.
      edit_lock_1 = vcfile_1->edit_mutex->lock();
  };
  return ({edit_lock_1, edit_lock_2});
}

array(object) wa_update_lock_2 (object/*(Workarea)*/ wa_1, object/*(Workarea)*/ wa_2)
// Lock two work areas in a canonical order to avoid deadlocks.
{
  object wa_lock_1, wa_lock_2;
  catch {			// Ignore recursive locks.
    if (wa_1->tag < wa_2->tag) {
      wa_lock_1 = wa_1->wa_update_mutex->lock();
      wa_lock_2 = wa_2->wa_update_mutex->lock();
    }
    else if (wa_1->tag < wa_2->tag) {
      wa_lock_2 = wa_2->wa_update_mutex->lock();
      wa_lock_1 = wa_1->wa_update_mutex->lock();
    }
    else			// Same work area.
      wa_lock_1 = wa_1->wa_update_mutex->lock();
  };
  return ({wa_lock_1, wa_lock_2});
}

#else  // THREADS

array(object) edit_lock_2 (object/*(VCFile)*/ vcfile_1, object/*(VCFile)*/ vcfile_2)
{
  sbdebug ("edit_lock_2 called without THREADS define\n" +
	   describe_backtrace (backtrace()));
  return ({class {}(), class {}()});
}

array(object) wa_update_lock_2 (object/*(Workarea)*/ wa_1, object/*(Workarea)*/ wa_2)
{
  sbdebug ("wa_update_lock_2 called without THREADS define\n" +
	   describe_backtrace (backtrace()));
  return ({class {}(), class {}()});
}

#endif


// VCFile/VCDir extras.

class VCLogEntry
{
  string revision;
  string date;			// Format: "yyyy/mm/dd hh:mm:ss".
  string action;		// "create", "edit", "delete", or "undelete".
  int userid;			// AC user ID.
  string user;
  string message;

  // The following are set for joins.
  string workarea_tag;		// The work area tag of the joined file.
  string join_revision;		// The other file revision used when doing the join.
}

array(object(VCLogEntry)) convert_log (array(object) lst)
{
  for (int i = 0; i < sizeof (lst); i++) {
    object(VCLogEntry) vcle = VCLogEntry();
    vcle->revision = lst[i]->revision;
    vcle->date = lst[i]->date;
    string actions, rest;
    sscanf (lst[i]->message, "%s by %d, %s, %*[^;\n]%s\n%s",
	    actions, vcle->userid, vcle->user, rest, vcle->message);
    actions = (actions / ", ")[0];
    vcle->action = (["join": "edit", "remove": "delete"])[actions] || actions;
    if (rest && sizeof (rest))
      sscanf (rest, "; join %s, %s", vcle->workarea_tag, vcle->join_revision);
    lst[i] = vcle;
  }
  return lst;
}

class Perm
// Used to lump together a permission and the ppoint it comes from to
// avoid races.
{
  int perm, ppid;
  void create (int _perm, int _ppid) {perm = _perm, ppid = _ppid;}
}

// Initialize these on load, since it's much easier to add messages in
// the reverse direction compared to the one suited for fast lookups.

mapping(string:mapping(string:string)) vcfile_invalid_op_msgs = ([
	"N/N": ([]),	"N/E": ([]),					"N/R": ([]),
			"E/E": ([]),	"E/M": ([]),	"E/MR": ([]),
	"A/N": ([]),			"A/M": ([]),	"A/MR": ([]),
			"M/E": ([]),	"M/M": ([]),	"M/MR": ([]),	"M/R": ([]),
			"R/E": ([]),					"R/R": ([]),
]);
mapping(string:mapping(string:string)) vcdir_invalid_op_msgs = ([
  "N": ([]), "E": ([]), "R": ([]),
]);

void add_msg (mapping msgs, string|array(string) op, array(string) stateglobs, string msg)
{
  foreach (stateglobs, string stateglob)
    foreach (glob (stateglob, indices (msgs)), string state)
      if (arrayp (op)) foreach (op, string o) msgs[state][o] = msg;
      else msgs[state][op] = msg;
}

#define VCFILEMSG(op, states, msg) add_msg (vcfile_invalid_op_msgs, op, states, msg)
#define VCDIRMSG(op, states, msg) add_msg (vcdir_invalid_op_msgs, op, states, msg)

void create()
{
  VCFILEMSG ("ascend", ({"*"}),
	     "The directory isn't accessible.");
  VCFILEMSG (({"view", "metadata", "download"}), ({"N/N", "N/R", "R/R"}),
	     "The file doesn't exist.");
  VCFILEMSG (({"view", "metadata", "download"}), ({"R/E"}),
	     "The file doesn't exist locally.");
  VCFILEMSG ("add", ({"N/E", "E/*", "A/*", "M/*"}),
	     "A file or directory with that name already exists.");
  VCFILEMSG (({"add", "upload"}), ({"N/R", "R/R"}),
	     "A deleted directory with that name exists. "
	     "It must be pruned before a file can take its name.");
  VCFILEMSG ("add", ({"R/E"}),
	     "The delete must be committed before a new file can be added.");
  VCFILEMSG (({"set_metadata", "upload"}), ({"N/E"}),
	     "To avoid lost updates, it's required that there already is a "
	     "local copy of the file.");
  // FIXME: This could also be the reason above:
  //	     "A directory with that name already exists."
  VCFILEMSG ("upload", ({"*/MR"}), // Warning.
	     "The file being changed has been deleted in the repository.");
  VCFILEMSG (({"set_metadata", "upload"}), ({"R/E"}),
	     "Can't change a file marked for deletion.");
  VCFILEMSG ("discard", ({"N/*", "R/R"}),
	     "There's no local copy to discard.");
  VCFILEMSG ("discard", ({"A/*", "M/*", "R/E"}), // Warning.
	     "There are local changes that will be lost.");
  VCFILEMSG ("update", ({"N/*", "M/E", "M/R", "R/E", "R/R"}),
	     "The file is already up to date.");
  VCFILEMSG ("update", ({"E/*"}),
	     "Use discard to remove the local copy."); // Not a user op.
  VCFILEMSG ("update", ({"A/N", "A/MR"}),
	     "There are no file in the repository to update to.");
  VCFILEMSG ("delete", ({"N/N", "N/R", "R/R"}),
	     "No file to mark for deletion.");
  VCFILEMSG ("delete", ({"A/*", "M/*"}), // Warning.
	     "There are local changes that will be lost.");
  VCFILEMSG ("delete", ({"R/E"}),
	     "The file already marked for deletion.");
  VCFILEMSG ("undelete", ({"N/N", "N/E", "E/E", "E/M", "A/N", "A/M", "M/E", "M/M"}),
	     "Not a deleted file.");
  VCFILEMSG ("undelete", ({"A/MR", "M/MR", "M/R"}),
	     "The file already exists locally.");
  VCFILEMSG ("revert", ({"N/N", "A/N"}),
	     "No such file in the repository.");
  VCFILEMSG ("revert", ({"A/M", "A/MR", "M/*"}), // Warning.
	     "There are local changes that will be overwritten.");
  VCFILEMSG ("revert", ({"R/E"}), // Warning.
	     "The mark for deletion will be overridden.");
  VCFILEMSG ("commit", ({"N/*", "E/*", "R/R"}),
	     "No changes to commit.");
  VCFILEMSG ("commit", ({"A/M", "M/M", "M/MR"}),
	     "The repository has changed; use update first.");
  // "copy" - handled as "view" at source, as "upload" at destination.
  // "move" - handled as "view" and "delete" at source, as "upload" at destination.
  VCFILEMSG ("join", ({"E/*", "A/*", "M/*", "R/*"}),
	     "There are local copies in one or both work areas; "
	     "use discard or update/commit first.");
  VCFILEMSG ("purge", ({"N/N"}),
	     "The file doesn't exist.");
  VCFILEMSG ("purge", ({"E/*", "A/*", "M/*", "R/*"}),
	     "There's a local copy; use discard first.");

  VCDIRMSG (({"metadata", "list", "list_recursive"}), ({"N", "R"}),
	    "The directory doesn't exist.");
  VCDIRMSG ("newfile", ({"N", "R"}),
	    "The directory to create the new file in doesn't exist.");
  VCDIRMSG ("newdir", ({"N", "R"}),
	    "The directory to create the new directory in doesn't exist.");
  VCDIRMSG ("add", ({"E"}),
	    "A file or directory with that name already exists.");
  VCDIRMSG ("add", ({"R"}),
	     "A deleted file with that name exists. "
	     "It must be pruned before a directory can take its name.");
  VCDIRMSG ("set_metadata", ({"E"}), // FIXME: Not ret by set_metadata.
	    "A file with that name already exists.");
  VCDIRMSG ("delete", ({"N", "R"}),
	    "No directory to delete.");
  VCDIRMSG ("delete", ({"E"}),	// Warning.
	    "The directory is not empty.");
  VCDIRMSG ("undelete", ({"N", "E"}),
	    "Not a deleted directory.");
  // "copy" - handled as "list" at source, as "add" at destination.
  // "move" - handled as "list" and "delete" at source, as "add" at destination.
  VCDIRMSG ("purge", ({"N"}),
	    "The directory doesn't exist.");
}
